program dirgra(dirgrap, list, output);
(* dirgra: directed graph Monte Carlo analysis

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.08; (* of dirgra.p 2016 Jan 24
2016 Jan 24, 1.08: remove reliance on random(0)
  sed -i.bak 's|seed = 0|seed := 0|g'               dirgra.p
2016 Jan 12, 1.07: remove reliance on random(0)
1995 mar 28, 1.06: running version
origin 1994 Dec 26,
Mon Dec 26 18:10:03 EST 1994 start
Mon Dec 26 18:58:55 EST 1994 reads graph and writes it out correctly
Mon Dec 26 20:13:21 EST 1994 run with 10000 iterations on a Sun Sparcstation
   20/61 this took 21.5 seconds.

*)
(* end module version *)

(* begin module describe.dirgra *)
(*
name
   dirgra: directed graph Monte Carlo analysis

synopsis
   dirgra(dirgrap, list: in, output: out)

files
   dirgrap:  Parameters to control the program
      first line: number of iterations to perform
      second line: numbers of the first and the last node in the net.
      later lines: each line contains two numbers that define the graph links.
   list:  analysis of the graph
   output: messages to the user

description
   Do a Monte Carlo flow analysis of a directed graph

examples

The following dirgrap defines Figure 1 of Adleman1994:

60000   number of iterations to perform
0 6    numbers of the first and last nodes
0 3
0 1
0 6
1 2
1 3
2 3
2 1
3 2
3 4
4 1
4 5
5 1
5 2
5 6

and gives this list file as a result:

dirgra 1.06
* first node: 0
*  last node: 6
* Nodes:
* 0 to:  3, 1, 6
* 1 to:  2, 3
* 2 to:  3, 1
* 3 to:  2, 4
* 4 to:  1, 5
* 5 to:  1, 2, 6

* Nodes with traversal counts and percentages: for 1811570 steps
* 0 to:  3 (19981, 1%), 1 (19932, 1%), 6 (20087, 1%)
* 1 to:  2 (212247, 12%), 3 (212948, 12%)
* 2 to:  3 (244654, 14%), 1 (245891, 14%)
* 3 to:  2 (238613, 13%), 4 (238970, 13%)
* 4 to:  1 (119693, 7%), 5 (119277, 7%)
* 5 to:  1 (39679, 2%), 2 (39685, 2%), 6 (39913, 2%)

documentation

@article{Adleman1994,
author = "L. M. Adleman",
title = "Molecular Computation of Solutions to Combinatorial Problems",
journal = "Science",
volume = "266",
pages = "1021-1024",
year = "1994"}

see also

author
   Thomas Dana Schneider

bugs

technical notes
   maxnodes and maxlinks are constants that define the maximum
   number of nodes and links the program can handle.  The program
   records nodes and links in an array for speed.

   The non-standard random(0) function is used to generate random numbers.  It
   is expected to give pseudo random numbers in the range 0 to 1.
   THIS WAS REMOVED on 2016 jan 12.

*)
(* end module describe.dirgra *)

(* begin module digra.const *)
   maxnodes = 50; (* maximum number of nodes + 1 that can be held *)
   maxlinks = 50; (* maximum number of links for each node. *)
(* end module digra.const *)

var
   dirgrap, list: text; (* files used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module random *)
(* random generator 1.  version = 1.00; 1986 December 31 *)
(* written by Thomas Schneider *)
procedure random(var seed: real);
(* WARNING: This random number generator is known to cycle on a Pyramid
90x computer.  It does not seem to cycle on a CDC Cyber.  The difference
is that the Pyramid has a 32 bit word, while the Cyber has 60 bits.
You should test the routine (use program tstrnd) before you use it.

   The seed value is used to generate a random number in the range 0 to 1.
The seed is then altered, to provide the seed of the next random number.
In other words, the seed is the random number.  To obtain the next random
number in the pseudo-random series, simply call random again with the same
seed.  Note that the series will not be truely random, since the previous
value determines the next value.  This is "pseudo-random".

An initial seed of 4.0 is suggested.  Note that this is not the first
random number.  You can use different seeds to get different pseudo-random
number series.

The expected mean and standard deviation of a flat distribution is
calculated using simple integrals:

       1
mean = / x dx = 0.5
       0
           1           2
variance = / (x - mean)  dx = 1/12;  stdev = sqrt(variance) = .2886751
           0

(The integral for the variance is solved easily by substituting y = x - mean.)

   The algorithm is derived from:
Pascal Programs For Scientists and Engineers, Alan R. Miller
Sybex, Berkeley CA (1981).
which in turn was an adaptation from HP-35 applications programs. *)
const
      pi = 3.141592653589; (* a value used to help mix up the result *)
var
      intermediate: real; (* an intermediate result *)
begin (* random *)
      intermediate := seed + pi;
      intermediate := exp(5.0*ln(intermediate));
      seed := intermediate - trunc(intermediate)
end; (* random *)
(* end module random *)

(* begin module dirgra.themain *)
procedure themain(var dirgrap, list: text);
(* the main procedure of the program *)
var
   firstnode: integer; (* start node number *)
   i: integer; (* index to iterations *)
   iterations: integer; (* number of traversals of the graph to do *)
   lastnode: integer; (* end node number *)
   n: integer; (* index to net nodes *)
   l: integer; (* index to net links *)
   net: array[0..maxnodes, 1..maxlinks] of integer; (* the net of links *)
   count: array[0..maxnodes, 1..maxlinks] of integer; (* number traversals *)
   links: array[0..maxnodes] of integer; (* number of links for each node *)
   node1, node2: integer; (* a directed segment goes from node 1 to node 2 *)
   nextnode: integer; (* next node chosen in the graph *)
   r: integer; (* randomly chosen link number *)
   steps: integer; (* total number of steps across the graph *)
   seed: real; (* a random number *)
begin
   writeln(output,'dirgra ',version:4:2);
   rewrite(list);
   writeln(list,'dirgra ',version:4:2);

   seed := 0; (* initial random seed *)

   (* clear the net *)
   for n := 0 to maxnodes do begin
      for l := 1 to maxlinks do begin
         net[n,l] := 0;
         count[n,l] := 0;
      end;
      links[n] := 0;
   end;

   (* read in the parameters and graph *)
   reset(dirgrap);
   readln(dirgrap, iterations);
   readln(dirgrap, firstnode, lastnode);
   if (firstnode < 0) or (lastnode < 0) then begin
      writeln(output,'first node and last node must be positive integers');
      halt;
   end;

   while not eof(dirgrap) do begin
      readln(dirgrap,node1,node2);
      links[node1] := succ(links[node1]);
      net[node1,links[node1]] := node2
   end;

   (* display the net *)
   writeln(list,'* first node: ',firstnode:1);
   writeln(list,'*  last node: ',lastnode:1);
   writeln(list,'* Nodes:');
   n := 0;
   while links[n] > 0 do begin
      write(list,'* ',n:1,' to: ');
      for l := 1 to links[n] do begin
         if l > 1 then write(list,',');
         write(list,' ',net[n,l]:1);
      end;
      writeln(list);
      n := succ(n);
   end;

   steps := 0;
   for i := 1 to iterations do begin
      writeln(output,'------ Iteration ',i:1,' ------');
      n := firstnode;
      repeat
         (* chose next node *)
         random(seed);
         r := trunc(links[n]*seed) + 1;
         nextnode := net[n,r];

         write(output,'at ',n:1, ' with ',links[n]:1,' links; ');
         write(output,'chose number ',r:1,' so ');
         writeln(output,'next node is ',nextnode:1);

         count[n,r] := succ(count[n,r]);
         steps := succ(steps);
         n := nextnode
      until n = lastnode
   end;

   (* display the net *)
   writeln(list);
   writeln(list,'* Nodes with traversal counts and percentages:',
                ' for ',steps:1,' steps');
   n := 0;
   while links[n] > 0 do begin
      write(list,'* ',n:1,' to: ');
      for l := 1 to links[n] do begin
         if l > 1 then write(list,',');
         write(list,' ',net[n,l]:1);
         write(list,' (',count[n,l]:1);
         write(list,', ',round(100*(count[n,l]/steps)):1,'%)')
      end;
      writeln(list);
      n := succ(n);
   end;

end;
(* end module dirgra.themain *)

begin
   themain(dirgrap, list);
1: end.
