program encsum(encseq,sum,output);
(* encsum: sums the vectors in an encoded sequence file
   by gary stormo
module libraries needed: delman, delmods, matmods  *)

label 1;  (* end of program *)

const
(* begin module version *)
version = 1.21; (* of encsum.p 1994 sep 5
origin: 1982 nov 15 *)
(* end module version *)

(* begin module describe.encsum *)
(*
name
      encsum: sum of the vectors of encoded sequences

synopsis
      encsum(encseq: in, sum: out, output: out)

files
      encseq: the file of encoded sequences; this is the output of
            the program 'encode'
      sum: the output of this program
      output: for messages to the user

description
      this program takes as input a file of encoded sequences, from the
      encode program, and sums the individual sequence vectors into one
      vector of their sums.  this is useful for doing histograms or
      compositions of many sequences.

see also
      encode.p

author
      gary stormo

bugs
      none known
*)
(* end module describe.encsum *)

(* begin module vector.const *)
      maxvecpart = 64; (* the number of elements in a 'part' of a vector *)
      vpagewidth = 64; (* the width (in characters) of the output vector
         file from procedure writevector *)
(* end module vector.const version = 'matmod 1.95 85 apr 18 tds/gds'; *)

type
(* begin module vector.type *)
(* vectors are useful for many applications, including the encoding of
   sequences.  this vector type is designed to be flexible enough to be
   used whenever one needs a vector.  it is designed as a linked list so
   it can contain as many elements as are ever needed.  the actual 'vector'
   is a record containing the length and a pointer to the first 'vectorpart'.
   that vectorpart is a record containing an array of the first 'maxvecpart'
   elements  (maxvecpart is a constant, from module vector.const, which must
   be included) and a pointer to the next 'vectorpart'.  the elements are
   of type real so that either integers or reals may be used.  *)
      partptr = ^vectorpart;
      vectorpart = record
         numbers : array[1..maxvecpart] of real;
         next : partptr;
      end;

      vector = record
         length : integer;
         part : partptr;
      end;
(* end module vector.type version = 'matmod 1.95 85 apr 18 tds/gds'; *)

var
(* begin module encsum.var *)
      encseq, (* the input encoded sequences *)
      sum: text;  (* the output vector sum *)
(* end module encsum.var *)

(* begin module package.primitive *)
(* ************************************************************************ *)
(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module copylines *)
function copylines(var fin, fout: text; n: integer): integer;
(* copy n lines of file fin to file fout.
the actual number of lines copied is returned. *)
var
      index: integer; (* the current line number *)
begin (* copylines *)
      index := 0;
      while (not eof(fin)) and (index < n) do begin
         copyaline(fin, fout);
         index := succ(index)
      end;

      copylines := index
end; (* copylines *)
(* end module copylines version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* ************************************************************************ *)
(* end module package.primitive version = 'delmod 6.51 85 apr 17 tds/gds' *)

(* begin module vector.functions *)
(* *********************************************************************** *)
(* begin module vector.io *)
(* *********************************************************************** *)
(* begin module vector.primitives *)
(* these functions and procedures do manipulations on vectors. *)

function vget(v:vector; pos:integer): real;
(* this returns from vector 'v' the value of the element at position 'pos' *)
var i: integer;
begin
      if (pos > v.length) or (pos < 1) then begin
         writeln(output,' error in call to function vget:',
            ' position ',pos:1,' is beyond the end of the vector');
         halt;
      end;

      (* move to the correct 'vectorpart' *)
      for i := 1 to ((pos -1) div maxvecpart) do  v.part := v.part^.next;

      (* get the proper array element from this part *)
      vget := v.part^.numbers[(pos -1) mod maxvecpart + 1];
end;

procedure vput(var v:vector; pos:integer; number: real);
(* this puts into vector 'v' the value of 'number' at position 'pos' *)
var i: integer;
      firstpart: partptr;  (* of the vector *)
begin
      if (pos > v.length) or (pos < 1) then begin
         writeln(output,' error in call to function vput:',
            ' position ',pos:1,' is beyond the end of the vector');
         halt;
      end;

      (* move to the correct 'vectorpart' *)
      firstpart := v.part;
      for i := 1 to ((pos -1) div maxvecpart) do  v.part := v.part^.next;

      (* put the 'number' into the proper array element for this part *)
      v.part^.numbers[(pos -1) mod maxvecpart + 1] := number;
      v.part := firstpart;
end;

procedure makevector(var v: vector; l: integer);
(* create the linked list of vector-parts which are needed for a vector
   of length 'l' *)
var   numparts,  (* number of parts needed for the vector *)
      i: integer; (* index to the parts of the vector *)
      firstpart,  (* of the vector *)
      newpart: partptr;  (* of the vector *)
begin
      if l < 1 then begin
         writeln(output,' makevector: positive length required');
         halt
      end;

      v.length := l;
      new(v.part);
      firstpart := v.part;
      numparts := (v.length - 1) div  maxvecpart + 1;
      for i := 1 to (numparts - 1) do begin
         new(newpart);
         v.part^.next := newpart;
         v.part := newpart;
      end;
      v.part^.next := nil;
      v.part := firstpart;
end;
(* end module vector.primitives version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module vector.readvector *)
procedure readvector(var thefile:text; var v: vector);
(* read the elements of v into v from 'thefile'.  v must already be set
   up (all the parts created and linked together) before calling.
   'thefile' must contain, from the cursor point until the end of the
    vector, only numbers, either integers or reals; otherwise it will
    bomb.  the vector ends with an end of line so that
    end of file can be tested for. *)
var i,j: integer;  (* indecies *)
      numparts,  (* number of parts in the vector *)
      lastpart: integer; (* the number of elements in the last vector part *)
      firstpart: partptr;  (* pointer to the first vector part *)
begin
      firstpart := v.part;
      numparts := (v.length - 1) div maxvecpart + 1;
      lastpart := (v.length - 1) mod maxvecpart + 1;
      for i := 1 to (numparts - 1) do begin
         for j := 1 to maxvecpart do read(thefile,v.part^.numbers[j]);
         v.part := v.part^.next;
      end;
      for j := 1 to lastpart do read(thefile,v.part^.numbers[j]);
      readln(thefile);
      v.part := firstpart;
end;
(* end module vector.readvector version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module vector.writevector *)
procedure writevector(var thefile: text; v: vector; y,z: integer);
(* writes the elements of 'v' to 'thefile'.  the integers y and z are:
    y: the size of the field for printing the number;
    z: the size of the decimal part of the field,
    if z = 0 then integers will be printed. *)
var   pos,   (* posititon in the vector *)
      i,j: integer;  (* indecies *)
      numparts,  (* number of parts in the vector *)
      lastpart: integer; (* the number of elements in the last vector part *)
      firstpart: partptr;  (* pointer to the first vector part *)
      x: integer; (* the number of elements to write on a line *)
begin
      x := trunc(vpagewidth/(y+1));
      pos := 0;
      firstpart := v.part;
      numparts := (v.length - 1) div maxvecpart + 1;
      lastpart := (v.length - 1) mod maxvecpart + 1;
      if (z = 0) then begin
         for i := 1 to (numparts - 1) do begin
            for j := 1 to maxvecpart do begin
               write(thefile,' ',round(v.part^.numbers[j]):y);
               pos := succ(pos);
               if (pos mod x = 0) then writeln(thefile);
            end;
            v.part := v.part^.next;
         end;
         for j := 1 to lastpart do begin
            write(thefile,' ',round(v.part^.numbers[j]):y);
            pos := succ(pos);
            if ((pos < v.length) and (pos mod x = 0)) then writeln(thefile);
         end
      end
      else begin
         for i := 1 to (numparts - 1) do begin
            for j := 1 to maxvecpart do begin
               write(thefile,' ',v.part^.numbers[j]:y:z);
               pos := succ(pos);
               if (pos mod x = 0) then writeln(thefile);
            end;
            v.part := v.part^.next;
         end;
         for j := 1 to lastpart do begin
            write(thefile,' ',v.part^.numbers[j]:y:z);
            pos := succ(pos);
            if ((pos < v.length) and (pos mod x = 0)) then writeln(thefile);
         end;
      end;
      writeln(thefile);
      v.part := firstpart;
end;
(* end module vector.writevector version = 'matmod 1.95 85 apr 18 tds/gds'; *)
(* *********************************************************************** *)
(* end module vector.io version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module vector.dotproduct *)
function dotproduct(vectora,vectorb: vector): real;
(* this returns the dotproduct of 'vectora' and 'vectorb' *)
var i: integer;  (* an index *)
    j: real;  (* partial products *)
begin
      if vectora.length <> vectorb.length then begin
         writeln(output,' function dotproduct: vector lengths must',
            ' be equal');
         halt
      end;
      j := 0;
      for i := 1 to vectora.length do
         j := j + vget(vectora,i) * vget(vectorb,i);
      dotproduct := j;
end;
(* end module vector.dotproduct version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module vector.magnitude *)
function magnitude(var v: vector): real;
(* find the magnitude (length) of the vector v *)
begin
      magnitude := sqrt(dotproduct(v,v));
end;
(* end module vector.magnitude version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module vector.normalize *)
procedure normalize(var v: vector);
(* this replaces the vector v with the congruent vector of unit length,
   i.e.,  the resulting vector v has the property that v.v = 1   *)
var i: integer;  (* index *)
      length: real;  (* of the unnormalized vector *)
begin
      length := magnitude(v);
      for i := 1 to v.length do
         vput(v,i,vget(v,i)/length);
end;
(* end module vector.normalize version = 'matmod 1.95 85 apr 18 tds/gds'; *)
(* *********************************************************************** *)
(* end module vector.functions version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module vector.vset *)
procedure vset(thevalue: real; var v: vector);
(* set the value thevalue into v *)
var   i: integer; (* an index *)
begin
      for i := 1 to v.length do vput(v,i,thevalue)
end;
(* end module vector.vset version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module vector.vadd *)
procedure vadd(a: vector; var b: vector);
(* add the vector a into the vector b *)
var   i: integer; (* an index *)
begin
      if a.length <> b.length then begin
         writeln(output,' function vadd: vector lengths must be equal');
         halt
      end;

      for i := 1 to a.length
         do vput(b,i,vget(b,i)+vget(a,i))
end;
(* end module vector.vadd version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module encode.enchead *)
procedure enchead(var encseq, fout: text; var length: integer);
(* copy the information from the start of the encoded vector
file encseq to fout.  to make fout have the same form as encseq,
it is necessary for the program which calls this procedure to
identify itself to the fout file on one line.
   the vector lengths is returned.  the encseq file is positioned
for reading of the vectors immediately after enchead.
   in contrast, the fout file is left at the end of the
vector length line so the user may add more information to the line.
after completing the line with writeln, a second writeln is
required before the vectors begin. *)
var   regions,  (* number of independently encoded regions *)
      i: integer; (* the number of lines of parameters *)
begin
      reset(encseq);
      readln(encseq); (* skip the previous program identification *)
      copyaline(encseq,fout); (* copy book information *)
      copyaline(encseq,fout); (* copy blank line *)
         (* copy the encoding parameters to the fout *)
      read(encseq,regions);
      write(fout,' ',regions:1);
      i := 6 * regions + 2;  (* 6 lines for each parameter and
         2 lines for the coded regions *)

      if copylines(encseq,fout,i) <> i then begin
         writeln(output,' encseq is missing parameters');
         halt
      end;

      (* now get the length *)
      readln(encseq,length);
      write(fout,' ',length:1,' is the vector length');
end;
(* end module encode.enchead version = 'matmod 1.95 85 apr 18 tds/gds'; *)

(* begin module encsum.themain *)
procedure themain(var encseq,sum: text);
(* the main procedure of the program *)
var
      sumvector,  (* the vector sum *)
      newvector: vector;  (* each new vector to be added to the sum *)
      i,   (* an index *)
      seqnum,  (* number of encoded sequences in encseq *)
      length: integer;  (* of the vectors *)
begin
      writeln(output,' encsum ',version:4:2);
      rewrite(sum);
      writeln(sum,' encsum version ',version:4:2,'; sum of vectors in');
      enchead(encseq,sum,length);

      makevector(sumvector,length);
      makevector(newvector,length);
      vset(0.0,sumvector);
      seqnum := 0;
      while not eof(encseq) do begin
         readvector(encseq,newvector);
         seqnum := succ(seqnum);
         vadd(newvector,sumvector);
      end;
      writeln(sum,';   ',seqnum:1,' sequences in this sum');
      writeln(sum);
      writevector(sum,sumvector,3,0);
end;
(* end module encsum.themain *)

begin
      themain(encseq,sum);
1: end.
