program fastcat(book, fastcatp, cat, output);
(* fastcat: create a Delila catalogue very quickly

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.13; (* of fastcat.p 2017 Jul 14
2017 Jul 14, 1.13: namelength/linelength set to 200/210 (see catal)
2017 Jul 10, 1.12: namelength set to 200 to allow even longer names
2007 Nov 23, 1.11: synopsis needs to mention fastcatp
2004 Sep  9, 1.10: upgrade for GPC
2001 Mar 20, 1.09: document limitation of program.
2001 Mar 15, 1.08: clean up
2001 Mar 15, 1.07: all book objects accepted.
2001 Mar 15, 1.06: functional!
2001 Mar 15, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.fastcat *)
(*
name
   fastcat: create a Delila catalogue very quickly

synopsis
   fastcat(book: in, fastcatp: in, cat: out, output: out)

files

   book:  A book of the Delila system.  This is preferably created
      by a reliable program.

   fastcatp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      NO OTHER PARAMETERS REQUIRED.

   cat:  A Delila catalogue file corresponding to the book.

   output: messages to the user

description

   The catal program checks the book structure before creating a delila
   catalogue entry.  When the book is large (>10,000 pieces) this takes a
   while and, for unknown reasons, catal slows down.  Fastcat does not check
   the book structure and so should be much faster than catal.

   WARNING: DO NOT USE THIS PROGRAM UNLESS THE BOOK STRUCTURE IS
   KNOWN TO BE CORRECT!

   WARNING: THIS PROGRAM WILL NOT CHANGE THE NAMES IN THE BOOK TO BE UNIQUE!
   YOU MUST DO THIS YOURSELF WITH ANOTHER TOOL!  The consequence is that if
   there are two pieces with the same name within an organism and chromosome
   Delila will only find the first instance of the piece.

examples

see also

   {parameter file:} fastcatp

   {normal (slower) catalogue creation program:} catal.p
   {program to look at the output of this program:} loocat.p
   {program containing standard book reading routines :} delmod.p

author

   Thomas Dana Schneider

bugs

technical notes

   The cat file is a file of records that contain the types, names and line
   numbers of the objects in the book.  Use the loocat program to look at
   it.

   The definitions of namelength, alpha, name, item and recordfile MUST match
   the catal and loocat definitions!

   The routine is based on standard routines in delmod.p, but trimmed to the
   bare essentials so that it should be fast.

*)
(* end module describe.fastcat *)

(* the definition of namelength MUST match the catal and loocat definitions!
*)

(* begin module book.const *)
(* constants needed for book manipulations *)
      dnamax = 1024;  (* length of dna arrays *)
      namelength = 200; (* maximum key name length *)
      linelength = 210; (* maximum line readable in book *)
(* end module book.const version = 7.67; {of delmod.p 2004 Sep 8} *)

(* this MUST match the value in delmod! *)
   datetimearraylength = 19; (* length of dataarray for dates *)

type
   (* the definitions of alpha and name MUST match the catal and loocat
      definitions! *)

(* begin module book.type ****************)
(* types needed for book manipulations *)

      chset = set of 'a'..'z';

      (* types defined in book definition *)

      alpha = packed array[1..namelength] of char; (* this is not alfa *)

      (* name is a left justified string with blanks following the
         characters *)
      name = record
         letters: alpha;
         length: 0..namelength (* zero means an unspecified structure *)
      end;

      lineptr = ^line;
      line = record (* a line of characters *)
         letters: packed array [1..linelength] of char;
         length: 0..linelength;
         next: lineptr
      end;

      direction = (plus, minus, dircomplement, dirhomologous);
      configuration = (linear, circular);
      state = (on, off);

      header = record (* header of key *)
         keynam: name; (* key name of structure *)
         fulnam: lineptr; (* full name of structure *)
         note:   lineptr (* note key *)
      end;

(* begin module base.type *)
   (* define the four nucleotide bases *)
   base = (a,c,g,t);
(* end module base.type version = 7.67; {of delmod.p 2004 Sep 8} *)

      (* sequence types *)
      dnaptr = ^dnastring;
      dnarange = 0..dnamax;
      seq = packed array[1..dnamax] of base;
      dnastring = record
         part: seq;
         length: dnarange;
         next: dnaptr
      end;


      orgkey = record (* organism key *)
         hea: header;
         mapunit: lineptr (* genetic map units *)
      end;

      chrkey = record (* chromosome key *)
         hea: header;
         mapbeg: real; (* number of genetic map beginning *)
         mapend: real (* number of genetic map ending *)
      end;

      pieceptr = ^piece;
      piekey = record (* piece key *)
         hea: header;
         mapbeg: real; (* genetic map beginning *)
         coocon: configuration; (* configruation (circular/linear) *)
         coodir: direction; (* direction (+/-) relative to genetic map *)
         coobeg: integer; (* beginning nucleotide *)
         cooend: integer; (* ending nucleotide *)
         piecon: configuration; (* configruation (circular/linear) *)
         piedir: direction; (* direction (+/-) relative to coordinates *)
         piebeg: integer; (* beginning nucleotide *)
         pieend: integer; (* ending nucleotide *)
      end;


      piece = record
         key: piekey;
         dna: dnaptr
      end;

      reference = record
         pienam : name;  (* name of piece referred to *)
         mapbeg : real;  (* genetic map beginning *)
         refdir : direction;  (* direction relative to coordinates *)
         refbeg : integer;  (* beginning nucleotide *)
         refend : integer;  (* ending nucleotide *)
      end;

      genkey = record  (* gene key *)
         hea : header;
         ref : reference;
      end;

      trakey = record  (* transcript key *)
         hea : header;
         ref : reference;
      end;

      markerptr = ^marker;
      markey = record (* marker key *)
         hea : header;
         ref : reference;
         sta : state;
         phenotype : lineptr;
         next : markerptr;
      end;

      marker = record
         key : markey;
         dna : dnaptr;
      end;

(* end module book.type version = 7.67; {of delmod.p 2004 Sep 8} *)

(* these items, which define the catalogue structure, come from loocat *)
   item = record (* information about an object in a library *)
      nodeletter: char; (* identification of library object type *)
      nam: name;  (* the name of the object *)
      line: integer (* the place of the object in the library *)
   end;
   recordfile = file of item;

var

   book, (* file used by this program *)
   fastcatp: (* file used by this program *)
   text; (* file used by this program *)

   cat: recordfile; (* a catalogue in the delila system *)

(* Note: fastcat does not use module book.var *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.67; {of delmod.p 2004 Sep 8} *)

(* begin module fastcat.themain *)
procedure themain(var book, fastcatp: text; var cat: recordfile);
(* the main procedure of the program *)
var
   catitem: item; (* one item in cat *)
   ch: char; (* first character on beginning of a book line *)
   i: integer; (* index to datetime being read in *)
   parameterversion: real; (* parameter version number *)
   theline: integer; (* the current line in the book *)

   orgopen: boolean; (* whether we are inside an organism *)
   chropen: boolean; (* whether we are inside a chromosome *)
   pieopen: boolean; (* whether we are inside a piece *)
   genopen: boolean; (* whether we are inside a gene *)
   maropen: boolean; (* whether we are inside a marker *)
   traopen: boolean; (* whether we are inside a transcript *)

procedure grabname(ch: char;
                   var book: text; var cat: recordfile;
                   var objopen: boolean);
(* for the object of type ch,
grab the name from the book file and put it into the cat file *)
var
   i: integer; (* index to the name *)
   c: char; (* a character of a name *)
begin
   objopen := not objopen;
   if objopen then begin

      (* store cat item *)
      catitem.nodeletter := ch;
      catitem.line := theline - 1;

      (* read key name.  see brname in delmod for original routine *)

      read(book,ch,ch); (* Skip '* ' *)
      with catitem.nam do begin
         length:=0;
         repeat
            length:=succ(length);
            read(book,c);
            letters[length] := c;
         until (eoln(book)) or
               (length>=namelength) or
               (letters[length]=' ');
         if letters[length]=' ' then length:=length-1;
         readln(book);
         theline := succ(theline);
      end;

      cat^ := catitem;
      put(cat);

      (* clear name for next time *)
      for i:=1 to catitem.nam.length do catitem.nam.letters[i]:=' ';
  
   end;
end;

begin
   writeln(output,'fastcat ',version:4:2);
   reset(fastcatp);
   readln(fastcatp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   reset(book);

   if eof(book) then begin
      writeln(output,'ERROR: the book is empty');
      halt;
   end;

   if book^ = 'h' then begin
      writeln(output,'ERROR: the book is bad');
      halt;
   end;

   rewrite(cat);

   (* set getocp variables *)
   orgopen := false;
   chropen := false;
   pieopen := false;
   genopen := false;
   maropen := false;
   traopen := false;
   theline := 1;

   (* set up the first item of the catalog, the datetime *)
   for i:=1 to namelength do catitem.nam.letters[i]:=' ';
   with catitem do begin
      nodeletter := '*';
      nam.length := datetimearraylength;
      read(book,ch,ch); (* skip the '* ' *)
      i := 1;
      for i := 1 to nam.length do begin (* grab datetime *)
          read(book, nam.letters[i]);
      end;
      line := theline;
   end;
   cat^ := catitem;
   put(cat);
   (* clear the date for reuse *)
   for i:=1 to catitem.nam.length do catitem.nam.letters[i]:=' ';
   readln(book);
   theline := succ(theline);

   (* the concept of this loop is based on getocp in delmod.p *)
   repeat
      readln(book,ch);
      theline := succ(theline);
      if ch in ['o','c','p','g','t','m'] then begin
         case ch of
            'o': grabname(ch, book, cat, orgopen);
            'c': grabname(ch, book, cat, chropen);
            'p': grabname(ch, book, cat, pieopen);
            'g': grabname(ch, book, cat, genopen);
            't': grabname(ch, book, cat, traopen);
            'm': grabname(ch, book, cat, maropen);
         end;
      end;
   until eof(book);

end;
(* end module fastcat.themain *)

begin
   themain(book, fastcatp, cat);
1: end.
