program fealoc(featab, data, fealocp, fealocreport, output);
(* fealoc: Example Delila System program (an empty fealoc)

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.13; (* of fealoc.p 2018 Mar 07
2018 Mar 07, 1.13: rename report to be fealocreport
2018 Mar 06, 1.12: clean up
2018 Mar 06, 1.11: tighten display
2018 Mar 06, 1.10: display functioning
2018 Mar 01, 1.09: check intragenic sites (continued)
   There were cases of coordinates being zero in featab.
   These came from join() not properly handled in dbinst!
   I fixed dbinst, dbinst 3.58
2018 Feb 26, 1.08: check intragenic sites (continued)
2018 Feb 25, 1.07: check intragenic sites
2018 Feb 25, 1.06: check between (inside features)
2018 Feb 23, 1.05: read data file
2018 Feb 23, 1.04: full featab data structure read in.
2018 Feb 23, 1.03: process featab into the data structure
2018 Feb 23, 1.02: read all featab parts by type
2018 Feb 23, 1.01: reading featab successfully!
2018 Feb 23, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.fealoc *)
(*
name
   fealoc: feature location in a genome

synopsis
   fealoc(featab: in, data: in, fealocp: in, fealocreport: out, output: out)

files

   featab:  feature table from dbinst giving locations of features.
      After an id line beginning with '*', the second line defines
      columns:
           accession       name    begin   end     orientation
      Remaining lines contain the data, for example:
           NC_000913.3     thrL    190     255     +
           NC_000913.3     thrA    337     2799    +

   data:  data table from scan giving locations of sites
          Lines that begin with '*' are ignored.
          Coordinate data is expected to be in the 6th space
          separated column

   fealocreport:  report of results in a table

   fealocp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      line 2: Lowest site in bits to identify

      line 3: Maximum number of sites to report.
              For testing set this to a small number.
              For production, set it to -1 (meaning all sites).

   output: messages to the user

description

   Read in a feature table (featab) created by dbinst and then read
   through data from scan and locate the sites identified by scan in
   the features.

examples

documentation

see also

   {creates featab:} dbinst.p
   {creates data:} scan.p

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.fealoc *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

type

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module featabstructure *)
   (* define a data structure to read in the featab table *)
   featabstructureptr = ^featabstructure;
   featabstructure = record
      accession: string; (* GenBank accession number *)
      name: string; (* feature name *)
      first: integer; (* first base of feature *)
      last : integer; (* last base of feature *)
      orientation: char; (* orientation of feature, + or - *)
      next: featabstructureptr;
   end;
(* end module featabstructure *)

var
   featab, data, (* file used by this program *)
   fealocp, (* file used by this program *)
   fealocreport: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer.
As of 2010 Dec 04 there is a previous pointer that
allows proper removal of the linked list using disposestring.
This clearstring routine is for the initial clearing
of the variables. *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      previous := nil;
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure disposestring(var ribbon: stringptr);
(* remove the linked list from the string in revere order *)
var
   t: stringptr; (* a link in the string *)
begin
   t := ribbon;
   (* find the end *)
   while t^.next <> nil do t := t^.next;
   (* move backwards and dispose links *)
   while t^.previous <> nil do begin
      t := t^.previous;
      dispose(t^.next);
   end;
   dispose(t);
end; (* disposestring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module onetoken *)
procedure onetoken(var afile: text; var buffer: string;
                  var gotten: boolean);
(* Get a string from a file not using string calls.  This lets one
obtain lines from a file without interactive prompts.  if end of file
is found, gotten is false. *)
const
   tabcharacter = 9; (* the ordinal of the tab character *)
var   index: integer; (* of buffer *)
      done: boolean;
begin (* onetoken *)
   skipblanks(afile);
   emptystring(buffer);
   done := false;
   if eof(afile)
   then gotten := false
   else begin
      index := 0;
      while (not eoln(afile)) and (index < maxstring)
         and not done
      do begin
         index := succ(index);
         read(afile, buffer.letters[index]);
         if    (buffer.letters[index] = ' ')
            or (ord(buffer.letters[index]) = tabcharacter)
         then begin
            done := true;
            index := pred(index);
         end
      end;
      buffer.length := index;
      buffer.current := 1;
      gotten := true
   end
end;
(* end module onetoken version = 5.67; (@ of prgmod.p 2017 Sep 29 *)

(* begin module writefeatab *)
procedure writefeatab(var afile: text; f: featabstructureptr);
(* write the featuretab structure out to a file *)
begin
   with f^ do begin
      write(afile, 'accession: ');
      writestring(afile, accession); writeln(afile);
      write(afile, 'name: ');
      writestring(afile, name); writeln(afile);
      writeln(afile, 'first: ',first:1);
      writeln(afile, 'last: ',last:1);
      writeln(afile, 'orientation: ',orientation);
   end;
end;
(* end module writefeatab *)

(* begin module showfeatab *)
procedure showfeatab(var afile: text; f: featabstructureptr);
(* show the featuretab structure on a single line to a file *)
const
   debugging = false;
begin
   with f^ do begin
      write(afile,' of');
      if debugging then write(afile, 'accession: ')
                   else write(afile, ' ');
      writestring(afile, accession);
      if debugging then write(afile, ' name: ')
                   else write(afile, ' ');
      writestring(afile, name);
      write(afile,' =');
      if debugging then write(afile, 'first: ',first:1);
      write(afile, ' ',first:1);
      if debugging then write(afile, 'last: ',last:1);
      write(afile, ' ',last:1);
      if debugging then write(afile, 'orientation: ')
                   else write(afile, ' ');
      write(afile, orientation);
{
      writeln(afile);
}
   end;
end;
(* end module showfeatab *)

(* begin module readfeatab *)
procedure readfeatab(var featab: text; var ftsp: featabstructureptr);
(* read the feature table from featab and store into the linked
list ftsp.  Make sure that first < last for later speed of figuring
out whether a site is in a given interval *)
const debugging = false; (* debug the code *)
var
   skip: integer; (* counter for skipping featab header *)
   accession: string;  (* accession buffer for onetoken *)
   name: string;        (* name buffer for onetoken *)
   first: integer; (* first base of the feature *)
   last: integer; (* last base of the feature *)
   orientation: char; (* orientation *)
   gotten: boolean; (* did we get the buffer? *)
   n: integer; (* count of data read in *)
   f: featabstructureptr; (* pointer to the data structure *)
   hold: integer; (* for switching first and last *)

procedure justread;
(* simple read of the file *)
begin
   onetoken(featab, accession, gotten);
   if not gotten then begin
      writeln(output,'readfeatab: failed to get buffer for accession');
      halt;
   end;
   onetoken(featab, name, gotten);
   if not gotten then begin
      writeln(output,'readfeatab: failed to get buffer for name');
      halt;
   end;
   read(featab, first);
   read(featab, last);
   skipblanks(featab); read(featab, orientation);
   readln(featab);
   write(output, 'accession: ');
   writestring(output, accession); writeln(output);
   write(output, 'name: ');
   writestring(output, name); writeln(output);
   writeln(output, 'first: ',first:1);
   writeln(output, 'last: ',last:1);
   writeln(output, 'orientation: ',orientation);
end;

begin
   writeln(output,'readfeatab: reading featab file');

   reset(featab);
   (* skip the header lines *)
   for skip := 1 to 2 do begin 
      if eof(featab) then begin
         writeln(output,'featab is empty');
         halt;
      end;
      readln(featab);
   end;

(* for testing:
   justread;
*)

   new(ftsp);
   f := ftsp;
   n := 0;

   while not eof(featab) do begin
      n := n + 1;
      with f^ do begin
         onetoken(featab, accession, gotten);
         if not gotten then begin
            writeln(output,'readfeatab: failed to get buffer for accession');
            halt;
         end;

         onetoken(featab, name, gotten);
         if not gotten then begin
            writeln(output,'readfeatab: failed to get buffer for name');
            halt;
         end;

         read(featab, first);
         read(featab, last);
         skipblanks(featab); read(featab, orientation);

         readln(featab);
         if debugging then begin
            writeln(output);
            writefeatab(output, f);
         end;

         if first > last then begin
{
writeln(output,'******BUBA before');
writefeatab(output, f);
}
            hold := first;
            first := last;
            last := hold;
{
writeln(output,'******BUBA after');
writefeatab(output, f);
writeln(output,'****************');
halt;
}
         end;

      end;
      if not eof(featab) then begin
         new(f^.next);
         f := f^.next;
      end
      else begin
         f^.next := nil; (* finish the list *)
      end;

      if debugging then begin
         if (n >= 3) then halt;
      end;
   end;
   writeln(output,n:1,' features read from featab');
{zzqq}
end;
(* end module readfeatab *)

function between(a,b,c: integer): boolean;
(* is b between a and c? *)
(* this is an inclusive between *)
begin
      between:=((a<=b) and (b<=c))
                       or
               ((c<=b) and (b<=a))
end;

function orderedbetween(a,b,c: integer): boolean;
(* is b between a and c?  a < c for speed *)
(* this is an inclusive between *)
begin
      orderedbetween:=((a<=b) and (b<=c))
(*
                       or
               ((c<=b) and (b<=a))
*)
end;


(* begin module fealoc.themain *)
procedure themain(var featab, data, fealocp, fealocreport: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   features: featabstructureptr; (* featab data *)
   afeature: featabstructureptr; (* a single feature *)
   previousfeature: featabstructureptr; (* the previoius feature *)
{
   buffer: string;  (* buffer for onetoken *)
}
   gotten: boolean; (* did we get the buffer? *)
   done: boolean; (* done searching for a feature for a site? *)

(* data file variables:
* 1 piece number
* 2 piece length
* 3 piece name
* 4 piece coordinate
* 5 matrix orientation (+1 = as in book, -1 = complement)
* 6 Ri evaluation (bits per site)
*)
   piecenumber: integer;
   piecelength: integer;
   piecename: string;
   piececoo: integer;
{
   orientation: char;
}
   orientation: integer;
   ri: real;

   ricutoff: real; (* parameter for minimum acceptable Ri *)
   line: integer; (* count of data lines *)
   sites: integer; (* number of sites at or above the ricutoff *)
   maxsites: integer; (* maximum number of sites to report (for testing) *)
   b: char; (* backspace *)

procedure writedata;
(* write the data from the data file out on several lines *)
begin
writeln(output,'1 piecenumber: ',piecenumber:1);
writeln(output,'2 piecelength: ',piecelength:1);
write(output,  '3 piecename: '); writestring(output,piecename); writeln(output);
writeln(output,'4 piececoo: ',piececoo:1);
writeln(output,'5 orientation: ',orientation:1);
writeln(output,'6 ri: ',ri:4:2);
end;

procedure showdata(var afile: text);
(* write the data from the data file out on one line *)
begin
   write(afile,piececoo:8);
   write(afile, ' ');
   if orientation = 1
   then write(afile, '+')
   else write(afile, '-');
   write(afile,' ',ri:4:2,' bits');
   write(afile,' |');
end;

procedure showinside(var afile: text; feature: featabstructureptr);
(* Dislay the feature with inside information *)
begin
   showdata(afile);
   write(afile,' INSIDE  -      ');
   showfeatab(afile, afeature);
   writeln(afile);
end;

procedure showoutside(var afile: text; feature: featabstructureptr;
beforeafter: char);
(* Dislay the feature with outside information *)
begin
   showdata(afile);
   write(afile,' OUTSIDE ');
   if feature^.orientation = beforeafter
   then write(afile,'CONTROL')
   else write(afile,'-      ');
   showfeatab(afile, feature);
   writeln(afile);
end;

begin
   writeln(output,'fealoc ',version:4:2);
   reset(fealocp);
   readln(fealocp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
   readln(fealocp,ricutoff);
   writeln(output,'Ri cutoff: ',ricutoff:4:2,' bits');
   readln(fealocp,maxsites);

   (* read data from featab into f *)
   readfeatab(featab, features);
   writeln(output);
   writeln(output, 'The first feature read is:');
   writefeatab(output, features);
   writeln(output);

   (* read data from data file *)
   reset(data);
   rewrite(fealocreport);
   line := 0;
   sites := 0;
   b := chr(8); (* backspace character *)

   (* header line: *)
   writeln(fealocreport,
'*coo',
' orient',
' bits',
' bits',
' |',
' in/out',
' control?',
' of',
' Accession',
' gene',
' =',
' first',
' last',
' orient'
);

   writeln(output,'data file line:');
   while not eof(data) do begin
      line := line + 1;
      if (data^ = '*') then begin
         readln(data);
      end
      else begin
         write(output,line:1,b,b,b,b,b,b,b,b,b);
{
}
         read(data,piecenumber);
         read(data,piecelength);
         onetoken(data,piecename,gotten);
         if not gotten then begin
            writeln(output,'unable to get a piece name');
            halt;
         end;
         read(data,piececoo);
         skipblanks(data);
         read(data,orientation);
         read(data,ri);
         readln(data); (* SKIP THE LAST COLUMNS FOR NOW *)

         if (ri >= ricutoff) then begin
             sites := sites + 1;
             (* At this point we ask where the coordinate is in the genome.
                Maybe an array would be faster?
                Maybe a tree would be faster? *)
             previousfeature := features;
             afeature := features;
             done := false;
             while not done do begin
                if orderedbetween(afeature^.first,
                                            piececoo,
                                  afeature^.last) then begin
                   showinside(fealocreport, afeature);

if afeature^.first = 0 then begin
   writeln(output,'Something is terribly wrong if a coordinate is zero!');
   writeln(output,'These come from join() not properly handled in dbinst!');
   halt;
end;
                end;
                if (piececoo < afeature^.first) and
                   (piececoo >= previousfeature^.last)
                then begin
                   showoutside(fealocreport, previousfeature,'-');
                   showoutside(fealocreport, afeature,'+');
                end;
                previousfeature := afeature;
                afeature := afeature^.next;
                if afeature = nil then done := true;
             end
         end;

         if maxsites >= 0 then begin
            if (sites >= maxsites) then begin
               writeln(output, 'sites = ',sites:1,
                               ' >= maxsites = ',maxsites:1);
               halt;
            end;
         end;

      end
   end;
   writeln(fealocreport,'* sites: ',sites:1,' ');
   writeln(output,'sites: ',sites:1,' ');
   writeln(output,'fealoc is done');

(* read through the data file using a parameter to throw
away sites below a cutoff defined in fealocp
This might be fast enough!
*)
   

end;
(* end module fealoc.themain *)

begin
   themain(featab, data, fealocp, fealocreport);
1: end.
