program gauss(gaussp,data,genhisparameters,output);
(* generate a bunch of gaussianly distributed numbers *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.03; (* of gauss.p 2011 May 06
2011 May 06, 1.03: upgrade documentation; genhisp->genhisparameters
1994 sep  5, 1.02: functional
1993 Jul 07, 1.02: origin 1993 July 7 from gentst *)
(* end module version *)

(* begin module describe.gauss *)
(*
name
   gauss: generate a bunch of gaussianly distributed numbers

synopsis
   gauss(gaussp:in, data: out, genhisparameters: out, output: out) 

files 

   gaussp: parameter file controlling the program.
      Two numbers, one per line:
         seed: random seed to start the process
         total: the number of numbers to generate

   data: the input file for genhis.  this is a set of numbers which
      should have gaussian distribution if the random number generator
      is a reasonable one.  It will be N(0,1), a normal distribution
      with mean 0 and standard deviation 1.

   genhisparameters: control file for genhis.  Copy or move to the
      name genhisp to control genhis.  This gives the user flexibility.

   output: messages to the user

description 

   Test of a random number generator by creating a Gaussian
   distribution of numbers for plotting by genhis.  Further details of
   the method are in email communications appended to the end of this
   program in a comment.

example
  seed := 0.5;
  total := 10000;

see also
   tstrnd.p, genhis.p, gentst.p

author
   Thomas D. Schneider

bugs
   none known
  
technical notes 
%
   the constant n in procedure randomtest determines how many times  
   the random number generator will be in a series of tests.  if n 
   is small, the the test will be poor, if it is large then the test may 
   take a long time. 

*)
(* end module describe.gauss *)

var
(* begin module gauss.var *)
   gaussp,data, genhisparameters: text; (* output files *)
(* end module gauss.var *)

(* begin module random *) 
procedure random(var seed: real); 
(* random generator 2.  version = 1.00; 1986 December 31
   Test this routine with the program tstrnd.
   written by David Masternarde *)

(* This random number generator is based on a shift register with a single bit
of feedback, as described in Electronics for Neurobiologists, by Brown, Maxfield
and Moraff, MIT press 1973, referencing Random Process Simulation and
Measurement by Korn, McGraw-Hill 1966.  The random seed rand, a number between 0
and 1 exclusive, is converted to an integer between 1 and 2**23-1, inclusive.
This 23-bit number is shifted right one bit and the output of the last (23rd)
bit and the 9th bit are added modulo 2 (exclusive orred) and fed back into the
new first bit.  This is done between 4 and 11 times, depending on the last 3
bits of the original number.  The result is converted back to a real number
between 0 and 1 from which the 23 bit integer can be recovered on the next call.
The 23-bit shift register goes through all 2**23-1 values before repeating;
the repetition frequency of this algorithm could be less or greater depending
on the seed, because of the random number of multiple shifts per call. *)
(* powers of 2 *)
const pow14=16384; pow15=32768; pow22= 4194304; pow23=8388608;
var iseed, (* integer shift register *)
    i, nrep: integer;  (* index, number of times to do shift *)
begin (* random *)  
      iseed := round(seed*pow23);  (* convert to 23 bit number *)
      if (iseed<1) or (iseed>=pow23) then iseed := 1;
      nrep := 4 + iseed mod 8;  (* do it 4 to 11 times based on last 3 bits *)
      for i:= 1 to nrep do
         (* if last bit and 9th bit are equal, feed back a 0, otherwise a 1 *)
	 if( odd(iseed) = ((iseed mod pow15) >= pow14))
	    then iseed := iseed div 2
	    else iseed := (iseed div 2) + pow22;
      seed := iseed/pow23;
end; (* random *) 
(* end module random *) 

(* begin module pic.polrec *)
procedure polrec(r,theta: real; var x,y: real);
(* convert polar to rectangular coordinates,
theta is in radians *)
begin
   x := r*cos(theta);
   y := r*sin(theta)
end;
(* end module pic.polrec version = 1.40; (@ of domod.p 1989 Aug 9 *)

(* begin module gaussian *)
procedure gaussian(var seed: real;
                   var haveone: boolean;
                   var x,y: real);
(* Produce a Gaussianly distributed number.  The seed is for the random number
generator.  Haveone is a Boolean which should be set to false the first time
the routine is called.  The routine produces two Gaussian numbers, and returns
one in x, and the other in y.  The next time the routine is called, the y value
is returned in x (the y value equals the x value).  This allows one to keep
calling the routine and using the x value.
*)
const
  twopi = 6.283185307179586476925; (* twice pi *)
var
  radius: real; (* radius *)
  angle: real; (* angle *)
begin
  if haveone
  then x := y
  else begin
     random(seed);
     angle := seed * twopi;
     random(seed);
     radius := sqrt(-2.0*ln(seed));
     polrec(radius,angle,x,y);
  end;
  haveone := not haveone
end;
(* end module gaussian *)

(* begin module gauss.themain *)
procedure themain(var data, genhisparameters: text);
(* the main procedure of gauss *)
const
  xnumber = 20; (* number of points on x axis to plot *)
var
  haveone: boolean; (* for the gauss to keep track of which
                       number it has given *)
  indexfortotal: integer; (* index for total *)
  total: integer; (* total numbers to make *)
  seed: real; (* the current random number seed *)
  x, y: real; (* two Gaussian numbers *)
begin
  writeln(output,'gauss ',version:4:2);

  reset(gaussp);
  readln(gaussp,seed);
  readln(gaussp,total);

  rewrite(genhisparameters);
  (* set x scale to fit a page: *)
  writeln(genhisparameters,'x n ', xnumber:1);
  writeln(genhisparameters,'p g'); (* plot Gaussian *)

  rewrite(data);
  writeln(data,'* gauss ',version:4:2);
  writeln(data,'* ',seed:10:9, ' random seed');
  writeln(data,'* ',total:1,' numbers, each the sum of');

  random(seed);
  haveone := false;
  for indexfortotal := 1 to total do begin
    gaussian(seed,haveone,x,y);
    writeln(data,x);
  end
end;
(* end module gauss.themain *)

begin
   themain(data,genhisparameters);
1: end.

(*  The following describes the process used in the program
_____________________________________________________________________________

From daemon Wed Jun 30 17:21:16 1993
Return-Path: <daemon>
Message-Id: <9306302121.AA23640@fcs280s.ncifcrf.gov>
From: Marcelo Magnasco <marcelo@mjf.rockefeller.edu>
Subject: reprints.
To: toms@alum.mit.edu (Tom Schneider) (Tom Schneider)
Date: Wed, 30 Jun 93 17:33:49 EDT
In-Reply-To: <9306222308.AA12118@fcs280s.ncifcrf.gov>;
   from "Tom Schneider" at Jun 22, 93 7:08 pm
Mailer: Elm [revision: 64.9]


Hi, Tom, I was glancing through your reprints (haven't had time to seriously
read them yet) and noticed that you were computing gaussian random numbers
through addition of 100 uniform random numbers. There are much better 
techniques than this! Are you aware of them? (you generate two at a time)

Anyway, best regards,
Marcelo
_____________________________________________________________________________

From toms Wed Jun 30 17:24:22 1993
To: marcelo@mjf.rockefeller.edu
Subject: Re:  reprints.
Content-Length: 311

Marcelo:  No, I didn't know there was a better way, though it makes
sense.  For the purposes of that paper, adding 100 did the job
but it did cost me since I couldn't easily explore the higher
dimensional cases.  I would be interested in the method though.
Could you post it to bionet.info-theory?
Regards,
Tom
_____________________________________________________________________________

From daemon Wed Jun 30 17:47:32 1993
Return-Path: <daemon>
Message-Id: <9306302147.AA23725@fcs280s.ncifcrf.gov>
From: Marcelo Magnasco <marcelo@mjf.rockefeller.edu>
Subject: Re:  reprints.
To: toms@alum.mit.edu
Date: Wed, 30 Jun 93 18:00:07 EDT
In-Reply-To: <9306302124.AA27263@fcs260c2.ncifcrf.gov>;
   from "toms@alum.mit.edu" at Jun 30, 93 5:24 pm
Mailer: Elm [revision: 64.9]

Hi Tom,

The method is extremely well known by people who integrate Langevin equations
or do numerical stochastic stuff, so I don't really think I should post it. 

Suppose you want independent univariate gaussian random numbers. 
Lets assume you have two such numbers. The joint probability distribution
for the two is a two-dimensional gaussian (the D=2 case in your paper). 

Change variables to r, theta. The prob distribution for theta is uniform 
in [0,2 pi). The prob distribution for r is proportional to r exp(-r^2/2)
The cumulative p.d. for this (the integral between 0 and x) is 1-exp(-r^2/2),
which has an explicit inverse: r= sqrt(-2*log( . )).

So you use one uniform random number to generate a theta, a second 
uniform random number, map it through the inverse of the cum.p.d. to get
r with the appropriate distribution, and then compute x = r * cos(theta)
and y = blah blah. You just got yourself two independent univariate 
Gaussian random numbers. 

(You cannot do this with just one UGRN because the cumulative p.d. has no
explicit inverse, i.e. errorfunction^{-1} has to be computed rather 
awkwardly and its a loss of time. The existence of an explicit inverse
to the cumulative p.d. is the key here)

BTW, such a routine is in Numerical Recipes. Unfortunately, it uses the 
NumRec ran1 random number generator for the uniform RNs, which is an 
incredibly inferior routine. NEVER use ran1 for any serious work at all!
It's better to hack your own, or to use the unix buitin if you are on a 
unix machine. 

With this scheme, I generate about 200.000 uniform gaussian random numbers
per second on my iris. Which means I can duplicate your plots (the ones
showing data on top of the grayshade prob dist funct) in under a couple
of seconds... that's why I though I should tell you!

Best Regards,
Marcelo Magnasco
_____________________________________________________________________________

From toms Wed Jun 30 17:54:27 1993
To: marcelo@mjf.rockefeller.edu
Subject: Re:  reprints.
Content-Length: 329

Marcelo:  Geez that is neat - the only part I'll have to think about
is the crux - how you use the inverse.  Although this may be well known
by certain folks, I still think that a posting on it would be neat.
It might get the discussion rolling in that direction.  Maybe I'll put
together some code for that method.  Thanks, Tom
_____________________________________________________________________________

From daemon Wed Jun 30 18:14:03 1993
Return-Path: <daemon>
Received: by fcs280s.ncifcrf.gov (4.1/NCIFCRF-3.0/AWF-2.0)
	id AA23787; Wed, 30 Jun 93 18:14:01 EDT
Message-Id: <9306302214.AA23787@fcs280s.ncifcrf.gov>
Received: from mjf.rockefeller.edu by fcs280s.ncifcrf.gov (4.1/NCIFCRF-3.0/AWF-2.0)
	id AA23784; Wed, 30 Jun 93 18:14:00 EDT
Received: by mjf.rockefeller.edu
	(15.11/15.6) id AA09347; Wed, 30 Jun 93 18:26:44 edt
From: Marcelo Magnasco <marcelo@mjf.rockefeller.edu>
Subject: Re:  reprints.
To: toms@alum.mit.edu
Date: Wed, 30 Jun 93 18:26:42 EDT
In-Reply-To: <9306302154.AA27412@fcs260c2.ncifcrf.gov>;
    from "toms@alum.mit.edu" at Jun 30, 93 5:54 pm
Mailer: Elm [revision: 64.9]
Status: R


If you want to assemble this into a coherent posting that's ok with me;
but I still insist that this method has nothing of mine inside, and that
it is in all the books on stochastic computations...

The inverse. 

Take any probability distribution p(x). Let me define F(x) as 
the integral p(x) dx between minus infinity and x. Then F(x) maps 
the line to the interval (0,1). (-infinity gets mapped to 0, because
there's no range, and +infinity gets mapped to 1, because p(x) is assumed
to be normalized; F(x) is monotonic nondecreasing because its derivative,
p(x), is a nonnegative function ).

Take now a random number x generated according to the distribution p. 
Then F(x) will be a random number in (0,1); furthermore, it will be 
uniformly distributed. (Basically, F tells you the "rank" of a number,
in the sense it tells you how much of the distribution lies below that
number; like the GRE "percent below" numbers, yeah?)

So, if I take a random number y uniformly distributed in (0,1), then 
F^{-1}(y) is a random number whose probability distribution is p. 

(this again is a well known method)

This method is useful only when you can get F^{-1} in closed form. In 
one dimension, F is the dreaded error function and F^{-1} is hideous to 
compute. In two dimensions, F has an extremely simple form. 

Trivial example: take a uniform random number y, and compute 
x = tan( pi(y-1/2) ). The distribution function of x is the derivative of 
the inverse of the tan, i.e., 1/(1+x^2). So you just generated a 
Lorentzian random number (bad guys, infinite variance and logarithmically
nonconvergent mean, but nice for some real rough stuff). 

Code:  (in C, using Unix's random() )

/* Generate a Gaussian random number with sigma=1       */
/* Base RNG: random() from unix                         */

#define RNG (random()/2147483648.0)
#define TWOPI 6.283185307179586476925

double grand()
{
static int needmore=0;
static double x,y,r,t;

if(needmore==0) {
  t=RNG*TWOPI; r=sqrt(-2.0*log(RNG));
  x=r*cos(t); y=r*sin(t); needmore=1;
  return(x); }
else { needmore=0; return(y); }
}

This code (the obvious code!) generates two random numbers the first time
you call it and returns the first one; the second time it remembers it 
still has a spare and gives you the second one, and so on and so forth. 


Happy simulations!

Marcelo Magnasco

*)
