program gifnormalize(input, gifnormalizep, output);
(* gifnormalize: normalize a gif file to a given size

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.07; (* of gifnormalize.p 2016 Jan 29
2016 Jan 29, 1.07: use zerovalue = chr(0); instead of non-ascii character
                   that caused p2c/tc to have a segmentation fault.
1996 Nov 14, 1.06: functional
1996 Nov 14, 1.00: origin
origin 1996 Nov 14 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.gifnormalize *)
(*
name
   gifnormalize: normalize a gif file to a given size

synopsis
   gifnormalize(input: in, gifnormalizep: in, output: out)

files
   input:  a gif file

   output:  the gif file resized according to the parameters

   gifnormalizep:  parameters to control the program
      version: lowest version number acceptable 
      startheight, startwidth: the height and width to start writing out
      endheight, endwidth: the height andwidth to end writing out

   output: messages to the user

description
   The program allows one to regularize a set of gifs for making a movie
   with gifmerge.  It reads in a pnm format file created by:

   giftopnm < $1 > $1.pnm

   and writes it out again.  The rectangular region that is written
   out is given by the two points
      (startheight, startwidth)
      (endheight, endwidth)

examples

documentation

   Manual page from ppm:

********************************************************************************
ppm(5)             Headers, Tables, and Macros             ppm(5)

NAME
     ppm - portable pixmap file format

DESCRIPTION
     The portable pixmap format is a  lowest  common  denominator
     color image file format.  The definition is as follows:

     - A "magic number" for identifying the  file  type.   A  ppm
       file's magic number is the two characters "P3".

     - Whitespace (blanks, TABs, CRs, LFs).

     - A width, formatted as ASCII characters in decimal.

     - Whitespace.

     - A height, again in ASCII decimal.

     - Whitespace.

     - The maximum color-component value, again in ASCII decimal.

     - Whitespace.

     - Width * height pixels, each  three  ASCII  decimal  values
       between 0 and the specified maximum value, starting at the
       top-left  corner  of  the  pixmap,  proceeding  in  normal
       English  reading  order.   The three values for each pixel
       represent red, green, and blue, respectively; a value of 0
       means  that color is off, and the maximum value means that
       color is maxxed out.

     - Characters from a "#" to the next end-of-line are  ignored
       (comments).

     - No line should be longer than 70 characters.

     Here is an example of a small pixmap in this format:
     P3
     # feep.ppm
     4 4
     15
      0  0  0    0  0  0    0  0  0   15  0 15
      0  0  0    0 15  7    0  0  0    0  0  0
      0  0  0    0  0  0    0 15  7    0  0  0
     15  0 15    0  0  0    0  0  0    0  0  0

     Programs that read this format should be as lenient as  pos-
     sible, accepting anything that looks remotely like a pixmap.


     There is also a variant on the format, available by  setting
     the  RAWBITS  option  at  compile  time.   This  variant  is
     different in the following ways:

     - The "magic number" is "P6" instead of "P3".

     - The pixel values are stored as  plain  bytes,  instead  of
       ASCII decimal.

     - Whitespace is not allowed in the pixels area, and  only  a
       single  character  of  whitespace (typically a newline) is
       allowed after the maxval.

     - The files are smaller and many times faster  to  read  and
       write.

     Note that this raw format can only be used for maxvals  less
     than or equal to 255.  If you use the ppm library and try to
     write a file with a larger  maxval,  it  will  automatically
     fall back on the slower but more general plain format.

THE LATTER IS THE FORM CREATED BY gifpnm.


********************************************************************************

see also
   man ppm

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.gifnormalize *)

var
   gifnormalizep (* file used by this program *)
   : text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module gifnormalize.themain *)
(*
procedure themain(var finput, gifnormalizep, foutput: text);
*)
procedure themain(var gifnormalizep: text);
(* the main procedure of the program *)
const
   maxwidth  = 1000; (* maximum width of image *)
   maxheight = 1000; (* maximum height of image *)
   zerovalue = chr(0); (* value to fill array with *)
var
   c: integer; (* index to colors (red, green, blue) *)
   endwidth: integer; (* user desired end of width *)
   endheight: integer; (* user desired end of height *)
   h: integer; (* index to height *)
   height: integer; (* actual height of image *)
   m1, m2: char; (* characters of the magic number *)
   maximumcolorcomponentvalue: integer; (* maximum color component value *)
   parameterversion: real; (* parameter version number *)
   store: array[1..maxwidth,1..maxheight,1..3] of char; (* store the image *)
   width: integer; (* actual width of image *)
   w: integer; (* index to width *)
   startwidth: integer; (* user desired start of width *)
   startheight: integer; (* user desired start of height *)
begin

(*
   writeln(output,'# gifnormalize ',version:4:2);
*)
   reset(gifnormalizep);
   readln(gifnormalizep, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(gifnormalizep, startwidth, startheight);
   readln(gifnormalizep, endwidth, endheight);
   if endwidth > maxwidth then begin
      writeln(output, '# endwidth exceeds maxwidth = ', maxwidth:1);
      halt
   end;
   if endheight > maxheight then begin
      writeln(output, '# endheight exceeds maxheight = ', maxheight:1);
      halt
   end;

   if endwidth < startwidth then begin
      writeln(output, '# endwidth must exceeds startwidth');
      halt
   end;

   if startwidth < 1 then begin
      writeln(output, '# startwidth must be positive');
      halt
   end;

   if endheight < startheight then begin
      writeln(output, '# endheight must exceeds startheight');
      halt
   end;

   if startheight < 1 then begin
      writeln(output, '# startheight must be positive');
      halt
   end;

   if startwidth > maxwidth then begin
      writeln(output, '# startwidth exceeds maxwidth = ', maxwidth:1);
      halt
   end;
   if startheight > maxheight then begin
      writeln(output, '# startheight exceeds maxheight = ', maxheight:1);
      halt
   end;

   read(input,m1,m2);
   if (m1 <> 'P') or
      (m2 <> '6') then begin
      writeln(output, '# magic number is not P6');
      halt
   end;
   readln(input,width,height,maximumcolorcomponentvalue);

   (* zero the array for safety *)
   for h := 1 to maxheight do begin
      for w := 1 to maxwidth do begin
         for c := 1 to 3 do begin
            store[w,h,c] := zerovalue;
         end;
      end;
   end;

   (* read the image into the array *)
   for h := 1 to height do begin
      for w := 1 to width do begin
         for c := 1 to 3 do begin
            read(input, store[w,h,c]);
         end;
      end;
   end;


   (********************************************************************)
   (* now write it out. *)

   readln(input);
   if not eof(input)
   then writeln(output,'# did not find end of file!');

   writeln(output,m1,m2);
   writeln(output,(endwidth-startwidth+1):1,
              ' ',(endheight-startheight+1):1);
   writeln(output,maximumcolorcomponentvalue:1);

   for h := startheight to endheight do begin
      for w := startwidth to endwidth do begin
         for c := 1 to 3 do begin
            write(output, store[w,h,c]);
         end;
      end;
   end;

     
end;
(* end module gifnormalize.themain *)

begin
   themain(gifnormalizep);
1: end.














