program kenin(allgen, keninp, inst, output);
(* kenin: create Delila instructions from Ken's all.gen instructions

  Tom Schneider
  network address: toms@alum.mit.edu


  1991
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.24; (* of kenin.p 1991 Mar 23
origin 1991 Mar 23 *)
(* end module version *)

(* begin module describe.kenin *)
(*
name
   kenin: create Delila instructions from Ken's all.gen instructions

synopsis
   kenin(allgen: in, inst: out; output: out)

files
   allgen:  gene instructions of the form provided by Kenn Rudd:
      piecename l1 l2 geneA l3 l4 geneB
      These are on a single line.  The first location is the start
      of the gene.  If l1>l2, the gene is on the complementary strand
   keninp: parameters to control the program.
      First line: FROM and TO of the output instructions.
      Second line: if the first character is 'b' then both open reading frames
	 and identified genes are written to inst.  If it is 'n' then no open
	 reading frames (orf) genes are not made into instructions.  If it is
	 'o' then ONLY orfs are used.
   inst:  Delila instructions corresponding to allgen:
      piece piecename;
      get from l1 -FROM to l1 + TO;
   output: messages to the user

description
   This program converts Kenn Rudd's list of gene locations in his
   database into Delila instructions.

examples

documentation

see also

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.kenin *)

      maxstring = 500; (* the maximum string *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 4.10; (@ of prgmod.p 1991 March 20 *)

var
   allgen, (* Kenn instructions *)
   keninp, (* parameters to control the program *)
   inst: text; (* Delila instructions *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.10; (@ of prgmod.p 1991 March 20 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.10; (@ of prgmod.p 1991 March 20 *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 4.10; (@ of prgmod.p 1991 March 20 *)

(* begin module interact.figurestring *)
procedure figurestring(
      var line: string; (* a string of characters to figure out *)
      var first: integer; (* first found non-blank character in the line *)
      var last:  integer; (* last character before a blank after first *)
      var whzat: char; (* what the token is *)
      var c:     char;  (* the first character of the token *)
      var i:     integer; (* integer value of token if it is integer; or 0 *)
      var r:     real); (* the real value if it is real; or 0.0 *)
(* figurestring figures out the tokens in a string.
 it recognizes words, integers, reals and poorly formed numbers.
 you can easily use it to parse lines.

our goal is to figure out what thing is on a string.
start looking at the current place on the line.
first and last are the first 'token' in line after start.
the current place is updated to the letter after last.
the thing found is described by the value of whzat:
      'c': character (when the token does not begin with a digit, '+', or '-')
      'i': integer
      'r': real
      ' ': blank line
      'g': garbage, cannot figure it out
and the value of the thing found is the appropriate variable *)
var
      numbers: set of '0'..'9';
      sign: integer; (* sign of a number *)
      numberstart: integer; (* the point a number starts,
                               beyond its sign, if any *)
      point: integer; (* location of decimal point *)
      power: integer; (* of 10 representing a place value in the number *)

      l: integer; (* an index for dissecting numbers *)

function figureinteger(first,last:integer):integer;
(* figure the integer in the token *)
var
      i: integer; (* index *)
      sum, increment: integer;
begin (* figureinteger *)
      power:=1; (* start at ones place *)
      sum:=0; (* start sum at zero *)
      for i:=last downto first do begin
         case line.letters[i] of
            '0': increment:=0;
            '1': increment:=1;
            '2': increment:=2;
            '3': increment:=3;
            '4': increment:=4;
            '5': increment:=5;
            '6': increment:=6;
            '7': increment:=7;
            '8': increment:=8;
            '9': increment:=9
         end;
         sum:=sum+power*increment;
         power:=power*10
      end;
      figureinteger:=sum
end; (* figureinteger *)

begin (* figurestring *)
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
(* c:=' '; i:=0; r:=0.0; do not affect these variables unless necessary *)
      point:=0;
      whzat := '.'; (* assume that we have someting to work on *)
      (* now to see if that is true: *)

      with line do if (length = 0) or (current < 1) or (current > length)
      then whzat := ' '
      else begin
         (* figure out where the first token is in the line *)
         first:=line.current;
         while (line.letters[first]=' ') and (first < line.length)
            do first:=succ(first);
         if (first = line.length) and (line.letters[first] = ' ')
            then whzat := ' ';
      end;

      if whzat <> ' ' then begin
         last:=first;
         while (line.letters[last]<>' ') and (last < line.length)
            do last:=succ(last);
         if line.letters[last] = ' ' then last := pred(last);

         (* the token is between inclusive first and last *)

         c:=line.letters[first];
         if (c in numbers) or (c in ['+','-']) then begin
            if c in ['+','-'] then begin
               case c of
                  '+': sign:=+1;
                  '-': sign:=-1;
               end;
               numberstart:=succ(first)
            end
            else begin
               sign:=+1;
               numberstart:=first
            end;

            whzat:='i';
            for l:=numberstart to last do begin
               if not(line.letters[l] in numbers)
               then if line.letters[l]='.' (* we found a period *)
                  then if whzat='i' (* if so far it is numbers *)
                     then begin
                        whzat:='r'; (* it is actually real *)
                        point:=l
                     end
                     else whzat:='g' (* it is a second '.', ie garbage *)
                  else whzat:='g' (* it is garbage *)
            end;
            (* if it is only numbers, it is integer *)

            (* build number *)
            (* if it ends in a period, it is integer *)
            if (whzat = 'r') and (point = last) then whzat:='i';

            if whzat = 'i' then begin
               if point = last (* had an ending decimal point *)
               then i:=sign * figureinteger(numberstart,pred(last))
               else i:=sign * figureinteger(numberstart,last);
               r:=i
            end
            else if whzat = 'r' then begin
               i:=figureinteger(numberstart,point-1);
               r:=sign * (i+figureinteger(point+1,last)/power);
               i:=sign * i
            end
         end
         else begin
            whzat:='c';
         end;

         (* move the start to just beyond the last character of the token *)
         line.current:=succ(last)
      end
end; (* figurestring *)

(* end module interact.figurestring version = 4.10; (@ of prgmod.p 1991 March 20 *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 4.10; (@ of prgmod.p 1991 March 20 *)

(* begin module interact.token *)
procedure token(var buffer, atoken: string; var gotten: boolean);
(* get a token from the buffer *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      int: integer;
      rea: real;
      index: integer; (* to the buffer *)
begin
      figurestring(buffer,first,last,what,cha,int,rea);

      if what = ' '
      then gotten := false
      else begin
         clearstring(atoken);
         for index := first to last
            do atoken.letters[index-first+1] := buffer.letters[index];
         atoken.length := last - first + 1;
         atoken.current := 1;

         gotten:=true
      end
end;
(* end module interact.token version = 4.10; (@ of prgmod.p 1991 March 20 *)
 
(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a string from a file not using string calls.  this lets one
obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;

         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                            maxstring:1,')');
            halt
         end;

         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)
(* end module interact.getstring version = 4.10; (@ of prgmod.p 1991 March 20 *)

(* begin module kenin.themain *)
procedure themain(var allgen, keninp, inst: text);
(* the main procedure of the program *)
var
   atoken: string; (* a token found in the buffer string *)
   count: integer; (* count of instructions written *)
   buffer: string; (* a string that the user types *)
   gotten: boolean; (* used by interactive routines *)
   l1,l2: integer; (* two locations in a piece *)
   nocount: integer; (* count of instructions not written *)
   orfs: char; (* 'n': no orfs, 'o': only orfs, otherwise both *)
   orffound: boolean; (* if true, an orf gene was found *)
   thefrom,theto: integer; (* the FROM and TO parameters *)
   piecename: string; (* a token found in the buffer string *)
   writeinst: boolean; (* if true write the instruction, if false, put
                          comments around it *)

   (* variables for calling figurestring: *)
   first: integer;
   last: integer;
   what: char;
   cha: char;
   int: integer;
   rea: real;
procedure testforinteger;
(* test that atoken contains an integer *)
begin
   token(buffer,atoken,gotten);
   if gotten then begin
      figurestring(atoken,first,last,what,cha,int,rea);
      if what <> 'i' then begin
         write(output,'For piece "');
         writestring(inst,piecename);
         writeln(output,'"');
         write(output,'figurestring found the token "');
         writestring(inst,atoken);
         write(output,'" of type ',what);
         writeln(output,'.  An integer was expected.');
         halt
      end;
   end;
end;
function sign(n: integer): char;
(* produce the sign of n if it is positive... *)
begin
   if n >= 0 then sign := '+'
             else sign := ' '
end;
procedure conclusions(var thefile: text);
(* write out what happend *)
begin
   writeln(thefile);
   writeln(thefile,'(* ');
   writeln(thefile, count:1,' get instructions written');
   writeln(thefile,nocount:1,' get instructions commented out');
   if (orfs='n') or (orfs='o') then begin
      write(thefile, 'deleted instructions were');
      if orfs='o' then write(thefile,' not orfs')
                  else write(thefile,' orfs');
   end;
   writeln(thefile,' *)');
end;
begin
   writeln(output,'kenin ',version:4:2);
   count := 0;
   nocount := 0;

   reset(keninp);
   readln(keninp, thefrom, theto);
   if eof(keninp) then orfs := 'b'
   else readln(keninp,orfs);
   if not (orfs in ['b','n','o']) then begin
      writeln(output,'second parameter must be one of: bno');
      halt
   end;

   rewrite(inst);
   writeln(inst,'title "from kenin ',version:4:2,'";');

   reset(allgen);
   writeln(inst);
   writeln(inst,'(*');
   while allgen^ <> ' ' do copyaline(allgen,inst);
   writeln(inst,'*)');
   writeln(inst);

   writeln(inst,'organism ecoli; chromosome ecoli;');
   writeln(inst);
   writeln(inst,'(* number only the pieces, starting at 1 *)');
   writeln(inst,'      default numbering piece;');
   writeln(inst,'      default numbering 1;');
   writeln(inst,'      default out-of-range reduce-range;');

   while not eof(allgen) do begin
      getstring(allgen,buffer,gotten);
      if gotten then begin
         token(buffer,piecename,gotten);
         if gotten then begin
            writeln(inst);
            write(inst,'piece ');
            writestring(inst,piecename);
            writeln(inst,';');

            while gotten do begin
               testforinteger;
               l1 := int;

               testforinteger;
               l2 := int;

               token(buffer,atoken,gotten);

               if gotten then begin
                  with atoken do orffound := (letters[1] = 'o')
                                         and (letters[2] = 'r')
                                         and (letters[3] = 'f');
   
                  writeinst := (orffound and (orfs = 'o')) or
                         ((not orffound) and (orfs = 'n')) or
                         (orfs = 'b');

                  (* comment out the instruction if we don't want it *)
                  if not writeinst then write(inst,'(* ');

                  (* write out the whole instruction *)
                  if l1 < l2 then begin
                     write(inst,'get from ',l1:1,' ',sign(thefrom),thefrom:1);
                     write(inst,     ' to ',l1:1,' ',sign(theto),theto:1);
                  end
                  else begin
                     write(inst,'get from ',l1:1,' ',sign(-thefrom),-thefrom:1);
                     write(inst,     ' to ',l1:1,' ',sign(-theto),-theto:1);
                     write(inst, ' direction -');
                  end;
                  write(inst,'; ');

                  if writeinst then write(inst,'(* ');
                  writestring(inst,atoken);

                  if writeinst then count := succ(count)
                               else nocount := succ(nocount);

                  if writeinst then write(inst,' #',count:1);

                  writeln(inst,' *)');
               end
            end
         end
      end
   end;

   conclusions(output);
   conclusions(inst);

end;
(* end module kenin.themain *)

begin
   themain(allgen, keninp, inst);
1: end.















