program log(histog, logp, xyin, output);
(* log: compute log of a data column

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of log.p 2002 Oct 25
2002 Oct 25, 1.02: documentation
2002 Oct 25, 1.01: functional
2002 Oct 25, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.log *)
(*
name
   log: compute log of a data column

synopsis
   log(histog: in, logp: in, xyin: out, output: out)

files

   histog:  a histogram file from genhis.
      First column: position
      Second column: value

   xyin:  output file for xyplo.  The second column is used to compute
      the log, base 2.
      Negative or zero values are set to a very large negative number.

      First column: position
      Second column: log2(value)
      Third column: value

      This substitutes the log for the value, but keeps
      the original value around for inspection.

   logp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      not used, could be used to set base of log.

   output: messages to the user

description

   Compute the base 2 log of the second data column of a histog file
   from the genhis program.  This allows the double log to be   
   plotted using xyplop's log feature.  On a double log plot, a
   Gaussian distribution should show up as a slope of exactly -2.

examples

documentation

see also

   {source of histog file:} genhis.p

   {program to plot the result:} xyplo.p

author

   Thomas Dana Schneider

bugs

   Only the second column is computed.
   The parameter file could be used to set the base.

technical notes

*)
(* end module describe.log *)

(* taken from genhis: *)
(* begin module histogram.const *)
      maxslots = 20000;  (* maximum number of slots in the histogram *)
      defslots = 100;  (* the default number of slots *)
      pageheight = 53; (* was 105; *) (* room on page for plotting;
             used to set scaling.  pageheight should be set to 27
             less than the number of spaces
             allowed across a page for your printer. *)
      dch = '+';  (* character for plotting data *)
      sch = ':';  (* character for plotting standard *)
      bch = '*';  (* character for plotting coincidence of standard and data *)
      wid = 10; (* width of numbers to use for data *)
      dec = 5; (* number of decimal places to use for data *)
(* end module histogram.const *)

var
   histog, (* file used by this program *)
   logp, (* file used by this program *)
   xyin: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.57; (@ of prgmod.p 2002 Sep 5 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.57; (@ of prgmod.p 2002 Sep 5 *)

(* begin module log.themain *)
procedure themain(var histog, logp, xyin: text);
(* the main procedure of the program *)
var
   logvalue: real; (* the log of the value *)
   ln2: real; (* ln(2) *)
   parameterversion: real; (* parameter version number *)
   position: real; (* position in histogram *)
   value: real; (* value in histogram at that position *)
begin
   writeln(output,'* log ',version:4:2);
   reset(logp);
   readln(logp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   ln2 := ln(2);

   reset(histog);
   rewrite(xyin);
   writeln(xyin,'* log ',version:4:2);

   while not eof(histog) do begin
{writeln(output,'here 1');}
      if histog^ = '*' then begin
{writeln(output,'here 1.5');}
         copyaline(histog, xyin);
      end
      else begin
{writeln(output,'here 2');}
         read(histog, position, value);

         if value > 0
         then logvalue := ln(value)/ln2
         else logvalue := -maxint;

{
write(output, position:wid:dec,
                  ' ', value:wid:dec,
                  ' ', logvalue:wid:dec);
}

(*
         write(xyin, position:wid:dec,
                ' ', value:wid:dec,
                ' ', logvalue:wid:dec);
*)

         write(xyin, position:wid:dec,
                ' ', logvalue:wid:dec,
                ' ', value:wid:dec);

{
writeln(output,'here 3');

if eoln(histog) then halt;
writeln(output,'here 3.7');

if eof(histog) then halt;
writeln(output,'here 3.5');
}

         copyaline(histog, xyin); (* copy the original graphic *)
{
writeln(output,'here 4');
}

(*
         readln(histog);
         writeln(xyin);
*)

      end
   end;

end;
(* end module log.themain *)

begin
   themain(histog, logp, xyin);
1: end.


























