program malopt(optalign, maloptp, xyin, output);
(* malopt: 2d plot of optimal alignments

  Tom Schneider
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.09; (* of malopt.p 1996 March 22
origin 1996 March 21 *)
(* end module version *)

(* begin module describe.malopt *)
(*
name
   malopt: 2d plot of optimal alignments

synopsis
   malopt(optalign: in, maloptp: in, xyin: out, output: out)

files
   optalign:  output of malign

   maloptp:  parameters to control this program.  Lines:
       1: doclasses: integer number of classes to process
       2: colorbound: highest distance of the color scale
       3: xrectangle, yrectangle: size of rectangles of colors.
          Values of 1 and 1 give nice squares that just fit together.
          Values of 0.9 and 0.9 squares that don't touch each other.
          (This would allow more precise location of particluar
           pieces of data.)

   xyin:  input to xyplo.  Columns:
       1: x class
       2: y class
       3: x H, uncertainty in bits
       4: y H, uncertainty in bits
       5: x rectangle width
       6: y rectangle height
       7: distance from x class to y class
       8: hue for color computed from the distance
       9: saturation for color
      10: brightness for color

       A color scale is put on the left side of the graph,
       running from 0 to colorbound.
       The lower right triangle of the graph gives the distances
       according to these colors.
       
   output: messages to the user

description
   How can we look at all the classes produced by malign?  We can first
   normalize each alignment vector by subtracting the first value from all
   other values.  Then we can ask what the distance is between all classes.
   These data are put into xyin for plotting by the xyplo program.

   The distances are computed and reported to xyin.  The distances are also
   linearly transformed to give a color scale.  The user can define the upper
   bound of the color scale, to be able to look at details for smaller
   distances.  Distances larger than this "colorbound" are set to black.

examples

documentation

see also
   malign.p, maloptp, xyplo.p, xyplop.malopt

author
   Thomas Dana Schneider

bugs

technical notes
   See xyplo for notes on color.

*)
(* end module describe.malopt *)

(* begin module malopt.const *)
   maxclasses = 100; (* maximum number of alignment classes that can be
                        handled *)
   maxsequences = 300; (* maximum number of sequences that can be handled *)
   nwid = 4; (* width of integers *)
   wid = 10; (* width of numbers *)
   dec = 5; (* decimal places of numbers *)
(* end module malopt.const *)

var
   optalign, (* file used by this program *)
   maloptp, xyin: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.15; (@ of prgmod.p 1994 November 12 *)

(* begin module malopt.themain *)
procedure themain(var optalign, maloptp, xyin: text);
(* the main procedure of the program *)
var
   avalue: integer; (* a value of the vector *)
   brightness: integer; (* color brightness *)
   class: integer; (* supposedly the class we are working on *)
   classes: integer; (* number of classes of alignments *)
   colorbound: real; (* highest distance to give colors for *)
   d: real; (* square of difference *)
   distance: real; (* a distance between two classes *)
   doclasses: integer; (* number of classes of alignments to handle *)
   H: array[0..maxclasses] of real; (* uncertainty for each class *)
   saturation: integer; (* color saturation *)
   s: integer; (* index to sequences *)
   sequences: integer; (* number of sequences being aligned *)
   sumsq: real; (* sum of squares *)
   theclasses: array[1..maxclasses,1..maxsequences] of integer;
      (* the vectors of each class *)
   xclass: integer; (* x class of alignment we are working on *)
   xrectangle: real; (* the length of the x axis of the rectangles
      to be drawn by xyplo *)
   yclass: integer; (* y class of alignment we are working on *)
   yrectangle: real; (* the length of the y axis of the rectangles
      to be drawn by xyplo *)
   zerovalue: integer; (* the first value of the vector for normalizing *)

function hue(color: real): real;
(* color hue.  User linear transform gives range to use.
Then the next transform gives a good color range. *)
begin
hue := color/colorbound
end;

procedure columndefinition;
(* define xyin columns *)
begin
   writeln(xyin, '*');
   writeln(xyin, '* data columns:');
   writeln(xyin, '*  1: x class');
   writeln(xyin, '*  2: y class');
   writeln(xyin, '*  3: x H, uncertainty in bits');
   writeln(xyin, '*  4: y H, uncertainty in bits');
   writeln(xyin, '*  5: x rectangle width');
   writeln(xyin, '*  6: y rectangle height');
   writeln(xyin, '*  7: distance from x class to y class');
   writeln(xyin, '*  8: distance converted to color');
   writeln(xyin, '*  9: hue for color');
   writeln(xyin, '* 10: saturation for color');
   writeln(xyin, '* 11: brightness for color');
   writeln(xyin, '*');
end;

procedure generateoutput;
(* generate the output to xyin *)
begin
   (* compute cutoff for colors *)
   if distance > colorbound
   then brightness := 0
   else brightness := 1;
   write(xyin,' ',xclass:nwid);
   write(xyin,' ',yclass:nwid);
   write(xyin,' ',H[xclass]:wid:dec);
   write(xyin,' ',H[yclass]:wid:dec);
   write(xyin,' ',xrectangle:nwid:dec);
   write(xyin,' ',yrectangle:nwid:dec);
   write(xyin,' ',distance:wid:dec);
   write(xyin,' ',hue(distance):wid:dec);
   write(xyin,' ',saturation:nwid);
   write(xyin,' ',brightness:nwid);
   writeln(xyin);
end;

begin
   writeln(output,'malopt ',version:4:2);
   reset(optalign);

   (* read parameters *)
   reset(maloptp);
   readln(maloptp, doclasses);
   readln(maloptp, colorbound);
   readln(maloptp, xrectangle, yrectangle);

   (* determine number of classes *)
   while optalign^ = '*' do readln(optalign);
   readln(optalign); (* skip blank line *)
   readln(optalign); (* skip sequences *)
   readln(optalign); (* skip max R line *)
   classes := 0;
   while not eof(optalign) do begin
      (* count blank lines *)
      if eoln(optalign)
      then classes := succ(classes);
      readln(optalign)
   end;

   (* start output *)
   rewrite(xyin);
   writeln(xyin,'* malopt ',version:4:2);
   reset(optalign);
   while optalign^ = '*' do copyaline(optalign,xyin);
   readln(optalign); (* skip blank line *)
   readln(optalign, sequences);
   readln(optalign); (* skip max R line *)

   writeln(xyin, '* ',classes:1,  ' classes found');
   writeln(output,    classes:1,  ' classes found');
   writeln(xyin, '* ',sequences:1,' sequences aligned');
   writeln(output,    sequences:1,' sequences aligned');

   writeln(xyin, '*');
   writeln(xyin, '* maloptp:');
   writeln(xyin,'* ',doclasses:10,'1: doclasses: integer number',
                   ' of classes to process');
   writeln(xyin,'* ',colorbound:wid:dec,'2: colorbound:',
                   ' maximum color distance');
   columndefinition;

   if doclasses > classes then begin
      writeln(output,'You asked to process ',doclasses:1,
                     ' but there are only ',classes:1,' classes');
      writeln(output,classes:1,' will be processed');
      doclasses := classes;
   end;

   if doclasses > maxclasses then begin
      writeln(output,'You asked to process ',doclasses:1,
                     ' but I can only handle ',maxclasses:1,' classes');
      writeln(output,maxclasses:1,' will be processed');
      doclasses := maxclasses;
   end;

   (* generate color scale *)
   saturation := 1;
   brightness := 1;
   for yclass := 0 to doclasses - 1 do begin
      xclass := 0;
      distance := colorbound*(yclass/doclasses);
      (* set these into the H columns in case those are used *)
      H[xclass] := 0;
      H[yclass] := yclass;
      generateoutput;
   end;

   (* load up class vectors *)
   for xclass := 1 to doclasses do begin
      readln(optalign); (* skip blank line *)
      read(optalign,class);
      (* locate the = sign: *)
      while (optalign^ <> '=') do get(optalign);

      get(optalign); (* move past the '=' *)
      readln(optalign, H[xclass]);
      if class <> xclass then begin
         writeln(output,'class misalignment');
         halt
      end;
      (* read the value for zeroing *)
      read(optalign,zerovalue);
      theclasses[xclass,1] := 0; (* subtract zerovalue from itself! *)
      for s := 2 to sequences do begin
         read(optalign, avalue);
         theclasses[xclass,s] := avalue - zerovalue;
      end;
      readln(optalign);
   end;
   writeln(output,'class vectors loaded');

   for xclass := 1 to doclasses do begin
      for yclass := 1 to xclass do begin
         (* compute distance *)
         sumsq := 0.0;
         for s := 1 to sequences do begin
            d := theclasses[xclass,s]-theclasses[yclass,s];
            sumsq := sumsq + d*d;
         end;
         distance := sqrt(sumsq);

         (* compute cutoff for colors *)
         if distance > colorbound
         then brightness := 0
         else brightness := 1;

         generateoutput;
      end;
   end;
   writeln(output,'done');

end;
(* end module malopt.themain *)

begin
   themain(optalign, maloptp, xyin);
1: end.
