program merge(afile,bfile,apfile,bpfile,output,input);
(* merge: compare two files and merge them
      by thomas schneider
      copyright 1989

modules from: delman, prgmods
*)

label 1; (* end of program *)

const
(* begin module version *)
version = 9.53; (* of merge 1989 May 1 *)
(* end module version *)

(* begin module describe.merge *)
(*
name
      merge: compare two files and merge them

synopsis
      merge(afile: in, bfile: in, apfile: out, bpfile: out,
            output: out, input: intty)

files
      afile: the first input file
      bfile: the second input file
      apfile: the afile with corrections from bfile or the user
      bpfile: the bfile with corrections from afile or the user
      output: messages to the user
      input: interactive input from the user.

description
      the merge program was designed to aid in the entry of sequences.
      merge will also compare any two files for differences.
      two typed copies of the data are made (afile and bfile).  merge will
      compare the files, ignoring spaces and end-of-lines.  this allows
      the two data files to be typed independently by two people in two
      formats.  differences between the files are flagged, and the user
      may then indicate which file is correct and merge will fix the other
      file.  the user may also modify the files using a small
      editing facility.  the changes go to the prime files (apfile, bpfile).
      to be sure that apfile and bpfile are identical after the merge,
      you can merge them again.  several commands can be put on one line
      separated by blanks.  if you type an unrecognizable command, or ask for
      help at any time then merge will list the commands available.

examples
      when two sequences were compared, the program gave this output:
i am 91% sure that this has a deletion in b (insertion in a) of 5 characters:
file a: line 1
aatccttatccctcctaatttcgtttttgct
       >iiiii<        x         at 9 insertion, 1 mismatch downstream
file b: line 1
aatccttacctaatttcctttttgct
       ><        x         at 9 deletion, 1 mismatch downstream
      the sequences matched before the points indicated.  file a had
      an insert of "tccct" followed several bases later by a c to g change.
      the mismatch made merge less sure of the deletion.
      one must look at the original sequence to make the correction.

author
      thomas d. schneider

bugs
      1. lines without any characters are not copied from the file to its
      pfile.  see procedure readline.
      2. excessive blank characters may fool the guess procedure, since it
      does not remove blanks before guessing.  removing blanks would make
      it difficult to autofix, and the spacings would be lost in the pfile.
      3. the program can not compare more than one line from each file at a
      time,  so the guess is limited to what is visible on a line.  the entire
      program must be rewritten to allow multiple line guessing.
*)
(* end module describe.merge *)

      blank = ' ';

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const version = 'prgmod 3.96  85 mar 18 tds'; *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type version = 'prgmod 3.96  85 mar 18 tds'; *)

      line = record  (* a line of text *)
         str: string;  (* a string as defined in interact.type *)
         number:  integer; (* the number of the line in its file *)
      end;

      (* record of corrections *)
      filecorrection = (ofa, ofb);
      typecorrection = (mismatch,deletion,insertion,unidentified);

var
      (* the two input files to compare, and the merged output files *)
      afile, bfile, apfile, bpfile: text;

      aline, bline: line; (* lines of afile and bfile *)
      done: boolean; (* is this task done? *)

      scanning: boolean; (* flag all further differences between
                            apfile and bpfile *)

      (* statistics records for detected differences *)
      characters, (* number of characters that matched so far *)
      mism, (* number of mismatches found *)
      deli,       (* number of deletiions/insertions found *)
      unid:       (* number of unidentified differences found *)
         integer;

      (* records of changes made *)
      correction: array[ofa..ofb, mismatch..unidentified] of integer;

      (* indicies for correction *)
      offile: filecorrection;
      oftype: typecorrection;
      total, subtotal: integer; (* to figure totals *)
      ofletter: char; (* to write out which file is being described *)

      (* input buffer string: *)
      buffer: string;
      gotten: boolean; (* true if there is stuff in the buffer *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.54 86 nov 12 tds/gds' *)

(* begin module package.interact *)
(* ************************************************************************ *)
(* begin module interact.prompt *)
procedure prompt(var afile: text);
(* prompt a file.
   the prompt is sent to the output file, and
   a line is read into the pascal line buffer.
   (for the cyber system this means to readln afile.)
   guarantee no bomb *)
begin (* prompt *)
      if eof(afile) then reset(afile);
      readln(afile);
end; (* prompt *)
(* end module interact.prompt version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.readchar *)
procedure readchar(var afile:text; var ch: char);
(* read a character from afile, guarantee no bomb *)
begin (* readchar *)
      if eof(afile) then prompt(afile);
      read(afile,ch);
(*writeln(output,'"',ch,'"') *)
end; (* readchar *)
(* end module interact.readchar version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.readstring *)
procedure readstring(var afile: text; var line: string);
(* read in a string from afile, protect against bombing *)
var
      index: integer; (* for line *)
      cha: char; (* a character read in *)
      done: boolean; (* used for removing trailing blanks from the line *)
      acceptable: boolean; (* was the line typed short enough? *)
begin (* readstring *)
      with line do begin
         repeat
            clearstring(line);
            prompt(afile);
            index := 0; (* we now count characters *)
            while (not eoln(afile)) and (index < maxstring) do begin
               index:=succ(index);
               readchar(afile,cha);
               line.letters[index]:=cha
            end;
            if not eoln(afile)
            then begin
               writeln(output, 'type lines shorter than ', (maxstring+1): 1,
                               ' characters.  please retype the line...');
               acceptable := false
            end
            else acceptable := true
         until acceptable;

         length := index;
         if length > 0 then begin
            done := false;
            repeat (* remove blanks from the line.  note that a while loop
                  can not be used because one must avoid letters[0], since that
                  position does not exist... *)
               if letters[length] = ' '
               then length := pred(length)
               else done := true;

               if length = 0 then done := true
            until done
         end;

         if length > 0 then current := 1
                       else current := 0
      end
end; (* readstring *)
(* end module interact.readstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.figurestring *)
procedure figurestring(
      var line: string; (* a string of characters to figure out *)
      var first: integer; (* first found non-blank character in the line *)
      var last:  integer; (* last character before a blank after first *)
      var whzat: char; (* what the token is *)
      var c:     char;  (* the first character of the token *)
      var i:     integer; (* integer value of token if it is integer; or 0 *)
      var r:     real); (* the real value if it is real; or 0.0 *)
(* figurestring figures out the tokens in a string.
 it recognizes words, integers, reals and poorly formed numbers.
 you can easily use it to parse lines.

our goal is to figure out what thing is on a string.
start looking at the current place on the line.
first and last are the first 'token' in line after start.
the current place is updated to the letter after last.
the thing found is described by the value of whzat:
      'c': character (when the token does not begin with a digit, '+', or '-')
      'i': integer
      'r': real
      ' ': blank line
      'g': garbage, cannot figure it out
and the value of the thing found is the appropriate variable *)
var
      numbers: set of '0'..'9';
      sign: integer; (* sign of a number *)
      numberstart: integer; (* the point a number starts,
                               beyond its sign, if any *)
      point: integer; (* location of decimal point *)
      power: integer; (* of 10 that a number is *)

      l: integer; (* an index for dissecting numbers *)

function figureinteger(first,last:integer):integer;
(* figure the integer in the token *)
var
      i: integer; (* index *)
      sum, increment: integer;
begin (* figureinteger *)
      power:=1; (* start at ones place *)
      sum:=0; (* start sum at zero *)
      for i:=last downto first do begin
         case line.letters[i] of
            '0': increment:=0;
            '1': increment:=1;
            '2': increment:=2;
            '3': increment:=3;
            '4': increment:=4;
            '5': increment:=5;
            '6': increment:=6;
            '7': increment:=7;
            '8': increment:=8;
            '9': increment:=9
         end;
         sum:=sum+power*increment;
         power:=power*10
      end;
      figureinteger:=sum
end; (* figureinteger *)

begin (* figurestring *)
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
(* c:=' '; i:=0; r:=0.0; do not affect these variables unless necessary *)
      point:=0;
      whzat := '.'; (* assume that we have someting to work on *)
      (* now to see if that is true: *)

      with line do if (length = 0) or (current < 1) or (current > length)
      then whzat := ' '
      else begin
         (* figure out where the first token is in the line *)
         first:=line.current;
         while (line.letters[first]=' ') and (first < line.length)
            do first:=succ(first);
         if (first = line.length) and (line.letters[first] = ' ')
            then whzat := ' ';
      end;

      if whzat <> ' ' then begin
         last:=first;
         while (line.letters[last]<>' ') and (last < line.length)
            do last:=succ(last);
         if line.letters[last] = ' ' then last := pred(last);

         (* the token is between inclusive first and last *)

         c:=line.letters[first];
         if (c in numbers) or (c in ['+','-']) then begin
            if c in ['+','-'] then begin
               case c of
                  '+': sign:=+1;
                  '-': sign:=-1;
               end;
               numberstart:=succ(first)
            end
            else begin
               sign:=+1;
               numberstart:=first
            end;

            whzat:='i';
            for l:=numberstart to last do begin
               if not(line.letters[l] in numbers)
               then if line.letters[l]='.' (* we found a period *)
                  then if whzat='i' (* if so far it is numbers *)
                     then begin
                        whzat:='r'; (* it is actually real *)
                        point:=l
                     end
                     else whzat:='g' (* it is a second '.', ie garbage *)
                  else whzat:='g' (* it is garbage *)
            end;
            (* if it is only numbers, it is integer *)

            (* build number *)
            (* if it ends in a period, it is integer *)
            if (whzat = 'r') and (point = last) then whzat:='i';

            if whzat = 'i' then begin
               if point = last (* had an ending decimal point *)
               then i:=sign * figureinteger(numberstart,pred(last))
               else i:=sign * figureinteger(numberstart,last);
               r:=i
            end
            else if whzat = 'r' then begin
               i:=figureinteger(numberstart,point-1);
               r:=sign * (i+figureinteger(point+1,last)/power);
               i:=sign * i
            end
         end
         else begin
            whzat:='c';
         end;

         (* move the start to just beyond the last character of the token *)
         line.current:=succ(last)
      end
end; (* figurestring *)

(* end module interact.figurestring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* ************************************************************************ *)
(* end module package.interact version = 'prgmod 3.96  85 mar 18 tds'; *)


(* begin module package.interact.gets *)
(* ************************************************************************ *)
(* begin module interact.nostring *)
function nostring(var buffer: string): boolean;
(* true if there are no characters in the rest of the buffer;
   false if there are characters.
   also, if there is no buffer, then buffer.length is set to 0 *)
var   answer: boolean; (* the answer returned *)
procedure kill;
(* destroy the line *)
begin (* kill *)
      answer := true; (* blood and gore *)
      clearstring(buffer) (* total death *)
end; (* kill *)
begin (* nostring *)
      with buffer do begin
         if length > 0 then begin
            if length < maxstring
            then while (letters[current] = ' ') and (current < length)
               do current := succ(current);

            if current <= maxstring
            then if letters[current] = ' '
                 then kill
                 else answer := false
            else kill
         end
         else kill
      end;
      nostring := answer
end; (* nostring *)
(* end module interact.nostring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.flagstring *)
procedure flagstring(var afile: text; var buffer: string);
(* flag an error in the buffer at the current place, and clear the buffer *)
begin (* flagstring *)
      with buffer do length := current; (* chop off the rest of the buffer *)
      writestring(afile, buffer); (* show the buffer *)
      write(afile,'? ');
      clearstring(buffer)
end; (* flagstring *)
(* end module interact.flagstring version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.getchar *)
procedure getchar(var afile: text; var buffer: string;
                  var cha: char; var gotten: boolean);
(* get a character from the buffer, or refill the buffer and
   let the calling program figure out whether the buffer has non
   blank characters in it. *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      int: integer;
      rea: real;
begin (* getchar *)
      if buffer.length = 0 then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else begin
         figurestring(buffer,first,last,what,cha,int,rea);
         gotten := (what <> ' ')
      end
end; (* getchar *)
(* end module interact.getchar version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.getinteger *)
procedure getinteger(var afile: text; var buffer: string;
                     var int: integer; var gotten: boolean);
(* get the integer int from the buffer or interactive file afile *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      rea: real;
begin (* getinteger *)
      if buffer.length = 0
      then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else begin
         figurestring(buffer,first,last,what,cha,int,rea);

         if what <> 'i'
         then begin
            flagstring(output,buffer);
            writeln(output,' please type an integer');
            gotten:=false
         end
         else gotten:=true
      end
end; (* getinteger *)
(* end module interact.getinteger version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.getreal *)
procedure getreal(var afile: text; var buffer: string;
                  var rea: real; var gotten: boolean);
(* get the real rea from the buffer or interactive file afile
integer values are also accepted. *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      int: integer;
begin (* getreal *)
      if buffer.length = 0
      then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else begin
         figurestring(buffer,first,last,what,cha,int,rea);

         if not (what in ['r', 'i'])
         then begin
            flagstring(output,buffer);
            writeln(output,' please type a real number');
            gotten:=false
         end
         else gotten:=true
      end;

      (* handle integers *)
      if what = 'i' then rea := int
end; (* getreal *)
(* end module interact.getreal version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.token *)
procedure token(var buffer, atoken: string; var gotten: boolean);
(* get a token from the buffer *)
var
      (* variables for calling figurestring: *)
      first: integer;
      last: integer;
      what: char;
      cha: char;
      int: integer;
      rea: real;
      index: integer; (* to the buffer *)
begin
      figurestring(buffer,first,last,what,cha,int,rea);

      if what = ' '
      then gotten := false
      else begin
         clearstring(atoken);
         for index := first to last
            do atoken.letters[index-first+1] := buffer.letters[index];
         atoken.length := last - first + 1;
         atoken.current := 1;

         gotten:=true
      end
end;
(* end module interact.token version = 'prgmod 3.96  85 mar 18 tds'; *)

(* begin module interact.gettoken *)
procedure gettoken(var afile: text; var buffer: string;
                   var atoken: string; var gotten: boolean);
(* get a token from the buffer or interactive file afile *)
begin (* gettoken *)
      if buffer.length = 0
      then begin
         gotten:=false;
         readstring(afile,buffer)
      end
      else token(buffer,atoken,gotten)
end; (* gettoken *)
(* end module interact.gettoken version = 'prgmod 3.96  85 mar 18 tds'; *)

(* ************************************************************************ *)
(* end module package.interact.gets version = 'prgmod 3.96  85 mar 18 tds'; *)


(****************************************************************************)
(* string manipulation procedures *)

procedure increment(var i: integer);
(* make incrementing easier *)
begin
      i:=succ(i)
end;

procedure clearline(var aline: line);
(* clear a line *)
begin
      with aline do begin
         clearstring(str);
         number:=0;
      end
end;

procedure writeline(var afile: text; aline: line);
(* write a line to a file *)
begin
      with aline do begin
         if str.length > 0 then begin
            writestring(afile,str);
            writeln(afile)
         end
      end
end;

procedure flagline(var afile: text; aline: line);
(* put a flag on a line: show its line number and line position *)
var
      p: integer; (* index to aline *)
begin
      with aline do begin
         writeln(afile,'line ',number:1);
         writeline(afile,aline);
         with str do begin
            for p:=1 to current-1 do write(afile,blank);
            writeln(afile,'^ at ',current:1)
         end
      end
end;

procedure show(a,b: line);
(* show two lines - this procedure is going the way of the dodo...  *)
begin
      writeln(output,'--');
      write(output,'file a: ');
      flagline(output,a);

      write(output,'file b: ');
      flagline(output,b);
end;

procedure fancyshow(aline, bline: line;
                    adel, bdel: integer; thetype: char);
(* like show, but rather fancy... *)
var   downstream: integer; (* the number of mismatches downstream *)
procedure pre(var afile: text; id: char; aline: line; del: integer);
(* preliminary presentation of a line to afile: print line id and
several blanks for showmis or showdel.  del is subtracted from the
number of blanks *)
var
      p: integer; (* index *)
begin (* pre *)
      with aline do begin
         writeln(afile,'file ',id,': line ' ,number:1);
         writeline(afile,aline);
         with str
            do for p := 1 to current - 1 - del
               do write(afile,blank)
      end
end; (* pre *)
procedure post(var afile: text;
               aline, bline: line; adel, bdel: integer);
(* write the mismatches on the rest of the line up to the end of either
line or a blank on either line *)
var
      pa, pb: integer; (* positions on the two lines *)
      a, b: char; (* letters at pa and pb *)
      done: boolean; (* are we done yet ma? *)
begin (* post *)
      downstream := 0;
      pa := aline.str.current + adel;
      pb := bline.str.current + bdel;

      repeat
         done := (pa > aline.str.length) or
                 (pb > bline.str.length);

         if not done then begin
            a := aline.str.letters[pa];
            b := bline.str.letters[pb];
            done := (a = ' ') or (b = ' ');

            if not done then begin
               if a <> b
               then begin
                  write(afile,'x');
                  increment(downstream)
               end
               else write(afile,' ');

               increment(pa);
               increment(pb)
            end
         end
      until done;

      write(afile,' at ',aline.str.current:1)
end; (* post *)
procedure showdown(var afile: text);
(* ok folks, stand back this will be a real shoot out...
no, kidding aside (get it, billy the kid??  ha ha)...
gee this could go on all day...
we are gathered here today to print the number of mismatches downstream. *)
begin (* showdown *)
      if downstream > 0
      then begin
         write(afile,', ',downstream:1,' mismatch');
         if downstream <> 1 then write(afile,'es');
         write(afile,' downstream')
      end
end; (* showdown *)
procedure showmis(var afile: text; id: char; aline, bline: line);
(* show mismatches for aline vs bline to afile *)
begin (* showmis *)
      pre(afile,id,aline,0);
      write(afile,'^'); (* the first mismatch *)
      post(afile,aline,bline,1,1);
      if thetype = 'm' then write(afile,' mismatch');
      showdown(afile);
      writeln(afile)
end; (* showmis *)
procedure showdel(var afile: text; id: char; aline, bline: line;
                                   adel,  bdel : integer);
(* show deletion for aline vs bline to afile *)
var   p: integer; (* index *)
begin (* showdel *)
      if aline.str.current > 1
      then begin
         pre(afile,id,aline,1);
         write(afile,'>');
      end
      else pre(afile,id,aline,0);

      for p := 1 to adel do write(afile,'i');
      write(afile,'<');
      post(afile,aline,bline,succ(adel),succ(bdel));

      if adel = 0
      then write(afile,' deletion')
      else write(afile,' insertion');
      showdown(afile);
      writeln(afile)
end; (* showdel *)
begin (* fancyshow *)
      case thetype of
         'm', 'u': begin
            showmis(output,'a',aline,bline);
            showmis(output,'b',bline,aline);
         end;
         'd': begin
            showdel(output,'a',aline,bline,adel,bdel);
            showdel(output,'b',bline,aline,bdel,adel)
         end
      end
end; (* fancyshow *)


procedure readline(var afile: text; var aline: line);
(* read a line from a file. aline.length=0 means eof(afile) was true
   and no line was read.  line.number is incremented to indicate the
   line number of a line in afile. *)
var
      skipping: boolean; (* true while skipping blank lines *)
begin
      with aline do begin
         (* skip extra lines.  they are not copied to the apfile.
         this loop is now designed so that the test for
         end of line is not done unless we are not at the
         end of file *)
         skipping := true;
         while skipping do begin
            if eof(afile)
            then skipping := false
            else if eoln(afile)
                 then begin
                    readln(afile);
                    increment(number)
                 end
                 else skipping := false
         end;

         (* read in a line *)
         with str do begin
            length:=0;
            if not eof(afile) then begin
               while (not eoln(afile)) and (length<maxstring) do begin
                  increment(length);
                  read(afile,letters[length])
               end;

               (* check for extra unread line *)
               increment(number);
               if not eoln(afile) then begin
                  writeln(output,'this line exceeds my buffer:');
                  current:=length;
                  flagline(output,aline);
                  writeln(output,'the excess will be lost.  continuing...');
               end;

               readln(afile)
            end;
            current:=0; (* force to zero *)
            end
      end
(*
;write(output,'readline:'); flagline(output,aline);
*)
end;

procedure nextnonblank(var afile, apfile: text; var aline: line);
(* get to the next non-blank character in aline, reading from
   afile if there is nothing left in aline *)
var
      done: boolean;
begin
      with aline do begin
         done:=false;
         with str do repeat
(*writeln(output,'loop nextnonblank, current=',current:1); *)
            increment(current);
            if current <= length
            then done:= (letters[current]<>blank)
            else begin
               (* dump previous line *)
               if length > 0 then writeline(apfile,aline);

               (* get next line *)
               readline(afile,aline)
            end;
            if length = 0 then done:=true  (* we are done if it is the eof *)
         until done
(*
;write(output,'nextnonblank:');
if length=0 then writeln(output,' eof')
            else writeln(output,' "',letters[current],'"');
*)
      end
end;

(*************************************************************************)
(* line editing procedures *)

procedure squack;
(* object to an insertion or replacement that is too big *)
begin
      writeln(output,'that insert would make the line longer than ',
                      maxstring:1,', my buffer length');
      writeln(output,'sorry, i cannot do it.');
      writeln(output,'your last typed line is lost.');
      writeln(output,'maybe you could use the break command.');
end;

function makewayline(amount: integer; var l: line): boolean;
(* make a gap of some amount in l starting just before l.str.current
   return true if the gap was made, false if it was not *)
var
      p: integer; (* position in a line *)
begin
      with l.str do begin
         if length + amount > maxstring
         then begin
            makewayline:=false;
            squack;
         end
         else begin
            makewayline:=true;

            for p:=length downto current
               do letters[p+amount]:=letters[p];
            length:=length+amount
         end
      end
end; (* makewayline *)

procedure replaceline(var i, l: line);
(* replace i line into l line at l.str.current *)
var
      p: integer; (* position in l *)
begin
         for p:=l.str.current to l.str.current+i.str.length-1
            do l.str.letters[p]:=i.str.letters[p-l.str.current+1]
end; (* replaceline *)

procedure copypartline(amount: integer; var fromline, toline: line);
(* copy from part of a line to another at the current places *)
var
      p: integer;
begin
      for p:=fromline.str.current to fromline.str.current+amount
         do toline.str.letters[p-fromline.str.current+toline.str.current]
            := fromline.str.letters[p]
end;

procedure insertinline(var i, l: line);
(* insert line i into line l *)
begin
      if makewayline(i.str.length,l)
      then replaceline(i,l)
end;

procedure deleteinline(d: integer; var l: line);
(* delete d letters from l at l.str.current *)
var
      p: integer;
begin
      with l.str do begin
         (* allow negative deletions *)
         if d<0 then begin
            if d+current < 1 (* case of deletion past start of line *)
            then begin
               d:=current-1;
               current:=1
            end
            else begin (* deletion in mid line *)
               current:=current+d;
               d:=-d
            end
            (* d is now a positive deletion *)
         end;

         (* delete in positive direction *)
         if d>0 then with l do begin
            if d+current > length
            then length:=current-1 (* delete very last character *)
            else begin
               for p:=current to length-d
                  do letters[p]:=letters[p+d];
               length:=length-d
            end
         end
      end
end;

(*************************************************************************)
(*************************************************************************)
(* procedures used by different *)

procedure guess(aline, bline: line;
                var adel, bdel: integer;
                var thetype: char);
(* guess how aline and bline differ.
   quality of a guess is (number of matches)/(number of characters
   looked at).  for dna 25 percent is expected by random match, so a good
   quality guess should be bigger than that.

   adel and bdel are the extent of deletions in a or b needed to best
   explain the difference.   (adel is how much aline must be deleted to make
   the lines the same.)


   thetype is the type of difference:
      thetype = 'm'; mismatch
      thetype = 'd'; deletion/insertion
      thetype = 'u'; unidentified

some notes on ranges of the constants
      setting foresight to 20 seemed to get guess into trouble with
      with downstream mismatches more often.  10 seems better.  when the
      quality is as low as 50 percent, the guesses are quite poor and often
      wrong, dispite the lower limit of 25 percent for dna.  for this reason
      i recommmend using a goodquality of at least 60 percent.
*)

const
      foresight = 10; (* how far ahead guess can look *)
      minsight = 5; (* minimum distance for allowing a guess *)
      goodquality = 0.60; (* the minimum quality of a printed guess *)

var
      besta, bestb: integer; (* the extent of deletions in afile or
            in bfile that leads to the highest quality (q). *)
      bestq: real; (* the best so far *)

      a, b: integer; (* tries for the best *)
      q: real;

function frame(adeletion, bdeletion: integer): real;
(* returns a number that is the quality of matches in this
   deletion frame (characters lost in afile or bfile),
   looking ahead a maximum distance: foresight. *)
var
      as, bs: integer; (* positions compared *)
      sight: integer; (* how far ahead we look *)
      matches: integer; (* number of matches *)
begin (* frame *)
      (* these -1 s compensate for moving as and bs foreward (see below) *)
      as:=aline.str.current+adeletion-1;
      bs:=bline.str.current+bdeletion-1;

      (* let mismatch test be scored one base ahead in the case of mismatch *)
      if (adeletion = 0) and (bdeletion = 0) then begin
         increment(as);
         increment(bs)
      end;

      sight:=0; matches:=0;

      while (sight <= foresight) and
            (as <= aline.str.length) and
            (bs <= bline.str.length) do begin
         (* we must first move as and bs foreward to account for any spaces *)
         repeat
            increment(as)
         until (aline.str.letters[as] <> blank) or (as > aline.str.length);

         repeat
            increment(bs)
         until (bline.str.letters[bs] <> blank) or (bs > bline.str.length);

         if (as <= aline.str.length) and (bs <= bline.str.length) then begin
            increment(sight);
            if aline.str.letters[as] = bline.str.letters[bs]
               then increment(matches)
         end
      end;

      (* quality of match *)
      if sight < minsight then frame:=0.0
                          else frame:=matches/sight
(*
;if sight=0 then sight:=1; (@ part of debug statements... *)
(*
;writeln(output,'frame adel:',adeletion:1,
                     ' bdel:',bdeletion:1,
                ' quality:', trunc(100*matches/sight):1);
*)
end; (* frame *)

begin (* guess *)
      (* initial guess is a mis match *)
      besta:=0;
      bestb:=0;
      bestq:=frame(besta,bestb);

      (* now see if deletion in a is better *)
      b:=0;
      for a:=1 to foresight do begin
         q:=frame(a,b);
         if q > bestq then begin
            besta:=a;
            bestb:=b;
            bestq:=q
         end
      end;

      (* now try deletions in b *)
      a:=0;
      for b:=1 to foresight do begin
         q:=frame(a,b);
         if q > bestq then begin
            besta:=a;
            bestb:=b;
            bestq:=q
         end
      end;

      adel:=besta; bdel:=bestb;

      (* decide if it is worth telling about *)
      if bestq >= goodquality then begin
         if bestq > 0.99 then bestq:=0.99; (* avoid being presumptious *)

         write(output,'i am ',round(100*bestq):1,
                      '% sure that this has a ');
         if (besta = 0) and (bestb = 0)
         then begin
            thetype:='m';
            writeln(output,'single mismatch:');
            increment(mism)
         end
         else begin
            thetype:='d';
            write(output,'deletion in ');
            if besta=0
            then write(output,'a (insertion in b) of ',bestb:1)
            else write(output,'b (insertion in a) of ',besta:1);
            write(output,' character');
            if besta+bestb>1 then writeln(output,'s:')
                             else writeln(output,':');

            increment(deli)
         end
      end
      else  begin
         thetype:='u';
         increment(unid);
      end
end;

(*************************************************************************)

procedure insert(var l: line);
(* insert text into a line *)
var
      i: line; (* temporary store for the insert *)
begin
      repeat
         if nostring(buffer) then writeln(output,'characters to insert: ');
         gettoken(input,buffer,i.str,gotten);
      until gotten;

      if i.str.length <> 0
         then insertinline(i,l)
         else writeln(output,'no insert')
end;

procedure delete(var l: line);
(* delete characters from a line *)
var
      d: integer; (* number of characters to delete *)
begin
      repeat
         if nostring(buffer)
            then writeln(output,'number of characters to delete');
         getinteger(input,buffer,d,gotten);
      until gotten;

      if d <> 0
         then deleteinline(d,l)
         else writeln(output,'no deletion')
end;

procedure jump(var l: line);
(* jump around on line *)
var
      distance: integer;
begin
      repeat
         if nostring(buffer) then writeln(output,'distance to jump');
         getinteger(input,buffer,distance,gotten);
      until gotten;

      with l.str do begin
         current:=current+distance;
         if current<1 then current:=1;
         if current>length+1 then current:=length+1; (* allow insert there *)
      end
end;

procedure replace(var l: line);
(* replace some characters in l *)
var
      r: line; (* temporary store for the replacement insert *)
begin
      repeat
         if nostring(buffer) then writeln(output,'new replacing characters:');
         gettoken(input,buffer,r.str,gotten);
      until gotten;

      if l.str.current + r.str.length > maxstring
         then squack
         else replaceline(r,l)
end;

procedure break(var lpfile: text; var l: line);
(* break l in two *)
var
      p: integer; (* position on l *)
      tc, tl: integer; (* temporary store for l.current and l.length *)
begin
      with l.str do begin
         if (current>1) and (current <= length) then begin
            (* store away current and length *)
            tc:=current;
            tl:=length;

            (* prepare to write lower part out *)
            current:=pred(current);
            length:=current;
            writeline(lpfile,l);

            (* restore sanity *)
            current:=tc;
            length:=tl;

            (* shove down *)
            for p:=1 to length-current+1
               do letters[p]:=letters[current+p-1];
            length:=length-current+1;
            current:=1
         end
      end
end;

procedure modifyhelp;
begin
flagstring(output,buffer);
writeln(output,'modification commands are:');
writeln(output,'i(nsert)  characters before current spot');
writeln(output,'d(elete)  characters starting from just before current spot');
writeln(output,'j(ump)    around on the line');
writeln(output,'r(eplace) several characters');
writeln(output,'b(reak)   line into two lines prior to current spot');
writeln(output,'          editing continues for second line only.');
writeln(output,'v(erify)  display both lines');
writeln(output,'q(uit)    to merge level');
writeln(output)
end;

procedure modifyline(var lpfile: text; var l: line; c: char);
(* modify line l. c identifies the file. *)
var
      command: char; (* users command *)
begin
      command:=blank;

      while command<>'q' do begin
         repeat
            if nostring(buffer) then writeln(output,'modify');
            getchar(input,buffer,command,gotten)
         until gotten;

         if not(command in ['i','d','r','j','b','v','q'])
         then modifyhelp
         else case command of
            'i': insert(l);
            'd': delete(l);
            'r': replace(l);
            'j': jump(l);
            'b': break(lpfile,l);
            'v': begin
                    fancyshow(aline,bline,0,0,'u');
                    writeln(output,'modifying ',c)
                 end;
            'q': ;
         end;

         if not (command in ['q','v']) and nostring(buffer) then begin
            writeln(output,'--');
            write(output,'file ',c,': ');
            flagline(output,l)
         end
      end;

      (* prevent strange current position but allow being off the end *)
      with l.str do begin
         if current > length then current:=length + 1;

         if length <= 0 then begin
            (* force rejection of this line by nextnonblank: *)
            length:=0;
            current:=length+5;  (* 5 = arbitrary number larger than length *)
            case c of
               'a': nextnonblank(afile,apfile,l);
               'b': nextnonblank(bfile,bpfile,l);
            end
         end
      end
end;

procedure modify(var apfile, bpfile: text; var a, b: line);
(* chose which line to modify *)
var
      choice: char; (* the line to be modified *)
begin
      repeat
         if nostring(buffer) then writeln(output,'chose a or b to modify');
         getchar(input,buffer,choice,gotten);
         if gotten then if not (choice in ['a','b'])
         then begin
            flagstring(output,buffer);
            gotten:=false
         end
      until gotten;

      case choice of
         'a': begin
            increment(correction[ofa,unidentified]);
            modifyline(apfile,a,choice)
         end;
         'b': begin
            increment(correction[ofb,unidentified]);
            modifyline(bpfile,b,choice)
         end;
      end;

      fancyshow(a,b,0,0,'u')
end;

procedure autofix(var aline, bline: line;
                  var adel, bdel: integer;
                  var thetype: char);
(* autofix automatically corrects one of the two lines.
   the user specifies which line is correct.
   inputs defined in different and guess.
   thetype is set to 'u' after autofix, so that a second call is ignored.
*)
var
      choice: char; (* choice of the correct file *)

procedure fix(    xdel, ydel: integer;
              var xline, yline: line;
                  ofy: filecorrection);
(* x is the correct line; fix y *)
begin (* fix *)
      case thetype of
         'm': begin
            increment(correction[ofy,mismatch]);
            yline.str.letters[yline.str.current]
                  := xline.str.letters[xline.str.current]
         end;
         'd': begin
            if xdel = 0
            then begin (* delete in y *)
               increment(correction[ofy,deletion]);
               deleteinline(ydel,yline)
            end
            else begin (* insert in y *)
               increment(correction[ofy,insertion]);
               if makewayline(xdel,yline)
               then copypartline(xdel,xline,yline)
            end
         end
      end
end; (* fix *)

begin (* autofix *)
(*writeln(output,'autofix,adel,bdel,thetype ',adel:1,' ',bdel:1,' ',thetype);*)
      if thetype = 'u'
      then begin
         writeln(output,'i do not know how to fix this.');
         buffer.length:=0
      end
      else if thetype in ['m','d'] then begin
         repeat
            if nostring(buffer) then writeln(output,
                               'which file is correct? (a, b, n(either))');
            getchar(input,buffer,choice,gotten);
            if gotten then if not (choice in ['a','b','n'])
            then begin
               flagstring(output,buffer);
               gotten:=false
            end
         until gotten;

         if choice in ['a','b']
         then begin
            case choice of
               'a': fix(adel,bdel,aline,bline,ofb);
               'b': fix(bdel,adel,bline,aline,ofa)
            end;
            thetype:='u' (* a second call will be ignored *)
         end
         else writeln(output,'automatic correction aborted.')
      end
      else halt; (* undefined type *)

      fancyshow(aline,bline,adel,bdel,thetype)
end; (* autofix *)

procedure move(var afile, apfile: text;
               var aline: line);
(* move ahead to the next line *)
begin (* move *)
      (* dump out the line *)
      writeline(apfile, aline);

      (* do the actual move *)
      readline(afile,aline);

      (* move away from current = 0 on the line *)
      nextnonblank(afile, apfile, aline)
end; (* move *)

procedure linejump(var afile, bfile, apfile, bpfile: text;
               var a, b: line);
(* skip over lines *)
var   choice: char; (* for the decision *)
begin (* linejump *)
      repeat
         if nostring(buffer)
         then writeln(output,'line to jump (a, b, 0 (none), or 2 (both))');
         getchar(input,buffer,choice,gotten);
         if gotten then if not(choice in ['a','b','0','2'])
            then begin
               flagstring(output,buffer);
               gotten := false
            end
         until gotten;

      case choice of
         'a': move(afile, apfile, a);
         'b': move(bfile, bpfile, b);
         '0': ;
         '2': begin
            move(afile, apfile, a);
            move(bfile, bpfile, b);
         end
      end
end; (* linejump *)

procedure skipout(var thefile, thepfile: text; var theline: line);
(* copy out the rest of thefile to thepfile *)
begin
      while theline.str.length<>0 do nextnonblank(thefile,thepfile,theline)
end;

procedure differhelp;
(* help for difference procedure (merge command state) *)
begin (* differhelp *)
flagstring(output,buffer);
writeln(output,'merge commands are:');
writeln(output,'c(ont)   continue looking for differences');
writeln(output,'f(ix)    automatic correction using the correct line');
writeln(output,'j(ump)   over one or both lines (this will ignore any',
                         ' differences)');
writeln(output,'m(odify) one of the lines by hand');
writeln(output,'n(umber) of characters identical so far is shown.  blanks');
writeln(output,'         are ignored.  this helps locate a difference.');
writeln(output,'q(uit)   like scan, but do not show anything');
writeln(output,'s(can)   ignore all further differences (but show them)');
writeln(output,'v(erify) redisplay of both lines');
writeln(output,'you can type on one line seperated by spaces:');
writeln(output,'m a j 3 r a q');
writeln(output)
end; (* differhelp *)

procedure different(var afile, bfile, apfile, bpfile: text;
                    var aline,bline: line; var done: boolean);
(* aline<>bline.  deal with this:
   tell the user.
   let the user modify the lines.
*)
var
      command: char; (* users command *)
      adel, bdel: integer; (* guess of deletion extents *)
      thetype: char; (* guess of the type of differences *)
      finished: boolean; (* used to move back to just before the difference *)
begin (* different *)
      writeln(output);
      write(output,'---');

      if scanning
      then writeln(output,' ',characters:1,' characters were identical before',
                          ' this difference:')
      else writeln(output,' difference:');

      adel:=0;
      bdel:=0;
      thetype:=' ';

      guess(aline,bline,adel,bdel,thetype);
      fancyshow(aline,bline,adel,bdel,thetype);


      command:=blank;
      clearstring(buffer);

      while not (command in ['c','q','s']) and (not scanning) do begin
         repeat
            if nostring(buffer) then writeln(output,'merge');
            getchar(input,buffer,command,gotten);
         until gotten;

         if not (command in ['c','q','s','m','n','f','v','j'])
         then differhelp
         else case command of
            'c': ; (* continue *)
            'q': begin
               clearstring(buffer); (* no typing ahead allowed *)
               repeat
                  if nostring(buffer) then
                  writeln(output,'do you really want to quit? (y/n)');
                  getchar(input,buffer,command,gotten);
               until gotten;
               if command='y'
               then begin
                  skipout(afile,apfile,aline);
                  skipout(bfile,bpfile,bline);
                  writeln(output,'the rest of the input files was ',
                                 'copied to their respective output files.');
                  done:=true;
                  command:='q'
               end
               else command:=blank
            end;
            's': begin
               clearstring(buffer); (* no typing ahead allowed *)
               repeat
                  if nostring(buffer) then
                  writeln(output,'do you really want to scan? (y/n)');
                  getchar(input,buffer,command,gotten);
               until gotten;
               if command='y' then scanning := true
            end;

            'm': begin
                    thetype:='u'; (* prevent autofix from being used later *)
                    modify(apfile,bpfile,aline,bline)
                 end;
            'n': writeln(output,' number of characters found identical',
                                ' so far: ',characters:1);
            'f': autofix(aline,bline,adel,bdel,thetype);
            'v': fancyshow(aline,bline,adel,bdel,thetype);
            'j': linejump(afile,bfile,apfile,bpfile,aline,bline)
         end
      end;

      (* we now step back to recheck this last change.
         we cannot check everything, but for a single mismatch, or
         a replace, we may as well.
         we keep reducing current until we find a non blank character.
         if current = 0 then the nextnonblank procedure will move current
         back onto the line.

         if the command was halt, then we cannot backup (the line length is 0).

         if scanning = true, to backup would make an infinite loop. *)

      if command in ['c','s'] then begin
         with aline.str
            do repeat
               current:=pred(current);
               if current = 0
               then finished:=true
               else finished:=(letters[current]<>blank)
            until finished;
         with bline.str
            do repeat
               current:=pred(current);
               if current = 0
               then finished:=true
               else finished:=(letters[current]<>blank)
            until finished
      end
      else if scanning then begin
         if thetype = 'd' then begin
            (* skip the insertion *)
            with aline.str do current := current + adel;
            with bline.str do current := current + bdel
         end
      end

end; (* different *)

begin (* merge *)
      writeln(output,'merge ',version:4:2);

      reset(afile);
      reset(bfile);
      rewrite(apfile);
      rewrite(bpfile);
      scanning:=false;
      characters:=0;
      mism:=0;
      deli:=0;
      unid:=0;
      for offile:=ofa to ofb
         do for oftype:=mismatch to unidentified
            do correction[offile,oftype]:=0;

      clearline(aline);
      clearline(bline);

      readline(afile,aline);
      readline(bfile,bline);

      done:=false;
      while not done do begin
         (* step one character foreward in each file *)
         nextnonblank(afile, apfile, aline);
         nextnonblank(bfile, bpfile, bline);

         (* decide on doneness *)
         if aline.str.length=0 then done:=true;
         if bline.str.length=0 then done:=true;

         if done
         then begin (* if one file ended early... *)
            if ((aline.str.length =0) and (bline.str.length<>0)) or
               ((aline.str.length<>0) and (bline.str.length =0))
            then begin
               (* this keeps things seperate *)
               writeln(output,'--');

               (* tell which line it was *)
               if aline.str.length=0
                  then write(output,'first')
                  else write(output,'second');
               write(output,' file ended early.  ');

               (* show the other line *)
               write(output,'the other file is at ');
               if aline.str.length=0
                  then flagline(output,bline)
                  else flagline(output,aline)
            end;
         end

         (* not done yet *)
         else begin
            if aline.str.letters[aline.str.current] <>
               bline.str.letters[bline.str.current]
            then begin
               different(afile, bfile, apfile, bpfile, aline, bline, done);
               if scanning then increment(characters)
            end
            else increment(characters)
         end
      end;

      writeln(output);
      writeln(output,characters:1,' non-blank characters were identical.');
      writeln(output);

      if mism+deli+unid = 0
      then writeln(output,'no differences')
      else begin
         writeln(output,mism:5,' mismatches');
         writeln(output,deli:5,' deletion/insertions');
         writeln(output,unid:5,' unidentified changes');
         writeln(output,'=====');
         writeln(output,mism+deli+unid:5,' detected differences');
         writeln(output);

         (* give info on corrections made *)
         total := 0;
         for offile := ofa to ofb
            do for oftype := mismatch to unidentified
               do total := total + correction[offile,oftype];

         if total > 0 then begin
            writeln(output,'corrections made:');
            writeln(output,'to file ','  mismatch  ',' deletion',' insertion',
                           ' unidentified',' total');
            ofletter:='a';
            for offile:=ofa to ofb do begin
               write(output,'    ',ofletter); (* 4 blanks *)
               total:=0;
               for oftype:=mismatch to unidentified do begin
                  write(output,' ',correction[offile,oftype]:9);
                  total:=total+correction[offile,oftype]
               end;
               writeln(output,' ',total:9);
               ofletter:=succ(ofletter)
            end;
            total:=0;
            write(output,'total');
            for oftype:=mismatch to unidentified do begin
               subtotal:=correction[ofa,oftype]+correction[ofb,oftype];
               write(output,' ',subtotal:9);
               total:=total+subtotal
            end;
            writeln(output,' ',total:9);
         end
         else writeln(output,'no corrections made');

         writeln(output);
         writeln(output,'input files were not modified.');
         writeln(output,'default output is in apfile and bpfile.')
      end;
1: end. (* merge *)
