program rebasec(enzymelist, data, output);
(* rebasec: convert restriction enzyme list to dimensionality

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.00; (* of rebasec.p 2019 Sep 08
2019 Sep 08, 1.00 origin from rebasec.p 2018 Jan 06
*)
(* end module version *)

(* begin module describe.rebasec *)
(*
name
   rebasec: convert restriction enzyme list to dimensionality

synopsis
   rebasec(enzymelist: in, data: out, output: out)

files
   enzymelist: a simple list of restriction enzymes with only alphabetic
       characters, one per line.

       The data could be from Rich Roberts' restriction enzyme
       database, REBASE

   data: Twice Rsequence for each enzyme.  Columns:
        DNA sequence
        b bases
        R bits
        D dimensionality
     Lines that begin with '*' are comments.
   output: messages to the user

description

   Take a raw list of restriction enzyme sites and convert

   into its dimensionality, D = 2 Rsequence.  The genhis program can
   then be used to generate a histogram using the third column.

   The list could come from Rich Robert's restriction enzyme database.

examples

documentation

   The enzymelist file is available from
   ftp://vent.neb.com/pub/rebase/enzymelist*.

see also
   genhis.p
   rebased.p

   ftp://vent.neb.com/pub/rebase/

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.rebasec *)

   bufmax = 100; (* size of line buffer *)

type
   linebuffer = array[1..bufmax] of char; (* line buffer *)

var
   enzymelist: text; (* input to the program *)
   data: text; (* output of the program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.86; (@ of prgmod.p 2004 Sep 8 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module rebasec.themain *)
procedure themain(var infile, outfile: text);
(* the main procedure of the program *)
const
   wid = 5; (* width of numbers output *)
   dec = 2; (* decimals of numbers output *)
var
   bad: boolean; (* true when odd characters are found on a data line *)
   b: integer; (* index to the buffer *)
   blength: integer; (* count of characters on the line *)

   w: integer; (* index for writing the characters on the line *)
   buffer: linebuffer; (* line buffer *)
   c: char; (* a character in the outfile file *)
   Rs: real; (* Rsequence (or its running sum) *)
   T: real; (* the information for a base which choses 3 out of 4 *)

{
   name: linebuffer; (* primary '<1>' name of the enzyme *)
   nlength: integer; (* length of the name *)
   id: char; (* a character in the outfile file, the data type id *)
   clength: integer; (* count of characters on the companylist *)
   companies: linebuffer; (* '<7>' name of the companies *)
}

begin
   writeln(output,'rebasec ',version:4:2);

   rewrite(outfile);
   writeln(outfile,'* rebasec ',version:4:2);
   writeln(outfile,'* columns:');
   writeln(outfile,'*   DNA sequence');
   writeln(outfile,'*   b bases');
   writeln(outfile,'*   R bits');
   writeln(outfile,'*   D dimensionality');

   reset(infile);

   T := 2 - ln(3)/ln(2);

   while not eof(infile) do begin
      if eoln(infile)
      then readln(infile)
      else begin (* non-blank line *)
         (* read a data line *)
         blength := 0;
         while not eoln(infile) and (blength < bufmax) do begin
            blength := blength + 1;
            read(infile, buffer[blength]);
         end;
         readln(infile);

         (* analyze the data line *)
         (* scan the buffer for the bases *)
         Rs := 0.0;
         b := 1;
         bad := false;
         while b <= blength do begin
            c := buffer[b];

            if c = 'A' then Rs := Rs + 2;
            if c = 'C' then Rs := Rs + 2;
            if c = 'G' then Rs := Rs + 2;
            if c = 'T' then Rs := Rs + 2;

            if c = 'R' then Rs := Rs + 1; (* GA *)
            if c = 'Y' then Rs := Rs + 1; (* CT *)

            if c = 'M' then Rs := Rs + 1; (* AC *)
            if c = 'K' then Rs := Rs + 1; (* GT *)

            if c = 'S' then Rs := Rs + 1; (* CG *)
            if c = 'W' then Rs := Rs + 1; (* AT *)

            if c = 'B' then Rs := Rs + T; (* CGT *)
            if c = 'D' then Rs := Rs + T; (* AGT *)
            if c = 'H' then Rs := Rs + T; (* ACT *)
            if c = 'V' then Rs := Rs + T; (* ACG *)

            (* detect anomalous characters *)
            if not (c in ['A','C','G','T',
                          'R','Y','M','K','S','W',
                          'B','D','H','V',
                          '0','1','2','3','4','5','6','7','8','9',
                          'N',' ','^','-','/','(',')'])
            then bad := true;
            b := b + 1;
         end;
         if Rs = 0 then bad := true;

         (* put out the result *)
         if bad then write(outfile,'* '); (* force it to be a comment *)
{
         for w := 1 to nlength do write(outfile,name[w]);
}
         write(outfile,' ');
         for w := 1 to blength do write(outfile,buffer[w]);
         write(outfile,' ',(  Rs/2):wid:dec); (* compacted bases *)
         write(outfile,' ',(  Rs  ):wid:dec); (* bits *)
         write(outfile,' ',(2*Rs  ):wid:dec); (* D dimension *)
      end;

      (* last item so finish the line: *)
      writeln(outfile,' ',2*rs:wid:dec,'DIMENSION');
   end
end;
(* end module rebasec.themain *)

begin
   themain(enzymelist, data);
1: end.
