program ref2bib(refs,bib,output);
(* ref2bib: refer to bibtex converter
   Tom Schneider, 1988 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.46; (* of ref2bib 1988 December 14
origin 1988 jan 19 *)
(* end module version *)

(* begin module describe.ref2bib *)
(*
name
   ref2bib: refer to bibtex converter

synopsis
   ref2bib(refs: in, bib: out, output: out)

files
   refs: reference list used by refer
   bib: reference list used by bibtex
   output: messages to the user

description
   The program converts from refer to bibtex reference list formats.

documentation
   man refer, LaTeX reference manual

author
   Thomas Dana Schneider
   Frederick, Maryland
   toms@alum.mit.edu

bugs
   only a few of the refer types have been converted.
   see comments in the code for the entire list.

*)
(* end module describe.ref2bib *)

(* begin module refer.man *)
(*
a piece of the refer man page:
	  %A   Author's	name
	  %B   Book containing article referenced
	  %C   City (place of publication)
	  %D   Date of publication
	  %E   Editor of book containing article referenced
	  %F   Footnote	number or label	(supplied by _r_e_f_e_r)
	  %G   Government order	number
	  %H   Header commentary, printed before reference
	  %I   Issuer (publisher)
	  %J   Journal containing article
	  %K   Keywords	to use in locating reference
	  %L   Label field used	by -k option of	_r_e_f_e_r
	  %M   Bell Labs Memorandum (undefined) REPLACED BY MONTH
	  %N   Number within volume
	  %O   Other commentary, printed at end	of reference
	  %P   Page number(s)
	  %Q   Corporate or Foreign Author (unreversed)
	  %R   Report, paper, or thesis	(unpublished)
	  %S   Series title
	  %T   Title of	article	or book
	  %V   Volume number
	  %X   Abstract	- used by _r_o_f_f_b_i_b, not by _r_e_f_e_r
	  %Y,Z ignored by _r_e_f_e_r

     Except for	`A', each field	should be given	just once.  Only
     relevant fields should be supplied.  An example is:

	  %A   Bill Tuthill
	  %T   Refer - A Bibliography System
	  %I   Computing Services
	  %C   Berkeley
	  %D   1982
	  %O   UNX 4.3.5.

SEE ALSO
     refer(1), sortbib(1), roffbib(1), indxbib(1), lookbib(1)
*)
(* end module refer.man *)

(* begin module interact.const *)
      maxstring = 150; (* the maximum string *)
(* end module interact.const *)

type
(* begin module interact.type *)
      string = record (* a string of characters *)
         letters: array[1..maxstring] of char; (* the letters in the string *)
         length: integer; (* the number of characters in the string *)
         current: integer; (* the letter we are working on *)
      end;
(* end module interact.type *)

(* begin module ref2bib.type *)
   listptr = ^list;
   list = record (* list of strings *)
      s: string;
      next: listptr;
   end;
(* end module ref2bib.type *)

var
   refs, bib: text; (* refer input and the new bib input *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module interact.clearstring *)
procedure clearstring(var ribbon: string);
(* empty the string *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
      with ribbon do begin
         for index := 1 to maxstring do letters[index] := ' ';
         length := 0;
         current := 0;
      end
end; (* clearstring *)
(* end module interact.clearstring *)

(* begin module interact.writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module interact.writestring *)

(* begin module interact.getstring *)
procedure getstring(var afile: text; var buffer: string;
                    var gotten: boolean);
(* get a string from a file not using string calls.  this lets one
obtain lines from a file without interactive prompts *)
var   index: integer; (* of buffer *)
begin (* getstring *)
      clearstring(buffer);
      if eof(afile)
      then gotten := false
      else begin
         index := 0;
         while (not eoln(afile)) and (index < maxstring) do begin
            index := succ(index);
            read(afile, buffer.letters[index])
         end;

         if not eoln(afile) then begin
            writeln(output, ' getstring: a line exceeds maximum string size (',
                           maxstring:1,')');
            halt
         end;

         buffer.length := index;
         buffer.current := 1;
         readln(afile);
         gotten := true
      end
end; (* getstring *)      
(* end module interact.getstring *)

(* begin module interact.partstring *)
procedure partstring(var tofile: text; var s: string;
                     start: integer);
(* write part of the string s to file tofile
starting at 'start', no writeln *)
var   i: integer; (* index to s *)
begin
   with s do for i := start to length do write(tofile, letters[i])
end;
(* end module interact.partstring *)

(* begin module ref2bib.make *)
procedure make(var l: listptr);
(* make a new list element *)
begin
   if l = nil then begin
      new(l);
      clearstring(l^.s);
      l^.next := nil;
   end
   else clearstring(l^.s);
end;
(* end module ref2bib.make *)

(* begin module ref2bib.readentry *)
procedure readentry(var refs: text; var l: listptr);
(* read the entire entry into the list l *)
var
   c: listptr; (* current element of the list *)
   done: boolean; (* done finding an entry *)
   gotten: boolean; (* was a string gotten *)
begin
   c := l;
   done := false;
   while not done do begin
      getstring(refs,c^.s,gotten);
      done := (c^.s.length = 0) or (not gotten);
      if not done then begin
         make(c^.next);
         c := c^.next;
      end
   end
end;
(* end module ref2bib.readentry *)

(* begin module ref2bib.show *)
procedure show(var afile: text; var l: listptr; always: boolean);
(* show what is in the list l to a file.
if always is true, always show the strings, otherwise only show
the ones that have not been identified yet *)
var
   c: listptr; (* current element of the list *)
   missing: boolean; (* true if didn't catch something *)
begin
   (* determine if something is missing *)
   missing := false;
   c := l;
   while c^.s.length <> 0 do begin
      with c^.s do begin
         if letters[length]<>'!' then missing := true
      end;
      c := c^.next;
   end;

   if always or missing then writeln(afile,'comment{');
   if always or missing
   then writeln(afile,'current list is: ******************');

   c := l;
   while c^.s.length <> 0 do begin
      with c^.s do begin
         if always or (letters[length]<>'!')
         then writestring(afile,c^.s);

         if letters[length]<>'!'
         then write(afile,'<<<<<<<<<<<<<<<<<<');
      if always or missing then writeln(afile);
      end;
      c := c^.next;
   end;
   if always or missing then writeln(afile,'********************************');
   if always or missing then writeln(afile,'}');
end;
(* end module ref2bib.show *)

(* begin module ref2bib.got *)
procedure got(l: listptr);
(* change the string to show it was gotten *)
begin
   with l^.s do begin
      letters[length + 1] := ' ';
      letters[length + 2] := 'G';
      letters[length + 3] := 'O';
      letters[length + 4] := 'T';
      letters[length + 5] := ' ';
      letters[length + 6] := 'I';
      letters[length + 7] := 'T';
      letters[length + 8] := '!';
      length := length + 8;
   end
end;
(* end module ref2bib.got *)

(* begin module ref2bib.pluckauthor *)
procedure pluckauthor(l: listptr; var bib: text; id: char;
                      var close: boolean);
(* pluck the author (id=A) or editor (id=E) lines from the list l to bib *)
var
   c: listptr; (* current element of the list *)
   found: boolean; (* true if something was found *)
   outside: boolean; (* true until first author placed in bib *)
begin
   if close then writeln(bib,','); (* honor the last request *)
   if not(id in ['A','E']) then begin
      writeln(output,'pluckauthor: unknown identifier: ',id);
      halt
   end;
   found := false;
   outside := true;
   c := l;
   while c^.s.length <> 0 do begin
      with c^.s do begin
         if (letters[1]='%') then begin
            if letters[2] = id then begin
               if outside then begin
                  case letters[2] of
                     'A': write(bib,'author = "');
                     'E': write(bib,'editor = "');
                  end;
                  found := true;
                  outside := false
                  end
                  else begin
                     writeln(bib); (* move to new line *)
                     write(bib,' and ');
                  end;
               partstring(bib,c^.s,4);
               got(c);
            end
            else if not outside then begin (* close that one *)
               write(bib,'"');
               outside := true;
            end
         end
      end;
      c := c^.next;
   end;
   if not outside then write(bib,'"'); (* close shop *)
   if found then close := true (* request closing comma next time *)
            else close := false
end;
(* end module ref2bib.pluckauthor *)

(* begin module ref2bib.plucktitle *)
procedure plucktitle(l: listptr; var bib: text; id: char; 
                     what: char; var close: boolean);
(* pluck the title (id=T) and book (id=B) lines from the list l to bib,
what the publication is will determine what we call it *)
var
   c: listptr; (* current element of the list *)
   found: boolean; (* true if something was found *)
   outside: boolean; (* true until first title placed in bib *)
begin
   if close then writeln(bib,','); (* honor the last request *)
   if not(id in ['T','B']) then begin
      writeln(output,'plucktitle: unknown identifier: ',id);
      halt
   end;
   found := false;
   outside := true;
   c := l;
   while c^.s.length <> 0 do begin
      with c^.s do begin
         if (letters[1]='%') then begin
            if letters[2] = id then begin
               if outside then begin
                              case letters[2] of
                                 'T': write(bib,'title = "');
                                 'B': case what of
                                         'b': write(bib,'book = "');
                                         'i': write(bib,'booktitle = "');
                                         'a': begin
                                              writeln(output,'in plucktitle');
                                              halt
                                          end;
                                      end;
                              end;
                              found := true;
                              outside := false
                           end;
               partstring(bib,c^.s,4);
               got(c);
            end
            else if not outside then begin (* close that one *)
               write(bib,'"');
               outside := true;
            end
         end
         else if not outside then begin
            (* the line did not begin with %, was not empty and
               is inside: must be continuation of title *)
            writeln(bib); (* move from previous line *)
            writestring(bib,c^.s);
            got(c);
         end
      end;
      c := c^.next;
   end;
   if not outside then write(bib,'"'); (* close shop *)
   if found then close := true (* request closing comma next time *)
            else close := false
end;
(* end module ref2bib.plucktitle *)

(* begin module ref2bib.plucklines *)
procedure plucklines(l: listptr; var bib: text;
                     var close: boolean);
(* pluck the journal from l to bib *)
var
   c: listptr; (* current element of the list *)
begin
   if close then writeln(bib,','); (* honor previous request *)
   close := false;
   c := l;
   while c^.s.length <> 0 do begin
      with c^.s do begin
         if (letters[1]='%') then begin
            if letters[2] in ['C','D','I','J','M','P','V','Z'] then begin
               if close then writeln(bib,',');
               case letters[2] of
                  'C': write(bib,'address = "');
                  'D': write(bib,'year = "');
                  'I': write(bib,'institution = "');
                  'J': write(bib,'journal = "');
                  'M': write(bib,'month = "');
                  'P': write(bib,'pages = "');
                  'V': write(bib,'volume = "');
                  'Z': write(bib,'comment = "');
               end;
               partstring(bib,c^.s,4);
               got(c);
               write(bib,'"');
               close := true
            end
         end
      end;
      c := c^.next;
   end
end;
(* end module ref2bib.plucklines *)

(* begin module ref2bib.whatisit *)
procedure whatisit(l: listptr; var bib: text; var what: char);
(* what is the entry.  if '%T' exists, there is an article.
if '%B' exists, it considered to be a book or in a book*)
var
   article: boolean; (* title of article found *)
   book: boolean; (* book id found *)
   c: listptr; (* current element of the list *)
begin
   write(bib,'@');

   c := l;
   book := false;
   article := false;
   while c^.s.length <> 0 do begin
      with c^.s do begin
         if (letters[1]='%') then begin
            if letters[2] in ['B','T'] then begin
               case letters[2] of
                  'B': book := true;
                  'T': article := true
               end
            end
         end
      end;
      c := c^.next;
   end;

   if book and article then begin
      write(bib,'inbook');
      what := 'i'
   end
   else if book then begin
      write(bib,'book');
      what := 'b'
   end
   else if article then begin
      write(bib,'article');
      what := 'a'
   end
   else begin
      writeln(output,'whatisit: unidentified thing in: (assuming article)');
      show(output,l,true);
      write(bib,'article');
      what := 'a'
   end;
   write(bib,'{');
end;
(* end module ref2bib.whatisit *)

(* begin module ref2bib.crush *)
function crush(s: string): integer;
(* find the last word in the string *)
var i: integer; (* index *)
begin
   with s do begin
      i := length;
      while (i>1) and (letters[i-1]<>' ') do i := i - 1
   end;
   crush := i
end;
(* end module ref2bib.crush *)

(* begin module ref2bib.makekey *)
procedure makekey(l: listptr; var bib: text;
                  var close: boolean);
(* make up a key for the entry, a difficult task!! *)
var
   a: listptr; (* points at the first author *)
   d: listptr; (* points at the first date *)
   c: listptr; (* current element of the list *)
begin
   (* find an author or editor *)
   a := nil;
   d := nil;
   c := l;
   while c^.s.length <> 0 do begin
      with c^.s do begin
         if (letters[1]='%') then begin
            if letters[2] in ['A','E']
               then if a = nil then a := c;
            if letters[2] in ['D']
               then if d = nil then d := c
         end
      end;
      c := c^.next;
   end;

   if a <> nil
   then partstring(bib,a^.s,crush(a^.s))
   else begin
      writeln(output,'missing AUTHOR or EDITOR in');
      show(output,l,true);
      halt
   end;

   if d <> nil
   then partstring(bib,d^.s,crush(d^.s))
   else begin
      writeln(output,'missing DATE in');
      show(output,l,true);
      halt
   end;

   close := true; (* request closing comma in next stage *)
end;
(* end module ref2bib.makekey *)

(* begin module ref2bib.endentry *)
procedure endentry(var bib: text);
(* end the entry *)
begin
   writeln(bib,'}');
   writeln(bib);
end;
(* end module ref2bib.endentry *)

(* begin module ref2bib.themain *)
procedure themain(var refs,bib: text);
(* the main procedure of the program *)
var
   close: boolean; (* need to close this by a comma still *)
   count: integer; (* number of entries found *)
   l: listptr; (* points to the list of strings *)
   what: char; (* what kind of publication it is *)
begin
   writeln(output,'ref2bib ',version:4:2);
   reset(refs);
   rewrite(bib);
   l := nil; (* force creation of the first one *)
   make(l);
   count := 0;

   while not eof(refs) do begin
      readentry(refs,l);
      count := count + 1;

      whatisit(l,bib,what);
      makekey(l,bib,close);

      pluckauthor(l,bib,'A',close);
      pluckauthor(l,bib,'E',close);
      plucktitle (l,bib,'T',what,close);
      plucktitle (l,bib,'B',what,close);
      plucklines (l,bib,close);
      show(bib,l,false); (* only show the unidentified ones *)
      endentry(bib)
   end;
   writeln(output,count:1,' entries');
end;
(* end module ref2bib.themain *)

begin
   themain(refs,bib);
1: end.
