program spray(sprayp, drops, output);
(* spray: spray a genome with random locations

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of spray.p 2003 Aug 22
2003 Aug 22, 1.02: improve documentation
2003 Aug 22, 1.01: improve documentation
2003 Aug 21, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.spray *)
(*
name
   spray: spray a genome with random locations

synopsis
   spray(sprayp: in, drops: out, output: out)

files

   drops:  n random numbers between 1 and G.

   sprayp:  parameters to control the program.  The file must contain the
      following parameters, one per line:
      
      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      G (integer): size of the genome

      n (integer): number of sites to spray onto the genome.

description

   For a genome of G bases, place n sites randomly on the genome. 
   This is represented by a set of n numbers in the drops file.  Of
   course if one runs this through diffint and then genhis, the result
   should be a Poisson distribution.

examples

documentation

see also

   {example parameter file: } sprayp

   {Compute differences of sequential integers: } difint.p

   {Plot a series of numbers as a histogram: } genhis.p

   {Compute the poisson directly: } poisson.p

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.spray *)

var
   sprayp, (* file used by this program *)
   drops: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module random *)
procedure random(var seed: real);
(* random generator 2.  version = 1.00; 1986 December 31
   Test this routine with the program tstrnd.
   written by David Masternarde *)
 
(* This random number generator is based on a shift register with a single bit
of feedback, as described in Electronics for Neurobiologists, by Brown, Maxfield
and Moraff, MIT press 1973, referencing Random Process Simulation and
Measurement by Korn, McGraw-Hill 1966.  The random seed rand, a number between 0
and 1 exclusive, is converted to an integer between 1 and 2**23-1, inclusive.
This 23-bit number is shifted right one bit and the output of the last (23rd)
bit and the 9th bit are added modulo 2 (exclusive orred) and fed back into the
new first bit.  This is done between 4 and 11 times, depending on the last 3
bits of the original number.  The result is converted back to a real number
between 0 and 1 from which the 23 bit integer can be recovered on the next call.
The 23-bit shift register goes through all 2**23-1 values before repeating;
the repetition frequency of this algorithm could be less or greater depending
on the seed, because of the random number of multiple shifts per call. *)
(* powers of 2 *)
const pow14=16384; pow15=32768; pow22= 4194304; pow23=8388608;
var iseed, (* integer shift register *)
    i, nrep: integer;  (* index, number of times to do shift *)
begin (* random *)
      iseed := round(seed*pow23);  (* convert to 23 bit number *)
      if (iseed<1) or (iseed>=pow23) then iseed := 1;
      nrep := 4 + iseed mod 8;  (* do it 4 to 11 times based on last 3 bits *)
      for i:= 1 to nrep do
         (* if last bit and 9th bit are equal, feed back a 0, otherwise a 1 *)
         if( odd(iseed) = ((iseed mod pow15) >= pow14))
            then iseed := iseed div 2
            else iseed := (iseed div 2) + pow22;
      seed := iseed/pow23;
end; (* random *)
(* end module random version = 1.02; (@ of gauss.p 1994 sep 5 *)

(* begin module spray.themain *)
procedure themain(var sprayp, drops: text);
(* the main procedure of the program *)
const
{
   debugging = true; (* set true to debug *)
}
   debugging = false; (* set true to debug *)
var
   parameterversion: real; (* parameter version number *)
   d: integer; (* index to n *)
   G: integer; (* genome size *)
   l: integer; (* location of a site *)
   n: integer; (* number of sites *)
   r: real; (* a random number from the local generator *)
begin
   writeln(output,'spray ',version:4:2);
   reset(sprayp);
   readln(sprayp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(sprayp, G);
   readln(sprayp, n);

   rewrite(drops);
   writeln(drops,'* spray ',version:4:2);
   writeln(drops,'* size of genome:  G ',G:4);
   writeln(drops,'* number of sites: n ',n:4);

   if debugging then begin
      n := 10;
   end;

   writeln(output,'Starting generation of drops ...');

   r := 0.5;
   for d := 1 to n do begin
      (*
      The sun random number generator repeats terribly!!
      Don't use it!
      r := random(0);
      *)

      random(r);
      l := trunc(r*G);
      writeln(drops,l:10);

      if debugging then begin
         writeln(output,'* random number r ',r:8:5);
         writeln(output,round(r*G):10);
      end;

   end;
   writeln(output,'... end of generation of drops');

end;
(* end module spray.themain *)

begin
   themain(sprayp, drops);
1: end.
