program ssbread(ssb, report, output);
(* ssbread: read a sample sheet from the ABI sequencer

  Tom Schneider
  network address: toms@alum.mit.edu


  1993
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.27; (* of ssbread.p 1993 January 29
origin 1993 Jan 11 *)
(* end module version *)

(* begin module describe.ssbread *)
(*
name
   ssbread: read a sample sheet from the ABI sequencer

synopsis
   ssbread(ssb: in, report: out, output: out)

files
   ssb: A sample sheet from the ABI Sequencer
      Each sample must have plasmid and primer names in the
      sample name section.
   report:  reading of the identifying date, sample number, plasmid and primer
      names There are two parts to the report.  In the first, the program
      locates sections of the file by coordinate, and shows what it finds
      where.  In the second, it reads the data which tod will read (using the
      identical procedures.

   output: messages to the user

description
   The program allows one to test the reading modules for program tod.

examples

documentation

see also
   tod.p

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.ssbread *)

(* begin module tod.const *)
   maxseq = 1000; (* largest sequence the program can handle *)
   maxname = 30; (* largest name the program can handle *)
   maxfillname= 10; (* largest name the program can handle *)
   quicksequence = 25; (* number of bases to display on initial reading *)
(* end module tod.const version = 3.00; (@ of tod.p 1993 January 28 *)

type
   block = record
      data: packed array[1..1500] of char;
      length: integer;
   end;

(* begin module tod.type *)
   fillname = array[1..maxfillname] of char; (* name for filling names *)
   name = record
      letters: array[1..maxname] of char; (* name *)
      length: integer;  (* the length of the name *)
   end;
   seqptr = ^sequence;
   sequence = record (* a sequence *)
      bases: array[1..maxseq] of char; (* the sequence itself *)
      length: integer;  (* the length of the sequence *)
      lanenumber: integer; (* lane number *)
      plasmidname: name; (* plasmid name *)
      primername: name; (* primer used *)
      date: name; (* date of sequencing *)
      start: integer; (* first base to print *)
      stop: integer; (* last base to print *)
      editdeleted: boolean; (* true only if the sequence was deleted
         by editing commands *)
      next:  seqptr; (* next sequence in the data structure *)
   end;
   locptr = ^locations;
   locations = record (* a list of integer locations *)
      spot: integer; (* on spot *)
      next: locptr (* the next location *)
   end;
(* end module tod.type version = 3.00; (@ of tod.p 1993 January 28 *)

var
   ssb, report: text; (* a file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 3.00; (@ of tod.p 1993 January 28 *)

procedure readsegment(var f: text; fromplace, toplace: integer;
                      var b: block; var l: integer);
(* read from one place to another place
characters from f and put into b.  add n to counter l.
The block may have nulls internally, but the n+1th character is
forced to be a null. *)
var
   j: integer; (* location in a block *)
begin
  if l > fromplace then begin
     reset(f);
     l := 1; (* ready to read position 1 *)
  end;
  while l < fromplace do begin
     get(f);
     l := succ(l);
  end;
  for j := fromplace to toplace do read(f,b.data[j-fromplace+1]);
  b.length := toplace - fromplace + 1;
  b.data[b.length + 1] := chr(0);
  l := l + b.length;
end;

procedure uptonull(var f: text; b: block);
(* write to file f charcters in b, but stop before any null. *)
var
   j: integer; (* location in a block *)
begin
   j := 1;
   while (ord(b.data[j]) <> 0) and (j <= b.length) do begin
      write(f,b.data[j]);
      j := j + 1;
   end;
end;

procedure skipnulluptonull(var f: text; b: block);
(* write to file f charcters in b, but stop before any null. *)
var
   j: integer; (* location in a block *)
   nonulls: boolean; (* no null characters found yet *)
begin
   j := 1;
   nonulls := true;
   (* skip null characters *)
   while (ord(b.data[j]) = 0) and (j <= b.length) do begin
      if nonulls then begin
         write(f,'(NULLs found)');
         nonulls := false;
      end;
      j := j + 1;
   end;
   while (ord(b.data[j]) <> 0) and (j <= b.length) do begin
      write(f,b.data[j]);
      j := j + 1;
   end;
end;

procedure grabandshow(var f,t: text; fromplace, toplace: integer;
                      var b: block; var l: integer);
(* grab the segment fromplace to toplace from file f and show it to t.
block b is for efficiency of storage.  Keep the location of the file
in mind in l. *)
begin
   readsegment(f,fromplace,toplace,b,l);
   write(t,'from ',fromplace:5);
   write(t,' to ',toplace:5,' | ');
   write(t,'"');
(*   uptonull(t,b); *)
   skipnulluptonull(t,b);
   write(t,'"');
   writeln(t);
end;

(*************************************************)

(* begin module tod.bar *)
procedure bar(var f: text; n: integer; c: char);
(* write a bar of n characters c to f *)
var i: integer; (* index to the characters *)
begin
  for i := 1 to n do write(f,c);
end;
(* end module tod.bar version = 3.00; (@ of tod.p 1993 January 28 *)

(* begin module tod.clearname *)
procedure clearname(var n: name; f: fillname);
var i: integer;
begin
   with n do begin
      for i := 1 to maxfillname do letters[i] := f[i];
      length := maxfillname;
   end;
end;
(* end module tod.clearname version = 3.00; (@ of tod.p 1993 January 28 *)

(* begin module tod.copyname *)
procedure copyname(n1: name; var n2: name);
(* copy name n1 into name n2 *)
var
   i: integer; (* index *)
begin
   n2.length := n1.length;
   for i := 1 to n1.length do n2.letters[i] := n1.letters[i]
end;
(* end module tod.copyname version = 3.00; (@ of tod.p 1993 January 28 *)
 
(* begin module tod.clearseq *)
procedure clearseq(var seq: sequence);
(* clear the sequence *)
begin
   with seq do begin
      length := 0;
      lanenumber := 0;
      clearname(plasmidname,'plasmid?  ');
      clearname( primername,'primer?   ');
      clearname(       date,'date?     ');
      start := 0;
      stop := 0;
      editdeleted := false;
      next := nil;
   end;
end;
(* end module tod.clearseq version = 3.00; (@ of tod.p 1993 January 28 *)


(* begin module tod.readname *)
procedure readname(var f: text; var n: name; var location: integer);
(* read the name n from file f.  Names end in space, null or slash *)
var
   c: char; (* a character in f *)
begin
   n.length := 0;
   (* skip leading empty characters *)
{
   while (ord(f^) = 0) or (f^ in [' ','/']) do begin
}
   while (f^ in [' ','/']) do begin
{
      if ord(f^) = 0 then with n do begin (* VERY SPECIAL CASE *)
         length := succ(length);
         letters[length] := 'p';
         writeln(output,'CORRECTING NULL FIRST CHARACTER OF NAME TO "p"');
      end;
}
      get(f);
      location := succ(location);
   end;
   with n do begin
      repeat
         read(f,c);
         location := succ(location);
         if (c <> chr(0)) and (c <> ' ') and (c <> '/') then begin
            length := succ(length);
            letters[length] := c;
         end
      until (c = chr(0)) or (c = ' ') or (c = '/') or eoln(f)
   end;
end;
(* end module tod.readname version = 3.00; (@ of tod.p 1993 January 28 *)

(* begin module tod.writename *)
procedure writename(var f: text; n: name);
(* write the name n to file f *)
var
   i: integer; (* index to the name *)
begin
   with n do begin
      for i := 1 to length
      do write(f, letters[i]);
   end;
end;
(* end module tod.writename version = 3.00; (@ of tod.p 1993 January 28 *)

(* begin module tod.readssblane *)
procedure readssblane(var ssb: text;
                          lanenumber: integer;
                      var plasmidname, primername: name;
                      var location: integer);
(* read sample data from the ssb file.
Lane number lanenumber will be read.  If the data do not
match, the program will halt.  The plasmid and primer names
are obtained.  The variable location keeps track of the
location in the file.  It must be initialized to zero, and
then only this routine is allowed to change it. *)
const
   lanebot = 131;  (* location of the start of the lane # data *)
{  lanetop = 162;}  (* location of the end of the lane # data *)
   namebot = 163;  (* location of the start of the name data *)
{  nametop = 418;} (* location of the end of the name data *)
   lanelength = 586; (* length of the lane blocks *)
var
   numbers: set of '0'..'9';
   number1,number2: char; (* the characters of the lane number *)
function spot(a: integer): integer;
begin
   spot := a + lanelength*(lanenumber-1);
end;
procedure tolocation(a: integer);
(* move to the location a *)
begin
  if location > a then begin
     reset(ssb);
     location := 1; (* ready to read position 1 *)
  end;
  while location < a do begin
     get(ssb);
     location := succ(location);
  end;
{writeln(output,'tolocation: ',a:1);}
end;
begin
      numbers:=['0','1','2','3','4','5','6','7','8','9'];
   tolocation(spot(lanebot));
   while not (ssb^ in numbers) do begin
      get(ssb); location := succ(location);
   end;
   read(ssb,number1,number2); location := location + 2;
   if not (number2 in numbers) then begin
      writeln(output,'readssblane: bad number: "',number1,number2,'"');
      writeln(output,'(ords: ',ord(number1):1,' & ',ord(number2),')');
      writeln(output,'for lane number ',lanenumber:1);
      halt
   end;

   if (ord(number1) - ord('0')) * 10 + (ord(number2) - ord('0'))
      <> lanenumber then begin
      writeln(output,'readssblane: trying to read lane number ',lanenumber:1,
                     ' but found ',number1,number2);
      halt
   end;

   tolocation(spot(namebot));
   (* if the first character is nul, neither name exists *)
   if ord(ssb^) = 0
   then begin
      plasmidname.length := 0;
      primername.length := 0;
   end
   else begin
      readname(ssb, plasmidname, location);
      readname(ssb, primername, location);
   end;
end;
(* end module tod.readssblane version = 3.00; (@ of tod.p 1993 January 28 *)

(* begin module tod.readssb *)
procedure readssb(var ssb: text; now: name; var data: seqptr);
(* read the ssb table into the data structure *)
var
   d: seqptr; (* the data currently being read in *)
   location: integer; (* location in the ssb file *)
   lane: integer; (* lane number *)
begin
   clearseq(data^);
   d := data;
   copyname(now,d^.date);
   reset(ssb);
   location := 1;
   for lane := 1 to 24 do begin
      if lane > 1 then begin (* open space for the next data item *)
         new(d^.next);
         d := d^.next;
         clearseq(d^);
         copyname(now,d^.date);
      end;
      with d^ do begin
         readssblane(ssb, lane, plasmidname, primername,location);
         d^.lanenumber := lane;
      end;
   end;
end;
(* end module tod.readssb version = 3.00; (@ of tod.p 1993 January 28 *)

(* begin module tod.reportlanes *)
procedure reportlanes(var results: text; data: seqptr;
                      expectsequence: boolean);
(* Report all the lanes, and report on the DNA sequence if expectsequence
is true. *)
var
   d: seqptr; (* pointer to the data structure *)
   l: integer; (* index to the sequence *)
begin
   writeln(results);
   bar(results,80,'*'); writeln(results); (* all bars are 80 wide *)
   write(results,'*** Sample Sheet *');
   bar(results,62,'*'); writeln(results); (* all bars are 80 wide *)
   writeln(results);
   d := data; 
   while d <> nil do with d^ do begin
      write(results,'lane number: ',lanenumber:2);
      if plasmidname.length > 0 then begin
         write(results,'  plasmid: ');
         writename(results,plasmidname);
         write(results,'  primer: ');
         writename(results, primername);
      end;

      if expectsequence then if length > 0 then begin
         if plasmidname.length <= 0 then begin
            write(results,'  THAT''S ODD!  SEQUENCE WITHOUT A NAME:');
         end;
         write(results,' ');
         l := 1;
         while (l <= length) and (l <= quicksequence) do begin
            write(results,bases[l]);
            l := succ(l)
         end;
         if l < length then write(results,'...');
{zzz}
      end
      else write(results,' NO SEQUENCE DATA FOUND! ******');

      writeln(results);
      d := d^.next 
   end;
   writeln(results);
end;
(* end module tod.reportlanes version = 3.00; (@ of tod.p 1993 January 28 *)


(* begin module ssbread.themain *)
procedure themain(var ssb, report: text);
(* the main procedure of the program *)
const
   samplebot = 131;
   sampletop = 162;
   namebot = 163;
   nametop = 418;
   samplelength = 586;
var
   b: block;
   location: integer; (* location in the file *)

   samplenumber: integer; (* sample number *)

   data: seqptr; (* the sample sheet data structure *)
   now: name; (* dummy name *)

function spot(a: integer): integer;
begin
   spot := a + samplelength*(samplenumber-1);
end;
begin
   reset(ssb);
   rewrite(report);
   writeln(report,'ssbread ',version:4:2);
   writeln(output,'ssbread ',version:4:2);
   location := 1;
   grabandshow(ssb, report,3,64,b,location);
   writeln( report);

   for samplenumber := 1 to 24 do begin
      grabandshow(ssb, report,spot(samplebot),spot(sampletop),b,location); (* sample *)
      grabandshow(ssb, report,spot(namebot),spot(nametop),b,location); (* name *)
   end;

   writeln(report);

{zzz}
   new(data);
   readssb(ssb, now, data);
   reportlanes(report, data, false);

   write(output,'Name: "');
   readsegment(ssb,3,64,b,location);
   skipnulluptonull(output,b);
   write(output,'"');
   writeln(output);
   reportlanes(output, data, false);

{
   writeln(report);
   writeln(report,'The reason that the last one is missing is that');
   writeln(report,'the table is not straight!!');
   grabandshow(ssb,report,13642,13897,b,l); writeln(report);
   writeln(report,'(13641 has the character value of null)');
}

{
   grabandshow(ssb,report,131,162,b,l); writeln(report);
   grabandshow(ssb,report,163,418,b,l); writeln(report);

   grabandshow(ssb,report,717,748,b,l); writeln(report);
   grabandshow(ssb,report,749,1004,b,l); writeln(report);
}

{
   grabandshow(ssb,report,3,64,b,l); writeln(report);
   grabandshow(ssb,report,3,16,b,l); writeln(report);
   grabandshow(ssb,report,70,72,b,l); writeln(report);
   grabandshow(ssb,report,163,200,b,l); writeln(report);
   grabandshow(ssb,report,70,70,b,l); writeln(report);
   grabandshow(ssb,report,70,71,b,l); writeln(report);
}
end;
(* end module ssbread.themain *)

begin
   themain(ssb, report);
1: end.
