program tabify(input, output);
(* tabify: convert space delimited table to tab delmiters

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

const
(* begin module version *)
   version = 1.02; (* of tabify.p 2010 Sep 13
2010 Sep 13, 1.02: ignore leading blanks
2007 Aug 06, 1.01: fix documentation
2007 Apr 07, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.tabify *)
(*
name
   tabify: convert space delimited table to tab delmiters

synopsis
   tabify(input: in, output: out)

files

   input: file with space delimiting

   output: file with tab delimiting

description

   The xyplo and other programs in the Delila system use spaces
   to delimit tables.  Any number of spaces are allowed.  However,
   sometimes we would like to read a table into a word processor
   or spread sheet.  Tab delimiters make this easier.

   New as of 2010 Sep 13:  leading blank characters are ignored so
   the first column is really treated as a first column.

examples

documentation

see also

   {Related method: } rembla.p rb.p

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.tabify *)

procedure themain(var fin, fout: text);
(* the main procedure of the program *)
const
   tab = '	'; (* the tab character *)
var
   blankchars, (* the number of blank characters currently counted *)
   tabchars, (* the number of tab characters currently counted *)
   blanklines, (* the number of blank lines currently counted *)
   totalblankchars, (* the actual number of blanks removed *)
   index: (* a counter for blanks and lines *)
      integer;
   wasblank: boolean; (* was a line blank? *)
   firstblank: boolean; (* is this a blank at the beginning of the line? *)
begin (* tabify *)
{
   writeln(output,' tabify ',version:4:2);
}

   blanklines := 0;
   totalblankchars := 0;

(* the mechanism of the program is very simple:  copy fin to fout.  when
blanks appear, do not write them to fout, but keep track of how many
there are.  if another character appears on the line, then first write
any blanks out before the character.  if no other characters appear
on the line, we forget to write out the blanks...
   the same principle is applied to lines, so that extra blank lines
at the end of the file will not be written out. *)

   while not eof(fin) do begin
      blankchars := 0;
      tabchars := 0;
      wasblank := true;
      firstblank := true;

      (* copy a line *)
      while not eoln(fin) do begin
         if fin^ = ' '
         then blankchars := succ(blankchars)
         else if fin^ = tab
              then tabchars := succ(tabchars)
              else begin (* handle non-tab and non-space characters *)
                 if blanklines <> 0 then begin
                    for index := 1 to blanklines do writeln(fout);
                    blanklines := 0 (* reset *)
                 end;

                 if blankchars + tabchars <> 0 then begin
{
                    for index := 1 to blankchars do write(fout,' ');
}
{
                    if firstblank
                    then firstblank := false (* ignore initial blanks! *)
                    else write(fout,tab);
firstblank := false; (* ignore initial blanks! *)
}
                    if not firstblank then write(fout,tab);
                    blankchars := 0; (* reset *)
                    tabchars := 0; (* reset *)
                 end;
firstblank := false; (* ignore initial blanks! *)

{
                 if tabchars <> 0 then begin
                    for index := 1 to tabchars do write(fout,' ');
                    tabchars := 0 (* reset *)
                 end;
}

                 write(fout, fin^);
                 wasblank := false
              end;
         get(fin)
      end;

      (* count the ones left over *)
      totalblankchars := totalblankchars + blankchars;

      if wasblank
      then blanklines := succ(blanklines)
      else writeln(fout); (* the actual end of that line *)
      readln(fin)
   end;

{
      write(output,' ',totalblankchars:1,' blank');
      plural(output, totalblankchars);
      writeln(output,' removed');

      write(output,' ',blanklines:1,' blank line');
      plural(output, blanklines);
      writeln(output,' at the end removed')
}
end;

begin
   themain(input, output);
end. (* tabify *)
