program timenul(output);
(* timenul: modules to neutralize the date-time functions

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

*)

label 1;  (* the end of the program *)

const
(* begin module version *)
   version = 1.11; (* of timenul.p 2016 Jan 27
2016 Jan 27, 1.11 fix beg/end
2016 Jan 16, 1.10 clean out package!
2016 Jan 16, 1.09 don't call date()!
2000 Jul 25, 1.08 name changed from nulldate to timenul
2000 Feb 18, 1.05 upgrade types
1995 sep 5 previous changes
origin 1991 Mar 6 *)
(* end module version *)

(* begin module describe.timenul *)
(*
name
   timenul: modules to neutralize the date-time functions

synopsis
   timenul(output: out)

files
   output: where the (neutralized) date and time will appear.

description
   If transportation of a program or translation to C is hindered by the
   presence of the date-time modules, then one may want to blank out the
   function of those modules for the time being.  Thus all the dates produced
   will be zero, but one will be able to run the programs.   Nulldate
   contains modules that will replace corresponding modules in the other
   module libraries which are system dependent.  This will allow easy
   transportation of the Delila system to other computers.

   The output time string provided is '0000/00/00 00:00:00'.

documentation
   moddef, delman.describe.module

see also
   delman.describe.delmod, moddef, delman.describe.module
   delmod.p, prgmod.p, matmod.p

   {General discussion on compiling Delila programs:} 
   https://alum.mit.edu/www/toms/delila.html#How.To.Compile 

author
   Thomas D. Schneider

bugs
   none known

technical notes
   The datetime package required a const 'datetimearray' and a type
   'datetimearray'.  These are part of the book.const and book.type modules
   of delmod, and are identical to those types and consts.  Note:  programs
   which use the datetime package must have these types and consts either
   from delmod or manually declared.

*)
(* end module describe.timenul *)

(* begin module datetime.const *)
   datetimearraylength = 19; (* length of dataarray for dates,
       It is just long enough to include the 4 digit year - solving the
       year 2000 problem:
1980/06/09 18:49:11
123456789 123456789 
         1         2
*)
(* end module datetime.const version = 7.40; {of delmod.p 2000 Feb 18} *)

type

(* begin module datetime.type *)
   (* array for dates *)
   datetimearray = packed array[1..datetimearraylength] of char;
(* end module datetime.type version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module timenul.halt *)
procedure halt;
 (* halt the program and notify the user.
 the module name is not the standard one so that this procedure will not be
 transfered when timenul is used as a modlib. this avoids the possibility
 of transfering an obsolete halt module into a main module library
 unknowingly. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module timenul.halt *)

(* begin module package.datetime *)
(* ************************************************************************ *)

(* begin module getdatetime *)
(* end module getdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module readdatetime *)
(* end module readdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module writedatetime *)
(* end module writedatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module timeseed *)
(* end module timeseed version = 7.40; {of delmod.p 2000 Feb 18} *)
 
(*[[*)
(* begin module limitdate *) 
(* end module limitdate version = 7.40; {of delmod.p 2000 Feb 18} *)

(*]]*)
(* ************************************************************************ *)
(* end module package.datetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module getdatetime *)
procedure getdatetime(var adatetime: datetimearray);
(*  Empty the adatetime array INSTEAD of doing the following:
---
get the date and time into a single array from the system clock.
   adatetime contains the date:
       1980/06/09 18:49:11
         ye mo da ho mi se
(year, month, day, hour, minute, second).
As of 2000 February 18, the Sun Pascal compiler requires a formatting
statement.  This statement allows the date to be generated in this
standard Delila format in a single call.  Information about the
formatting statement is available on the manual page for date in Unix.
If a computer does not have this method, see the 'oldgetdatetime' routine
in delmod.p (https://alum.mit.edu/www/toms/delila/delmod.html)
for some conversion code.
*)
var
   i: integer; (* index to adatetime *)
begin
   (* date(adatetime,'%Y/%m/%d %H:%M:%S'); *)
   for i := 1 to datetimearraylength
   do adatetime[i] := '0';
   adatetime[5]  := '/';
   adatetime[8]  := '/';
   adatetime[11] := ' ';
   adatetime[14] := ':';
   adatetime[17] := ':';
end;
(* end module getdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module readdatetime  *)
procedure readdatetime (var thefile: text; var adatetime: datetimearray);
(* read the date and time from the file *)
var
      index: integer; (* to the udatetime *)
      (* the following is an unpacked date time array, to avoid reading
      into a packed array.  reading into a packed array is not transportable *)
      udatetime: array[1..datetimearraylength] of char;
begin
      for index:=1 to datetimearraylength do read(thefile,udatetime[index]);
      pack(udatetime, 1, adatetime);
      if (adatetime[3]='/') and (adatetime[12]=':') then begin
         writeln(output,' old datetime (only 2 year digits) read: ',
                          adatetime:datetimearraylength);
      end;
end;
(* end module readdatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module writedatetime *)
procedure writedatetime(var thefile: text; adatetime: datetimearray);
(* expand the date and time out and print in the file *)
var
      index: integer; (* index of datetime *)
begin
      for index:=1 to datetimearraylength do write(thefile,adatetime[index])
end;
(* end module writedatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

procedure themain (var fout: text);
(* main routine *)
var
   adatetime: datetimearray;  (* this is a global variable for use in the running
                           of timenul *)
begin
   rewrite(fout);
   writeln(fout,'timenul ',version:4:2);
   writeln(output,'getdatetime(adatetime);');
   getdatetime(adatetime);
   writeln(output,'writedatetime(adatetime);');
   writedatetime(fout,adatetime);
   writeln(output);
   writeln(output,'done timenul');
end;

begin  (* timenul *)
   themain (output);
1:end.   (* timenul *)
