program titer(plates,result,verbose,output);
(* analyse titertek optical density data
   by Gary Stormo, 1991 *)

label 1;  (* end of program *)

(* begin module version *)
const   version = 2.29;  (* of titer.p 1993 Jan 27
origin before 1986 august 1 *)
(* end module version *)

(* begin module describe.titer *)
(*
name
      titer: analyse titertek optical density data

files
      titer(plates: in, result: out, verbose: out, output: out)

synopsis
      plates: output of the tk program; containing a header line,
         and a series of plates that describe the names of wells,
         their optical densities at various times and od620 data.
      result: a tabulation of the beta-galactosidase values.
      verbose: more detail on how the calculations were done.
      output: messages to the user.

description
      take data from titertek plates and do the analysis:
        get sample names from id plate, find duplicates
        read in volume values
        read in od620 values
        read in od414 values for each time point
                calculate best slope from all time points
                calculate activity for each sample
        report beta-galactosidase data
           st.dev. is the standard deviation for the samples
           % dev.  is 100*st.dev./activity.  If this is larger than 10%
                   for 4 samples, we usually redo the measurement.

examples
        titer.plates  example input file
        titer.result  example output file
        titer.verbose  example output file

see also
        beta-galactosidase assay protocol
        tk program (in advanced IBM basic)
        tkod.p program

author
        Gary Stormo

*)
(* end module describe.titer *)

(*  more constants *)
        namelength = 10; (* maximum length of names *)
        rows = 8; (* number of wells in a row on a plate *)
        columns = 12; (* number of wells in a column on a plate *)
        platesize = 96; (* rows * columns; area of a plate *)
        percent = '%'; (* the percent symbol *)
type
        wellarray = array[1..platesize] of real; (* readings at time points *)
        
        nametype = array [1..namelength] of char;

        data = record    (* activity data *)
           ave: real;
           sd: real;
        end;

        meas = ^measurement;  (* linked list of measurements *)
        measurement = record
           time: real;
           od414: wellarray;
           next: meas;
       end;
        
        position = record   (* on plate *)
           row : char;
           column: integer;
        end;

        wellptr = ^well;   (* list of wells for each sample *)
        well = record
           pos: position;
           next: wellptr;
        end;

        samplelist = ^sample;  (* list of samples on plate *)
        sample = record
           name: nametype;
           condition: nametype;
           wells: wellptr;
           activity: data;
           next: samplelist;
        end;

        platelist = ^plate;  (* list of plates in file *)
        plate = record
           number: integer;
           name: nametype;
           volume: wellarray;
           od620: wellarray;
           assays: meas;
           slope: wellarray;   (* for data in each well *)
           cc: wellarray;     (* corr. coef. for data to slope *)
           samples: samplelist;
           gotod620: boolean; (* whether the od620 was read *) 
           next: platelist;
        end;


var     
    plates,  (* the input plates file *)
    verbose,  (* verbose output, telling the activities at each time *)
    result: text;  (* the output result file *)
    p: platelist;  (* the list of plates *)

(* begin module halt *)
procedure halt; 
(* stop the program.  the procedure performs a goto to the end of the   
   program.  you must have a label: 
      label 1;
   declared, and also the end of the program must have this label:  
      1: end. 
   examples are in the module libraries.
   this is the only goto in the delila system. *) 
begin 
      writeln(output,' program halt.'); 
      goto 1
end;  
(* end module halt *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text); 
(* copy a line from file fin to file fout *)
begin (* copyaline *) 
      while not eoln(fin) do begin  
         fout^ := fin^;   
         put(fout);   
         get(fin) 
      end;  
      readln(fin);  
      writeln(fout);  
end; (* copyaline *)
(* end module copyaline version = "delmod 6.21 84 may 19 tds/gds"; *) 

(* begin module titer.getname *)

procedure getname(var fin: text; var name: nametype);
(* read the name from the file fin.  names are characters less than 
   "namelength" long, separated by blanks  *)
var i,j: integer;  (* an index *)
    done: boolean;  (* when name end is reached *)
    ch: char;    (* part of the name *)
begin
           (* get to first non-blank *)
        repeat
                read(fin,ch)
        until (ch <> ' ');
           (* read in name until next blank *)
        i := 0;
        done := false;
        while ((ch <> ' ') and (not done)) do begin
                i := succ(i);
                if (i <= namelength) then name[i] := ch;
                if eoln(fin) then done := true 
                else read(fin,ch);
        end;
        for j := (i+1) to namelength do name[j] := ' ';
end;

(* end module titer.getname *)


(* begin module titer.writename *)
procedure writename(var fout: text; name: nametype);
(* write the name to the file fout *)
var i: integer; (* index to the name *)
begin
    for i := 1 to namelength do write(fout,name[i]);
end;
(* end module titer.writename *)

(* begin module titer.header *)
procedure header(var fin, result, verbose: text);
(* copy the plate name and date from the input file 
   to the output files result and verbose.  *)
begin
      writeln(result,' titer ',version:4:2,'  results of analysis of:');
      reset(fin);
      copyaline(fin,result);
      copyaline(fin,result);
      copyaline(fin,result);
      writeln(result);
      write(result,'isolate    ');
      write(result,'condition  ');
      write(result,'plate      ');
      write(result,'   activity');
      write(result,'    st.dev.');
      write(result,'     ',percent,' dev.');
      writeln(result);
      writeln(result);

      writeln(verbose,' titer ',version:4:2,' verbose analysis of:');
      reset(fin);
      copyaline(fin,verbose);
      copyaline(fin,verbose);
      copyaline(fin,verbose);
      writeln(verbose);
end;
(* end module titer.header *)

(* begin module titer.getids *)

procedure getids(var fin: text; var s: samplelist);
(* make the list of which samples are in which wells, from the 
  well id plate. *)
const debugging = false;  (* after development *)
type  welldata = record
         name: nametype;
         condition: nametype;
         row: char;
         column: integer;
      end;

      namearray = array[1..platesize] of welldata;

var   i,j: integer;  (* indecies of for loops *)
      num,           (* column number *)
      k: integer;   (* well array element *)
      well: namearray;  (* the array of well names, conditions and positions *)

procedure getcond(var w: namearray);
(* get the condition name from the plate, if it exists. 
   if it exists it must follow the name id plate, and is distinguishable
   by the word "condition".  if it doesn't exist, put the name "std" as
   the condition.  the volume plate will
   be there instead, and the word "volumes" will be seen. *)
var i,j,   (* indices *)
    num,   (* column number *)
    k:  integer;   (* well array index *)
    ch: char;    (* to see if its "v" or "c" *)
procedure skiptodata;
(* skip to distinguishing line of the plate *)
begin
    readln(fin); (* skip first line of plate *)
    readln(fin); (* skip type line of plate *)
        (* "c" or "v" will be third character, if condition or volume plate *)
    get(fin);
    get(fin);
    readln(fin,ch);
        (* skip to data lines, whatever they are *)
    readln(fin);
if debugging then writeln(output,ch);
end; (* skiptodata *)
begin
    skiptodata;
    if (ch = 'c') then begin (* read condition names *)
        for i := 1 to columns do begin
                read(fin,num);
                for j:= 1 to rows do begin
                        k := j + rows*(i-1);
                        getname(fin,w[k].condition);
                end;
                readln(fin);
        end;
        readln(fin); (* skip blank line between plates *)
        skiptodata;
     end
     else
        for i := 1 to platesize do begin
              w[i].condition[1] := 's';
           w[i].condition[2] := 't';
           w[i].condition[3] := 'd';
           for j := 4 to namelength do w[i].condition[j] := ' ';
        end;
     if ch <> 'v' then begin
        writeln(output,'you are missing a volume plate');
        halt
     end
end;

procedure addwells(w: welldata; var sawell: wellptr);
(* add new well to list, and fill with position data *)
begin
        new(sawell);
        sawell^.pos.row := w.row;
        sawell^.pos.column := w.column;
        sawell^.next := nil;
end;

procedure makelist(well: namearray; var s: samplelist);
(* take the array of well names and positions and reduce it to a list
   in which each name appears once, with a list of the well postions 
   for that sample name. *)
var n: nametype;  (* to store the name for comparison *)
    i,j: integer;  (* index *)
    lastwell,  (* last well in list *)
    w: wellptr;  (* for new wells on list *)
    lastsa,   (* last sample on list *)
    sa: samplelist;  (* for making new samples *)
    ch : char;  (* for name comparing *)
    newsamp: boolean;  (* true if new sample name or condition *)
begin
                (* read first name and position into list *)
        new(sa);
        sa^.next := nil;
        lastsa := sa;
        s := sa;
        for i := 1 to namelength do sa^.name[i] := well[1].name[i];
if debugging then writename(output,sa^.name);
        for i := 1 to namelength do sa^.condition[i] := well[1].condition[i];
if debugging then writename(output,sa^.condition);
        addwells(well[1],w);
        sa^.wells := w;
        lastwell := w;
                (* add new positions for old names, and get new
                   names and positions *)
        for j := 2 to platesize do begin
            newsamp := false;
                (* check if this is a new condition *)
               i := 0;
            repeat
                i := succ(i);
                ch := well[j].name[i];
            until ((ch <> sa^.name[i]) or (i = namelength));
            if (ch <> sa^.name[i]) then  (* new name *)
                newsamp := true
                else begin
                    (* check if new condition *)
                      i := 0;
                   repeat
                        i := succ(i);
                        ch := well[j].condition[i];
                   until ((ch <> sa^.condition[i]) or (i = namelength));
                   if (ch <> sa^.condition[i]) then  (* new name *)
                        newsamp := true;
                end;
            if newsamp then begin
                new(sa);
                sa^.next := nil;
                lastsa^.next := sa;
                lastsa := sa;
                for i := 1 to namelength do
                    sa^.name[i] := well[j].name[i];
if debugging then writename(output,sa^.name);
                for i := 1 to namelength do
                    sa^.condition[i] := well[j].condition[i];
if debugging then writename(output,sa^.condition);
                addwells(well[j],w);
                sa^.wells := w;
                lastwell := w
            end
            else begin
                addwells(well[j],w);
                lastwell^.next := w;
                lastwell := w;
            end;
        end;
end;


begin  (* getids *)
      for i := 1 to columns do begin
         read(fin,num);  (* column number *)
         for j := 1 to rows do begin
               (* this definition of k makes the array be in "column order";
                  that is, the first column makes the first c elements of
                  the array.  if "row order" is desired then make:
                        k := i + columns*(j-1);
                  to minimize sorting.                      *)
            k := j + rows*(i-1);
if debugging then writeln(output,'row=',i:1,'column=',j:1,'k=',k:1);
            getname(fin,well[k].name);
            well[k].column := i;
               (* the first row is a, second b ......  *)
            well[k].row := chr(ord('a') + j - 1); 
         end;
         readln(fin);
      end;
      readln(fin);  (* skip blank after plate data *)
      getcond(well); 
      makelist(well,s);
end;

(* end module titer.getids *)

(* begin module titer.getnumbers *)
procedure getnumbers(var fin: text; var w: wellarray);
(* get the numbers from the file and return in w *)
var  
     i,j,n: integer;  (* index *)

begin
    for i := 1 to columns do begin
        read(fin,n);
        for j := 1 to rows do
            read(fin,w[j + rows*(i-1)]);
        readln(fin);
    end; 
    readln(fin);
end;

(* end module titer.getnumbers *)

(* begin module titer.getvol *)
procedure getvol(var fin: text; var p: platelist);
(* get the volume data from the file and put it into the platelist *)
var  
     i,j,n: integer;  (* index *)

begin
    getnumbers(fin,p^.volume);
end;

(* end module titer.getvol *)

(* begin module titer.avedata *)
procedure avedata(var p: platelist);
(* get the slope from the input plates;
   calculate the activity as:
      slope * 60 * 1000 / (od620 * volume)
   calculate the average and sd of the activity and put into p^.samples *)

var  r,c,   (* row and column in plate *)
     i: integer;  (* index *)
     wpos: integer;  (* position in wellarray *)
     act, sumact,  (* values of data points and their sum *)
     sumsqdact : real;  (* sum of the data points squared *)
     w: wellptr;  (* pointer to wells of sample *)
     sa: samplelist;  (* current sample *)
begin
   sa := p^.samples;
   while (sa <> nil) do begin
      sumact := 0;
      sumsqdact := 0;
      i := 0;
      w := sa^.wells;
      while (w <> nil) do begin
         c := w^.pos.column;
         r := ord(w^.pos.row) - ord('a') + 1;
         i := succ(i);
         wpos := r + (c-1)*rows;

         if ((p^.od620[wpos] * p^.volume[wpos]) > 0.0001)
         then
            act := p^.slope[wpos] * 60000 / (p^.od620[wpos] * p^.volume[wpos])
         else
            act := 0;

         sumact := sumact + act; 
         sumsqdact := sumsqdact + act*act;
         w := w^.next;
      end;
      sa^.activity.ave := sumact / i;
         (* calculate the standard deviation of the data *)
      sa^.activity.sd := sqrt((sumsqdact / i) - (sumact/i) * (sumact/i)); 
      sa := sa^.next;
   end;
end;

(* end module titer.avedata *)


(* begin module titer.getod620 *)

procedure getod620(var fin: text; var p: platelist);
(* get the od620 data from the file and put it into the platelist *)
var  
     i: integer;  (* index *)

begin
       (* skip to data lines *)
    for i := 1 to 3 do readln(fin);
    getnumbers(fin,p^.od620);
    p^.gotod620 := true
end;

(* end module titer.getod620 *)

(* begin module titer.getdata *)

procedure getdata(var fin: text; var p: platelist);
(* get the od414 data from the file and put it into the platelist *)
const debugging=false;
var  
     i: integer;  (* index *)
     lastassay,    (* pointer to end of data set string *)
     assay: meas;  (* pointer to new data set *)

begin
       (* read the time of the data point *)
    get(fin);  (* skip '*' in file *)
    new(assay);
    if (p^.assays = nil) then p^.assays := assay
    else begin
       lastassay := p^.assays;
       while (lastassay^.next <> nil) do lastassay := lastassay^.next;
       lastassay^.next := assay;
    end;
    assay^.next := nil;
    readln(fin,assay^.time);
if debugging then writeln(output,'time ',assay^.time:6:2);
    readln(fin);   (* skip row names *)
    getnumbers(fin,assay^.od414);
end;

(* end module titer.getdata *)

(* begin module linear.regression.modified *)  
(* linear regression procedure  
      1980 august 5 
      by thomas schneider 
      copyright 1980 *) 
  
procedure regress ( control: char; (* described below *)  
                    x, y: real;    (* the data pairs *) 
                var sumx, sumy, sumxsqd, sumysqd, sumxy, (* internal records *) 
(* mean, variance and covariance of x and y: *) 
                    ex, ey, varx, vary, covxy,
(* correlation coefficient: *)
                    r,
(* m = slope, b = y intercept as in y = mx + b *) 
                    m, b  
                        : real; 
(* number of data pairs entered: *) 
                var n: integer);
(* linear regression for variables x and y. 
   the control variable has only three acceptable states: 
      c clear all variables to zero 
        (note that r is set to 2.0, so that a program can check whether 
         or not the results have been calculated yet. the recommended test
         is not for equality, since round-off could cause problems.  use  
              if r > 1.5 then (no calculations yet) 
                         else (calculations done)    )
      e enter the x and y values into the internal sums 
      r results are calculated from the sums
*)
begin 
      if not( control in ['c','e','r']) then begin  
         writeln(output,' linear regression control variable value, "',control, 
                        '" is not acceptable.');
         writeln(output,' it must be in ["c","e","r"]');  
         halt 
      end 
  
      else case control of  
         'c': begin (* clear *) 
            x:=0.0; y:=0.0; 
            sumx:=0.0; sumy:=0.0; 
            sumxsqd:=0.0; sumysqd:=0.0; 
            sumxy:=0.0; 
            ex:=0.0; ey:=0.0; 
            varx:=0.0; vary:=0.0; 
            covxy:=0.0; 
            r:=2.0; (* this (impossible) value can act as a flag that 
                       shows that the calculations have not been made. *) 
  
            m:=0.0; b:=0.0; 
            n:=0  
         end; 
  
         'e': begin (* enter data *)
            sumx:=sumx + x; 
            sumy:=sumy + y; 
            sumxsqd:=sumxsqd + x*x; 
            sumysqd:=sumysqd + y*y; 
            sumxy:=sumxy + x*y; 
            n:=succ(n)
         end; 
  
         'r': begin (* calculate results *) 
            ex:=sumx/n; 
            ey:=sumy/n; 
  
            varx:=(sumxsqd/n) - ex*ex;
            vary:=(sumysqd/n) - ey*ey;
  
            covxy:=(sumxy/n) - ex*ey; 
 (* modified version of module *) 
            if ((varx * vary) > 0.0001) then
            r:=covxy/(sqrt(varx * vary)) 
            else r := 0;
  

 (* modified version of module *) 
            if (varx > 0.0001) then
            m:=covxy/varx
            else m := 0;
            b:=ey - m*ex
         end
      end 
end;  
(* end module linear.regression.modified *)  


(* begin module titer.writedata *)
procedure writedata(p: platelist; var result,verbose: text);
(* write the data from the platelist to the files *)
const debugging=false;
var row: char;  (* of plate *)
    column,     (* of plate *)
    i: integer;  (* an index into wellarray *)
    act: real;  (* activity, calculated from slope *)
    a: meas;  (* for time point measurements *)
    s: samplelist;  (* current samplelist from the platelist *)
begin
    while (p <> nil) do begin
        s := p^.samples;
        while (s <> nil) do begin
                   (* write name, condition, plate, activity and st. dev. 
                      for each sample to the result file *)
                writename(result,s^.name);
                write(result,' ');
                writename(result,s^.condition);
                write(result,' ');
                writename(result,p^.name);
                write(result,' ');
                with s^.activity do begin
                   write(result,ave:10:2);
                   write(result,' ');
                   write(result,sd:10:2);
                   write(result,' ');
                   if ave > 0.0 then writeln(result,round(100*sd/ave):10)
                                else writeln(result,' ':10);
                end;

                   (* write name and condition and well list for each sample,
                      plus the activity and st.dev. *)
                writeln(verbose);   
                writename(verbose,s^.name);
                write(verbose,' ');
                writename(verbose,s^.condition);
                write(verbose,' ');
                writename(verbose,p^.name);
                writeln(verbose,' is averaged from wells:');
                writeln(verbose,' well  activity  correlation coefficient');
                    (* write the id wells to verbose *)
                while (s^.wells <> nil) do begin
                   column := s^.wells^.pos.column;
                   row := s^.wells^.pos.row;
                   write(verbose,'   ',row,column:1);
                      (* write the slope and cc for each well *)
                   i := (column -1) * rows + ord(row) - ord('a') + 1;
   if debugging then write(verbose,'  array pos = ',i:1);
                   if not p^.gotod620 then begin
                      writeln(output,'an od620 plate is missing');
                      halt
                   end;
                   if ((p^.od620[i] * p^.volume[i]) > 0.0001) then
                     act := p^.slope[i] * 60000 / (p^.od620[i] * p^.volume[i])
                   else act := 0;
                   write(verbose,'  ',act:8:2);
                   write(verbose,'  ',p^.cc[i]:8:2);
                    writeln(verbose);
                   s^.wells := s^.wells^.next;
                end;
                write(verbose,'actvity =',s^.activity.ave:8:2);
                writeln(verbose,'   st. dev. = ',s^.activity.sd:6:3);
                s := s^.next;
        end;
        p := p^.next;
    end;
end;
(* end module titer.writedata *)


(* begin module titer.getplatenumber *)
procedure getplatenumber(var fin:text; var n: integer);
(* read the plate number from the file fin *)
var ch: char;
begin
    (* number occurs after second blank *)
    repeat read(fin,ch) until ch = ' ';
    repeat read(fin,ch) until ch = ' ';
    read(fin,n);
end;
(* end module titer.getplatenumber *)

(* begin module titer.gettype *)
procedure gettype(var fin: text; var t: char);
(* get the type of plate data by reading the appropriate char *)
begin
    get(fin);
    get(fin);
    readln(fin,t);
end;
(* end module titer.gettype *)

(* begin module titer.skipplate *)
procedure skipplate(var fin:text);
(* skip the contents of one plate *)
const debugging=false;
var i: integer;
begin
    if debugging then writeln(output,'skipping plate');
    for i := 1 to (columns + 3) do readln(fin);
end;
(* end module titer.skipplate *)

(* begin module titer.readplates *)
procedure readplates(var fin: text; var p: platelist);
(* read input file fin into the platelist p, which points to beginning *)
const debugging=false;
var i,      (* index *)
    n,     (* platenumber *)
    lastn, (* highest platenumber seen *)
    currentn: integer;  (* plate currently being worked on *)
    t: char;  (* plate type *)
    lastp,          (* last plate in list *)
    pa: platelist;  (* pointer to current plate *)
    assay: meas;   (* an assay on the plate *)

begin  (* readplate *)
    lastn := 0;
    currentn := 0;
    while not eof(fin) do begin
        getplatenumber(fin,n);
if debugging then writeln(output,'platenumber ',n:1);
        if (n > lastn) then begin  (* new plate, get names *)
            lastn := n;
            currentn := n;
            new(pa);
            if (n = 1) then p := pa
            else begin  (* add pa to end of list *)
               lastp := p;
               while (lastp^.next <> nil) do lastp := lastp^.next;
               lastp^.next := pa;
            end;
            pa^.number := n;
            pa^.next := nil;
            pa^.assays := nil;
            pa^.samples := nil;
            pa^.gotod620 := false;
            getname(fin,pa^.name);
if debugging then writename(output,pa^.name);
            for i := 1 to 4 do readln(fin);
                (* get new plate names and conditions, if any *)
            getids(fin,pa^.samples);
               (* volume must follow name plate *)
            getvol(fin,pa)
        end
        else begin 
            readln(fin);
                (* get to the plate for the current number *)
            if (currentn <> n) then begin
                pa := p;
                while (pa^.number < n) do pa := pa^.next;
                currentn := n;
            end;
                (* find out what type of plate it is *) 
            gettype(fin,t);
 if debugging then writeln(output,'platetype ',t); 
            if (t = 'd') then getod620(fin,pa)
            else if (t = 't') then getdata(fin,pa)
            else begin
                writeln(output,'invalid platetype ',t);
                writeln(output,'REMAINDER OF THE INPUT FILE:');
                while not eof(fin) do copyaline(fin,output);
                halt;
           end;
        end;
    end;  (* while *)
    if debugging then begin
       pa := p;
       while (pa <> nil) do begin
          write(output,pa^.number,'  ');
          writename(output,pa^.name);
          writeln(output);
          assay := pa^.assays;
          while (assay <> nil) do begin
             write(output,assay^.time:6);
             assay := assay^.next;
          end;
          writeln(output);
          pa := pa^.next;
        end;
    end;
end;

(* end module titer.readplates *)

(* begin module titer.analyze *)
procedure analyze(var p: platelist);
(* analyze the data for each plate in p.
   put the activity for each sample in the samplelist for each plate. *)
const debugging=false;
        stopdiff = 0.3;  (* decrease in od414 that signifies stopping *)
var  i: integer;
     a: meas;  (* a particular assay data point *)
     pa: platelist;  (* the current plate *)
        (* variables needed for regression calls *)
     done : boolean;  (* true when analysis should stop *)
     x,y,   (* dummy data variables *)
     lastx, lasty,  (* previous values, for determining stopping analysis *)
     sx,sy,sxs,sys,sxy,  (* sumx, sumy, sumx**2, sumy**2, sumxy *)
     ex,ey,  (* means of the data variables *)
     vx,vy,covxy,  (* variance of data variables and covariance *)
     cc,m,b: real;  (* corr. coef., slope and intercept of regression line *)
     n: integer;  (* number of data points *)
     column: integer;  (* from plate *)
     row: char;  (* from plate *)
begin
     pa := p;
if debugging then writeln(output,' beginning regressions ');
     while (pa <> nil) do begin
        for i := 1 to platesize do begin
           lasty := 0;
           done := false;
              (* clear regression variables *)
           regress('c',x,y,sx,sy,sxs,sys,sxy,ex,ey,vx,vy,covxy,cc,m,b,n);
           a:=  pa^.assays;
           repeat
              x := a^.time;
              y := a^.od414[i];
              if ((lasty - y) > stopdiff) then begin  
                 (* y value has decreased significantly, stop analysis *)
                 done := true;
                 write(verbose,' plate ');
                 writename(verbose,pa^.name);
                   (* calculate row and column of stopped analysis *)
                 column := trunc((i-1)/rows) + 1;
                 row := chr(ord('a') + (i-1) mod rows);
                 write(verbose,'  well ',row,column:1);
                 writeln(verbose,' analysis stopped after time ',
                                 round(lastx):4, ' seconds')
              end
              else begin
                 lastx := x;
                 lasty := y;
                 regress('e',x,y,sx,sy,sxs,sys,sxy,ex,ey,vx,vy,covxy,cc,m,b,n);
                 a := a^.next;
                 if (a = nil) then done := true;
              end;
           until done;
if debugging then writeln(output,'time points =',n:2);
           regress('r',x,y,sx,sy,sxs,sys,sxy,ex,ey,vx,vy,covxy,cc,m,b,n);
           pa^.slope[i] := m;
           pa^.cc[i] := cc; 
        end;
if debugging then writeln(output,'regression done for plate ',pa^.number:2);
          (* calculate the average slope for the wells from the same sample,
             put into samplelist along with sd  *)
        avedata(pa);
        pa := pa^.next;
     end;
end;
(* end module titer.analyze *)

begin   (* titer *)

        rewrite(result);
        rewrite(verbose);
        writeln(output,'titer ',version:4:2);
        header(plates,result,verbose);
        readplates(plates,p);
        analyze(p);
        writedata(p,result,verbose);  

1: end.
